
# HMS_Data_Build.R
# -----------------------------------------------------------------------------
#
# An R script that rebuilds the HMS data.

library(data.table)
library(bit64)

input_directory = "/kilts/nielsen_extracts/HMS"
output_directory = "/dir" # Change accordingly
maxyear = 2016 
minyear = 2010

# Build Trip Data

print("Building Trips")
trips = NULL

for (year in minyear:maxyear){
  print(year)
  temp = fread(sprintf("%s/%s/Annual_Files/trips_%s.tsv", input_directory, year, year),
               showProgress = FALSE)
  trips = rbindlist(list(trips, temp), use.names = TRUE, fill = TRUE)
  
  # The method of payment column only appears after 2013, so we have to fill
  # rows before that year with NA
  
  rm(temp)
}

print("Changing Column Orders and Setting Keys")
setcolorder(trips, c("household_code", "retailer_code", "purchase_date",
                       "trip_code_uc", "store_code_uc", "total_spent",
                       "method_of_payment_cd", "panel_year", "store_zip3"))

setkey(trips, household_code, retailer_code, purchase_date)

print("Saving")
save(list = "trips", file = sprintf("%s/trips.RData", output_directory))

rm(trips)

# Build Panelists Data

print("Building Panelists")
panelists = NULL

for (year in minyear:maxyear){
  print(year)
  temp = fread(sprintf("%s/%s/Annual_Files/panelists_%s.tsv", input_directory, year, year),
               showProgress = FALSE)
  panelists = rbindlist(list(panelists, temp), use.names = TRUE, fill = TRUE)
  rm(temp)
}

print("Turning Codes into Factors")

# Turn the Demographic Variables into Factors
# Adapted from James Sams's scripts and minor corrections made to match the manual
# Household_income needs special handling as the meaning of the code changes
# over the years

Variable_Labels = list(
  household_size = c('Single Member'=1, 'Two Members'=2, 'Three Members'=3,
                     'Four Members'=4, 'Five Members'=5, 'Six Members'=6,
                     'Seven Members'=7, 'Eight Members'=8, 'Nine+ Members'=9),
  type_of_residence = c('One Family House'=1,
                        'One Family House (Condo/Coop)'=2,
                        'Two Family'=3, 'Two Family House (Condo/Coop)'=4,
                        'Three+ Family House'=5,
                        'Three+ Family House (Condo/Coop)'=6,
                        'Mobile Home or Trailer'=7),
  household_composition = c('Married'=1, 'Female Head Living with Others Related'=2,
                            'Male Head Living with Others Related'=3,
                            'Female Living Alone'=5,
                            'Female Living with Non-Related'=6,
                            'Male Living Alone'=7,
                            'Male Living with Non-Related'=8),
  age_and_presence_of_children = c('Under 6 only'=1, '6-12 only'=2,
                                   '13-17 only'=3, 'Under 6 & 6-12'=4,
                                   'Under 6 & 13-17'=5, '6-12 & 13-17'=6,
                                   'Under 6 & 6-12 & 13-17'=7,
                                   'No Children Under 18'=9),
  head_age = c('No Such Head'=0, 'Under 25 Years'=1, '25-29 Years'=2,
               '30-34 Years'=3, '35-39 Years'=4, '40-44 Years'=5,
               '45-49 Years'=6, '50-54 Years'=7, '55-64 Years'=8,
               '65+ Years'=9),
  head_employment = c('No Such Head'=0, 'Under 30 hours'=1, '30-34 hours'=2,
                      '35+ hours'=3, 'Not Employed for Pay'=9),
  head_education = c('No Such Head or Unknown'=0, 'Grade School'=1,
                     'Some High School'=2, 'Graduated High School'=3,
                     'Some College'=4, 'Graduated College'=5,
                     'Post College Grad'=6),
  head_occupation = c('No Such Head'=0, 'Professional'=1,
                      'Prop, Managers, Officials'=2, 'Clerical'=3, 'Sales'=4,
                      'Craftsman/Foreman(Skilled)'=5,
                      'Operative(Semi-Skilled)'=6, 'Military'=7,
                      'Service Workers&Private HH Workers'=8,
                      'Farmer (Manager, Owner, Worker)'=9,
                      'Students Employed <30 hours'=10, 'Laborers'=11,
                      'Retired, Unemployed'=12),
  marital_status = c('Unknown'=0, 'Married'=1, 'Widowed'=2,
                     'Divorced/Separated'=3, 'Single'=4),
  race = c('White'=1, 'Black'=2, 'Asian'=3, 'Other'=4),
  hispanic_origin = c('Hispanic'=1, 'Not Hispanic'=2),
  region_code = c('New England'=1, 'Middle Atlantic'=2,
                  'East North Central'=3, 'West North Central'=4,
                  'South Atlantic'=5, 'East South Central'=6,
                  'West South Central'=7, 'Mountain'=8, 'Pacific'=9),
  kitchen_appliances = c('Microwave Only'=1, 'Dishwasher Only'=2,
                         'Garbage Disposal Only'=3,
                         'Microwave & Dishwasher'=4,
                         'Microwave & Garbage Disposal'=5,
                         'Dishwasher & Garbage Disposal'=6,
                         'Microwave, Dishwasher, & Garbage Disposal'=7,
                         'None'=8, 'Not Available'=9),
  tv_items = c('No Television'=0, 'Television with No Cable'=1,
               'Reg Cable Only'=2, 'Reg & Pay Cable'=3),
  household_internet_connection = c('Yes'=1, 'No'=2),
  wic_indicator_current = c('On WIC Now'=1),
  wic_indicator_ever_notcurrent = c('On WIC Previously'=1),
  member_relationship_sex = c('Son'=1, 'Daughter'=2, 'Other Male Relative'=3,
                              'Other Female Relative'=4, 'Male Not Related'=5,
                              'Female Not Related'=6),
  member_employment = c('Employed'=1)
)

# Copy Necessary but Missing Labels

for (i in 1:8){
  Variable_Labels[[paste0("Member_",i,"_Relationship_Sex")]] = Variable_Labels[["member_relationship_sex"]]
  Variable_Labels[[paste0("Member_",i,"_Employment")]] = Variable_Labels[["member_employment"]]
}

for (sex in c("male", "female")){
  Variable_Labels[[paste0(sex, "_head_age")]] = Variable_Labels[["head_age"]]
  Variable_Labels[[paste0(sex, "_head_employment")]] = Variable_Labels[["head_employment"]]
  Variable_Labels[[paste0(sex, "_head_education")]] = Variable_Labels[["head_education"]]
  Variable_Labels[[paste0(sex, "_head_occupation")]] = Variable_Labels[["head_occupation"]]
}


# Now turn number codes into factors

colnums = which(colnames(panelists) %in% names(Variable_Labels))

for (i in colnums){
  label = Variable_Labels[[colnames(panelists)[i]]]
  panelists[[i]] = factor(panelists[[i]], levels = label, labels = names(label))
}

# This applies to 2006-2009
household_income1 = c('-$5000'=3, '$5000-$7999'=4, '$8000-$9999'=6,
                     '$10,000-$11,999'=8, '$12,000-$14,999'=10,
                     '$15,000-$19,999'=11, '$20,000-$24,999'=13,
                     '$25,000-$29,999'=15, '$30,000-$34,999'=16,
                     '$35,000-$39,999'=17, '$40,000-$44,999'=18,
                     '$45,000-$49,999'=19, '$50,000-$59,999'=21,
                     '$60,000-$69,999'=23, '$70,000-$99,999'=26,
                     '$100,000 - $124,999'=27, '$125,000 - $149,999'=28,
                     '$150,000 - $199,999'=29, '$200,000 + '=30)

# This applies to all other years
household_income2 = c('-$5000'=3, '$5000-$7999'=4, '$8000-$9999'=6,
                      '$10,000-$11,999'=8, '$12,000-$14,999'=10,
                      '$15,000-$19,999'=11, '$20,000-$24,999'=13,
                      '$25,000-$29,999'=15, '$30,000-$34,999'=16,
                      '$35,000-$39,999'=17, '$40,000-$44,999'=18,
                      '$45,000-$49,999'=19, '$50,000-$59,999'=21,
                      '$60,000-$69,999'=23, '$70,000-$99,999'=26,
                      '$100,000 + '=27)

panelists[,household_income := as.character(household_income)]
panelists[panel_year %in% 2006:2009, household_income:= 
            as.character(factor(household_income, levels = household_income1, 
                                labels = names(household_income1)))]
panelists[!panel_year %in% 2006:2009, household_income:= 
            as.character(factor(household_income, levels = household_income2, 
                                labels = names(household_income2)))]
panelists[,household_income := factor(household_income, 
                                      levels = unique(c(names(household_income2),
                                                        names(household_income1))))]

panelists[,fips_state_descr := factor(fips_state_descr)]
panelists[,fips_county_descr := factor(fips_county_descr)]
panelists[,scantrack_market_descr := factor(scantrack_market_descr)]
panelists[,dma_descr := factor(dma_descr)]

print("Setting Keys")
setkey(panelists, household_code, panel_year)

print("Saving")
save(list = "panelists", file = sprintf("%s/panelists.RData", output_directory))
rm(panelists)

# Build Purchases Data

print("Building Purchases")

# Load Products Data and extract Magnet UPCs
products = fread(sprintf("%s/Master_Files/Latest/products.tsv", input_directory),
                 showProgress = FALSE)

load(sprintf("%s/trips.RData", output_directory))
trips = trips[,.(trip_code_uc, household_code, retailer_code, purchase_date)]

purchases = NULL

for (year in minyear:maxyear){
  print(year)
  temp = fread(sprintf("%s/%s/Annual_Files/purchases_%s.tsv", input_directory, year, year),
               showProgress = FALSE)
  
  # Sanity Check : when the UPC's are not read in as 64 bit format
  if (class(temp$upc)!="integer64"){
    temp[, upc:= as.integer64(upc)] 
  } 
  
  temp[, panel_year := year]
  
  # The column deal_type only appears until 2013 so we should fill the missing values with NA
  purchases = rbindlist(list(purchases, temp), use.names = TRUE, fill = TRUE)
  rm(temp)
}

print("Appending additional information")
purchases = merge(purchases, trips, by = "trip_code_uc")
purchases[, trip_code_uc := NULL]
setcolorder(purchases, c("upc", "upc_ver_uc", "household_code", "retailer_code", 
                         "purchase_date", "total_price_paid", "quantity",
                         "coupon_value", "deal_flag_uc", "panel_year"))

print("Setting Keys")
setkey(purchases, household_code, purchase_date, upc, upc_ver_uc)

print("Saving")
save(list = "purchases", file = sprintf("%s/purchases.RData", output_directory))

# Save the first 100,000 rows separately for easy reference

purchases = head(purchases, 100000)
save(list = "purchases", file = sprintf("%s/purchases_subsample.RData", output_directory))

rm(purchases)

# Build Retailers Data

print("Building Retailers")

retailers = fread(sprintf("%s/Master_Files/Latest/retailers.tsv", input_directory),
                  showProgress = FALSE)
setkey(retailers, retailer_code)

save(list = "retailers", file = sprintf("%s/retailers.RData", output_directory))
rm(retailers)

print("Done!")
