
# productmeta.R
# -----------------------------------------------------------------------------
#
# An R script that will build a product level meta data
# using the tsv formatted Nielsen extracts files produced by
# the Kilts Center of Marketing. 

# Note: Nielsen RMS 2017 data had a revision for the week 07152017,
#       Hence the processing of 2017 data is changed accordingly.

library(data.table)
library(bit64)
library(foreach)
library(doParallel)

input_directory = "/kilts/nielsen_extracts/RMS"
output_directory = "/project/ghitsch/data/RMS-Build-2016" # Change appropriately
maxyear = 2014 # Change the latest year in the data accordingly
minyear = 2006

if (!dir.exists(output_directory)) {
  dir.create(output_directory,
             showWarnings = FALSE)
}

# Stack Store Data

if (!file.exists(sprintf("%s/Stores.RData", output_directory))){
  stores = NULL

  for (year in minyear:maxyear){

    if (year == 2017) {
      temp <- fread(paste0(input_directory,
                           "/2017_Week_07152017_Only/Annual_Files/stores_",
                           year,
                           ".tsv"))
    } else {
      temp <- fread(paste0(input_directory,
      					   "/",
      					   year,
      					   "/Annual_Files/stores_",
                           year,
                           ".tsv"))
    }

    stores <- rbind(stores, temp)
    rm(temp)

  }
  setkey(stores, store_code_uc, year)
  save(list = "stores",
       file = sprintf("%s/Stores.RData", output_directory))
  rm(stores)
}

# Stack UPC Versions Data

if (!file.exists(sprintf("%s/Versions.RData", output_directory))){
  versions = NULL

  for (year in minyear:maxyear){

    temp <- fread(sprintf("%s/%s/Annual_Files/rms_versions_%s.tsv",
                          input_directory, year, year))
    versions <- rbind(versions, temp)
    rm(temp)
  }

  setnames(versions, c("panel_year"), c("year"))

  setkey(versions, upc, year)
  save(list = "versions",
       file = sprintf("%s/Versions.RData", output_directory))
  rm(versions)
}

# Stack Product Extra Attribute Data

if (!file.exists(sprintf("%s/Products-Extra.RData", output_directory))){
  products_extra = NULL

  for (year in minyear:maxyear){

    if (year == 2017) {
      temp1 <- fread(paste0(input_directory,
                            "/",
                            year,
                            "/Annual_Files/products_extra_",
                            year,
                            ".tsv"))

      temp2 <- fread(paste0(input_directory,
                            "/2017_Week_07152017_Only/Annual_Files/products_extra_",
                            year,
                            ".tsv"))

      temp  <- funion(temp1, temp2)

      rm(temp1, temp2)

    } else {
      temp <- fread(sprintf("%s/%s/Annual_Files/products_extra_%s.tsv",
                            input_directory, year, year))
    }

    products_extra = rbind(products_extra, temp)
    rm(temp)
  }

  setnames(products_extra, c("panel_year"), c("year"))

  setkey(products_extra, upc, upc_ver_uc)
  save(list = "products_extra",
       file = sprintf("%s/Products-Extra.RData", output_directory))
  rm(products_extra)
}

# Build Product / Year Level Meta Data

products <- fread(paste0(input_directory,
                         "/Master_Files/Latest/products.tsv"),
                  showProgress = FALSE,
                  colClasses   = c(upc = "integer64"))


save(products, 
     file = paste0(output_directory, 
                   "/Products.RData"))

rm(products)

load(sprintf("%s/Stores.RData", output_directory))

# Define a function that will build the meta data

build_meta_data <- function(year, module){

  # Define a function that calculates the number of retail chains
  # in which a UPC has ever appeared for each year

  get_chain_num  <-  function(x){
    return(length(unique(stores[unique(x), parent_code])))
  }

  moduleno <- gsub(paste0("(.*)/(.*)_",
                          year,
                          ".tsv"),
                   "\\2", module) # Select the second (.*) of the string

  module_major <- gsub(paste0("(.*)/Movement_Files/(.*)"),
                       "\\2", module)

  print(moduleno)

  movement <-  fread(module,
                     showProgress = FALSE,
                     colClasses   = c(upc = "integer64"))

  if (year == 2017) {
    module2   <- paste0("/kilts/nielsen_extracts/RMS/2017_Week_07152017_Only/",
                        "Movement_Files/",
                        module_major)

    if (file.exists(file.path(module2))) {
      movement2 <- fread(module2,
                         showProgress = FALSE,
                         colClasses   = c(upc = "integer64"))

      # Check if 07152017 data can be matched to the original data
      stopifnot(unique(movement2$week_end) == "20170715")
      stopifnot(movement[week_end == "20170715", .N] == movement2[week_end == "20170715", .N])

      movement <- movement[week_end != 20170715]
      movement <- rbind(movement,
                        movement2)

      rm(movement2)
    }
  }

  # The date is read as numbers, so convert into the date format R recognizes
  movement[, week_end := as.Date(as.character(week_end), format='%Y%m%d')]


  RMS_info <-  movement[, .(
    revenue_RMS             = sum(price*units/prmult, na.rm = TRUE),
    first_date_observed_RMS = min(week_end),
    last_date_observed_RMS  = max(week_end),
    N_weeks_RMS             = as.numeric(max(week_end) - min(week_end))/7+1,
    N_stores_RMS            = length(unique(store_code_uc)),
    N_chains_RMS            = get_chain_num(.(store_code_uc, year))),
    # Year-specific in case a store moves to a different retail chain
    by = upc]

  return(RMS_info)
}

# Now Go Parallel, this part is modified for the new Mercury server.

Nslots <- as.numeric(Sys.getenv("SLURM_CPUS_ON_NODE"))
cat(as.character(Nslots), "\n")

for (year in minyear:maxyear){

  cl <- makeCluster(Nslots, outfile = "")

  registerDoParallel(cl)

  clusterExport(cl=cl,
                varlist = ls(.GlobalEnv))


  list_modules = list.files(sprintf("%s/%s/Movement_Files", 
  									input_directory, year),
                            full.names = TRUE, recursive = TRUE)

  ncases <- length(list_modules)

  RMS_output <- foreach(i = 1:ncases,
                        .packages = c("data.table",
                                      "bit64"),
                        .combine  = rbind) %dopar% {
                          build_meta_data(year,
                                          module = list_modules[i])
                        }
  stopCluster(cl)

  RMS_output[, "year" := year] # Add year info for future use

  save(list = "RMS_output", 
  	   file = sprintf("%s/RMS_%s.RData", output_directory, year))
  print(warnings())
  rm(RMS_output)
}


# Stack yearly info and attach upc version and module code

stacked = NULL

for (year in minyear:maxyear){
  load(sprintf("%s/RMS_%s.RData", output_directory, year))
  stacked = rbind(stacked, RMS_output)
  rm(RMS_output)
  # file.remove(sprintf("%s/RMS_%s.RData", output_directory, year))
}

load(sprintf("%s/Versions.RData", output_directory))

setkey(stacked, upc, year)
meta_data = merge(stacked, versions, all.x = TRUE)
meta_data = meta_data[, c(1,9,2:8), with = FALSE]

load(sprintf("%s/Products.RData", output_directory))

products = products[, c(1,2,4), with = FALSE] #Only leave upc, upc_ver_uc, and module code
setkey(products, upc, upc_ver_uc)
setkey(meta_data, upc, upc_ver_uc)
meta_data = merge(meta_data, products, all.x = TRUE)
setkey(meta_data, upc, upc_ver_uc, year)
save(list = "meta_data", file = sprintf("%s/Meta-Data.RData", output_directory))

print("Done!")
