
# movement.R
# -----------------------------------------------------------------------------
#
# An R script that will build R Formatted Movement Files based on the
# tsv formatted files produced by the Kilts center. Uses the product level
# meta data generated by productmeta.R

rm(list = ls())
library(data.table)
library(bit64)
library(doParallel)
library(lubridate)

input_directory = "/kilts/nielsen_extracts/RMS"
output_directory = "/project/ghitsch/data/RMS-Build-2016" # Change appropriately
maxyear = 2014 # Change the latest year in the data accordingly
minyear = 2006

load(sprintf("%s/Meta-Data/Versions.RData", output_directory))

# Threshold for qualifying as top UPC 
# (UPC that has ever made into the threshold in any year)
# Adjust appropriately

N = 75000

print("Loading Meta Data")
load(sprintf("%s/Meta-Data/Meta-Data.RData", output_directory))
meta_data[, rank_year    := rank(-revenue_RMS), by = year]
meta_data[, highest_rank := lapply(.SD, min), 
          by = upc, .SDcols = "rank_year"]

upcs = meta_data[highest_rank <= N, upc]
upcs = sort(unique(upcs)) #This is the list of UPCs that have ever made top N

# Obtain List of Modules and Groups

print("Obtaining List of Modules and Groups")

lists_grp_mod <- data.table()

for (year in minyear:maxyear){
  temp = list.files(sprintf("%s/%s/Movement_Files", input_directory, year), 
                    recursive  = TRUE,
                    full.names = FALSE)
  temp1 = gsub(paste0("(.*)_",
                      year,
                      "/(.*)_",
                      year,
                      ".tsv"), "\\2", temp) #Extract module code
  temp2 = gsub(paste0("(.*)_",
                      year,
                      "/(.*)_",
                      year,
                      ".tsv"), "\\1", temp) #Extract group code
  
  lists_grp_mod_sub <- data.table(
    module = temp1,
    group  = temp2
  )
  
  lists_grp_mod <- rbind(lists_grp_mod, lists_grp_mod_sub)
  
  rm(temp, temp1, temp2, lists_grp_mod_sub)
}

setkey(lists_grp_mod, module, group)

lists_grp_mod = unique(lists_grp_mod) # List of unique Product Group and Module Codes

unique_mod <- unique(lists_grp_mod[, module])
ncases     <- length(unique_mod)

# Create Directories for each module

print("Creating Directories")
for (modules in lists_grp_mod[,unique(module)]){
  print(sprintf("Creating Directory for Module %s", modules))
  dir.create(sprintf("%s/Modules/%s", output_directory, modules), showWarnings = FALSE)
}


# Define a function that will build the movement data

build_movement <- function(Module){
  # Some modules (e.g. 1362, 1484) have too many observations,
  # exceeding the limit of data.table, thus we split
  # time frame into [minyear, 2011], and [2012, maxyear]

  # As we have more and more data, this would probably happen for
  # more modules, thus I recommend splitting the time frame for all of the modules.

  unique_groups <- lists_grp_mod[module == Module]
  
  if (!(Module %in% c(1362, 1484))) {

    # Stack movement data for module
    stacked <- data.table()
    for (year in minyear:maxyear) {
      for (i in 1:nrow(unique_groups)) {
        Group <- unique_groups[i, group]
        
        move_file <- sprintf("%s/%s/Movement_Files/%s_%s/%s_%s.tsv", 
                             input_directory, year, Group, 
                             year, Module, year)
        
        if (file.exists(move_file)) {
          print(move_file)
          temp <- fread(move_file, 
                        showProgress = FALSE,
                        colClasses   = c(upc = "integer64"))
          
          if (year == 2017) {
            move_file2 <- sprintf("%s/%s/Movement_Files/%s_%s/%s_%s.tsv", 
                                  input_directory, 
                                  "2017_Week_07152017_Only", 
                                  Group, 
                                  year, Module, year)
            if (file.exists(move_file2)) {
              temp2 <- fread(move_file2, 
                             showProgress = FALSE,
                             colClasses   = c(upc = "integer64"))
              
              stopifnot(unique(temp2$week_end) == "20170715")
              stopifnot(temp[week_end == "20170715", .N] == temp2[week_end == "20170715", .N])
              
              temp <- temp[week_end != 20170715]
              temp <- rbind(temp,
                            temp2)
              
              rm(temp2)
            }
          }
          
          temp[, "year" := year]
          
          stacked <- rbind(stacked, temp)
          rm(temp)
        }
        
      }
    }
    
    setkey(stacked, upc, year)
    
    print(sprintf("Adding UPC versions for Module %s", Module))
    stacked = merge(stacked, versions, all.x = TRUE)
    
    stacked[, "year" := NULL]
    setcolorder(stacked, c("upc", "upc_ver_uc"))

    setkey(stacked, upc, upc_ver_uc, store_code_uc, week_end)
    
    # Saving Records of current module
    
    print(sprintf("Saving All Others in Module %s", Module))
    
    movement <- stacked[!(upc %in% upcs),]
    movement[, week_end := as.Date(as.character(week_end), format='%Y%m%d')]
    # We could have done this just once for the stacked file
    # But it tends to cause memory issues for some large modules
    save(list = "movement", file = sprintf("%s/Modules/%s/All-Other.RData", output_directory, Module))
    
    # Get unique upcs in module
    # for certain modules, unique(stacked$upc) would break because 
    # the number of upc exceeds the limit of unique()
    # Thus we break stacked into 10 groups
    upcs_in_module <- NULL
    try(upcs_in_module <- upcs[upcs %in% unique(stacked[,upc])])
    if (is.null(upcs_in_module)) {
      num_stacked <- stacked[, .N]
      split_size  <- 10
      group_size  <- num_stacked %/% split_size
      
      unique_stk_upc <- data.table()
      for (split in 1:split_size) {
        if (split != split_size) {
          sub_upc <- unique(stacked[(1 + (split - 1) * group_size):(split * group_size), 
                                    .(upc)])
        } else {
          sub_upc <- unique(stacked[(1 + (split - 1) * group_size):num_stacked, 
                                    .(upc)])
        }
        unique_stk_upc <- rbind(unique_stk_upc, sub_upc)
      }
      stopifnot(class(unique_stk_upc$upc) == "integer64")
      
      upcs_in_module = upcs[upcs %in% unique(unique_stk_upc[,upc])]
    }
    
    cat(length(upcs_in_module))
    
    if (length(upcs_in_module) > 0){ #There are Module that don't have any top UPCs
      
      for (j in 1:length(upcs_in_module)){
        upc_curr <- upcs_in_module[j]
        print(sprintf("Saving UPC %s in Module %s", upc_curr, Module))
        movement <- stacked[upc == upc_curr]
        movement[, week_end := as.Date(as.character(week_end), format='%Y%m%d')]
        save(list = "movement", file = sprintf("%s/Modules/%s/%s.RData", 
                                               output_directory, Module, upc_curr))
      }
    }
    rm(stacked)
    
    cat("Module",Module,"Done! \n")
  } else {
    stacked1 <- data.table()
    stacked2 <- data.table()
    
    for (year in minyear:2011) {
      for (i in 1:nrow(unique_groups)) {
        Group <- unique_groups[i, group]
        
        move_file <- sprintf("%s/%s/Movement_Files/%s_%s/%s_%s.tsv", 
                             input_directory, year, Group, 
                             year, Module, year)
        
        if (file.exists(move_file)) {
          print(move_file)
          temp <- fread(move_file, 
                        showProgress = FALSE,
                        colClasses   = c(upc = "integer64"))
          
          if (year == 2017) {
            move_file2 <- sprintf("%s/%s/Movement_Files/%s_%s/%s_%s.tsv", 
                                  input_directory, 
                                  "2017_Week_07152017_Only", 
                                  Group, 
                                  year, Module, year)
            if (file.exists(move_file2)) {
              temp2 <- fread(move_file2, 
                             showProgress = FALSE,
                             colClasses   = c(upc = "integer64"))
              
              stopifnot(unique(temp2$week_end) == "20170715")
              stopifnot(temp[week_end == "20170715", .N] == temp2[week_end == "20170715", .N])
              
              temp <- temp[week_end != 20170715]
              temp <- rbind(temp,
                            temp2)
              
              rm(temp2)
            }
          }
          
          temp[, "year" := year]
          
          stacked1 <- rbind(stacked1, temp)
          rm(temp)
        }
        
      }
    }
    
    for (year in 2012:maxyear) {
      for (i in 1:nrow(unique_groups)) {
        Group <- unique_groups[i, group]
        
        move_file <- sprintf("%s/%s/Movement_Files/%s_%s/%s_%s.tsv", 
                             input_directory, year, Group, 
                             year, Module, year)
        
        if (file.exists(move_file)) {
          print(move_file)
          temp <- fread(move_file, 
                        showProgress = FALSE,
                        colClasses   = c(upc = "integer64"))
          
          if (year == 2017) {
            move_file2 <- sprintf("%s/%s/Movement_Files/%s_%s/%s_%s.tsv", 
                                  input_directory, 
                                  "2017_Week_07152017_Only", 
                                  Group, 
                                  year, Module, year)
            if (file.exists(move_file2)) {
              temp2 <- fread(move_file2, 
                             showProgress = FALSE,
                             colClasses   = c(upc = "integer64"))
              
              stopifnot(unique(temp2$week_end) == "20170715")
              stopifnot(temp[week_end == "20170715", .N] == temp2[week_end == "20170715", .N])
              
              temp <- temp[week_end != 20170715]
              temp <- rbind(temp,
                            temp2)
              
              rm(temp2)
            }
          }
          
          temp[, "year" := year]
          
          stacked2 <- rbind(stacked2, temp)
          rm(temp)
        }
        
      }
    }
    
    setkey(stacked1, upc, year)
    setkey(stacked2, upc, year)
    
    print(sprintf("Adding UPC versions for Module %s", Module))
    stacked1 = merge(stacked1, versions, all.x = TRUE)
    stacked1 = stacked1[,c(1,10,3:9), with = FALSE]
    
    setkey(stacked1, upc, upc_ver_uc, store_code_uc, week_end)
    
    stacked2 = merge(stacked2, versions, all.x = TRUE)
    stacked2 = stacked2[,c(1,10,3:9), with = FALSE]
    
    setkey(stacked2, upc, upc_ver_uc, store_code_uc, week_end)
    # Saving Records of current module
    
    print(sprintf("Saving All Others in Module %s", Module))
    
    movement1 <- stacked1[!(upc %in% upcs),]
    movement1[, week_end := as.Date(as.character(week_end), format='%Y%m%d')]
    
    movement2 <- stacked2[!(upc %in% upcs),]
    movement2[, week_end := as.Date(as.character(week_end), format='%Y%m%d')]
    
    movement <- rbind(movement1, movement2)
    rm(movement1, movement2)
    
    # We could have done this just once for the stacked file
    # But it tends to cause memory issues for some large modules
    save(list = "movement", file = sprintf("%s/Modules/%s/All-Other.RData", output_directory, Module))
    rm(movement)
    
    upcs_in_module1 <- NULL
    try(upcs_in_module1 <- upcs[upcs %in% unique(stacked1[,upc])])
    if (is.null(upcs_in_module1)) {
      num_stacked1 <- stacked1[, .N]
      split_size1  <- 10
      group_size1  <- num_stacked1 %/% 10
      
      unique_stk_upc1 <- data.table()
      for (split in 1:10) {
        if (split!=10) {
          sub_upc <- unique(stacked1[(1 + (split - 1) * group_size1):(split * group_size1), 
                                    .(upc)])
        } else {
          sub_upc <- unique(stacked1[(1 + (split - 1) * group_size1):num_stacked1, 
                                    .(upc)])
        }
        unique_stk_upc1 <- rbind(unique_stk_upc1, sub_upc)
      }
      stopifnot(class(unique_stk_upc1$upc) == "integer64")
      
      upcs_in_module1 = upcs[upcs %in% unique(unique_stk_upc1[,upc])]
    }
    
    cat(length(upcs_in_module1))
    
    upcs_in_module2 <- NULL
    try(upcs_in_module2 <- upcs[upcs %in% unique(stacked2[,upc])])
    if (is.null(upcs_in_module2)) {
      num_stacked2 <- stacked2[, .N]
      split_size2  <- 10
      group_size2  <- num_stacked2 %/% 10
      
      unique_stk_upc2 <- data.table()
      for (split in 1:10) {
        if (split!=10) {
          sub_upc <- unique(stacked2[(1 + (split - 1) * group_size2):(split * group_size2), 
                                     .(upc)])
        } else {
          sub_upc <- unique(stacked2[(1 + (split - 1) * group_size2):num_stacked2, 
                                     .(upc)])
        }
        unique_stk_upc2 <- rbind(unique_stk_upc2, sub_upc)
      }
      stopifnot(class(unique_stk_upc2$upc) == "integer64")
      
      upcs_in_module2 = upcs[upcs %in% unique(unique_stk_upc2[,upc])]
    }
    
    cat(length(upcs_in_module2))
    
    upcs_in_module <- unique(c(upcs_in_module1, upcs_in_module2))
    
    if (length(upcs_in_module) > 0){ #There are Module that don't have any top UPCs
      
      for (j in 1:length(upcs_in_module)){
        upc_curr <- upcs_in_module[j]
        print(sprintf("Saving UPC %s in Module %s", upc_curr, Module))
        
        movement1 <- stacked1[upc == upc_curr]
        movement1[, week_end := as.Date(as.character(week_end), format='%Y%m%d')]
        
        movement2 <- stacked2[upc == upc_curr]
        movement2[, week_end := as.Date(as.character(week_end), format='%Y%m%d')]
        
        movement <- rbind(movement1, movement2)
        rm(movement1, movement2)
        
        save(list = "movement", file = sprintf("%s/Modules/%s/%s.RData", 
                                               output_directory, Module, upc_curr))
        rm(movement)
      }
    }
    rm(stacked1, stacked2)
    
    cat("Module",Module,"Done! \n")
  }
  
  

  return(NULL)

}

# Now go parallel

Nslots <- as.numeric(Sys.getenv("SLURM_CPUS_ON_NODE"))
cat(as.character(Nslots), "\n")

cl <- makeCluster(Nslots, outfile = "")

registerDoParallel(cl)

clusterExport(cl=cl,
              varlist = ls(.GlobalEnv))

foreach(i = 1:ncases,
        .packages = c("data.table",
                      "bit64",
                      "lubridate")) %dopar% {
                        build_movement(Module = unique_mod[i])
                      }

stopCluster(cl)



print(warnings())

print("Done!")
