
# storemeta.R
# -----------------------------------------------------------------------------
#
# An R script that will build store / year level meta data using the
# movement files in the RData format. So need to have run the R script
# that produces the RData Formatted Movement files (movement.R)
# BEFORE running this one.

library(data.table)
library(bit64)
library(doParallel)

output_directory = "/grpshares/hitsch_shapiro_ads/data/RMS" # Change appropriately

# Define a function that will build the store meta data

build_store_meta <- function(module_dir) {
  
  stacked  <- data.table()
  
  moduleno <- gsub("(.*)/Modules/(.*)",
                   "\\2", module_dir)
  
  cat("Processing Module", moduleno, "\n")
  
  file_list <- list.files(module_dir, full.names = TRUE)
  
  for (fn in file_list) {
    upcno <- gsub(paste0("(.*)/Modules/",
                         moduleno,
                         "/(.*).RData"),
                  "\\2", fn)
    
    # cat("Processing Module", moduleno, "upc", upcno, "\n")
    
    load(fn)
    
    temp    <- movement[, .(revenue = sum(units*price/prmult),
                            first_date = min(week_end),
                            last_date = max(week_end)), 
                        by = .(store_code_uc, year(week_end))]
    
    stacked <- rbind(stacked, temp)
    
    stacked <- stacked[,.(revenue = sum(revenue, na.rm=TRUE),
                          first_date = min(first_date, na.rm=TRUE),
                          last_date = max(last_date, na.rm=TRUE)), 
                       by = .(store_code_uc, year)]
    
    rm(temp, movement)
  }
  
  cat("Module", moduleno, "Done! \n")
  return(stacked)
}

# Go Parallel

Nslots <- as.numeric(Sys.getenv("SLURM_CPUS_ON_NODE"))
cat(as.character(Nslots), "\n")

cl <- makeCluster(Nslots, outfile = "")

registerDoParallel(cl)

clusterExport(cl=cl,
              varlist = ls(.GlobalEnv))

list_modules <- list.dirs(paste0(output_directory, "/Modules"),
                          full.names = TRUE, recursive = FALSE)
ncases <- length(list_modules)

RMS_store <- foreach(i = 1:ncases,
                     .packages = c("data.table",
                                   "bit64",
                                   "lubridate"),
                     .combine  = rbind) %dopar% {
  build_store_meta(list_modules[i])
                     }

stopCluster(cl)

RMS_store = RMS_store[,.(revenue    = sum(revenue, na.rm=TRUE),
                         first_date = min(first_date, na.rm=TRUE),
                         last_date  = max(last_date, na.rm=TRUE)), 
                      by = .(store_code_uc, year)]
RMS_store[, N_weeks := as.numeric(last_date - first_date)/7 + 1]

setkey(RMS_store, store_code_uc, year)

save(RMS_store,
     file= paste0(output_directory, "/Meta-Data/store_meta.RData"))

print(warnings())
