rm(list=ls())
library(data.table)
library(bit64)
library(Rcpp)
library(RcppArmadillo)
setNumericRounding(0)

Rcpp::sourceCpp(paste0(dir, "cpp-Source/Price-Algorithm.cpp"))

price.imputation <- function(DT){
	movement <- copy(DT)
	movement[, `:=`(price=price/prmult, week_end=as.numeric(week_end))]
	setkey(movement, upc, store_code_uc, week_end)
	# movement <- unique(movement)

	# Remove conflicting observations
	movement[, temp:=(1:.N), by=.(upc, store_code_uc, week_end)]
	movement <- movement[temp==1]
	movement[, temp:=NULL]

	# Filter for process
	filter_table <- movement[, .(span=(max(week_end)-min(week_end))/7+1, obs=.N), by=.(upc, store_code_uc)]
	filter_table[, percent_covered:=obs/span]
	filter_table[, processed:=(percent_covered>=0.2)]

	setkey(filter_table, upc, store_code_uc)
	setkey(movement, upc, store_code_uc)

	movement <- filter_table[,.(upc, store_code_uc, processed)][movement]

	not_processed <- movement[processed==FALSE]
	movement <- movement[processed==TRUE]

	# Create move_lookup_table for the CPP algorithm
	move_lookup_table <- movement[, .(index_begin=.I[1], index_end=.I[.N]), by=.(upc, store_code_uc)]

	# Processing
	# -------------------------------------------------- #
	# Step 1: Calculating Base Price
	# -------------------------------------------------- #
	structures <- movement[, .(upc, store_code_uc, week_end)]
	base_prices <- as.data.table(BasePriceDetectionHybridAlgorithm(movement, move_lookup_table))
	base_prices <- cbind(structures, base_prices)
	# -------------------------------------------------- #
	# Step 2: Imputing Base Price
	# -------------------------------------------------- #	
	# 1. Get full week structures for each upc-version-store combination
	# (Equivalent to "ExpandMissingWeeks" in C++)
	tryCatch(
		{
			structures[, `:=`(min_date=min(week_end), max_date=max(week_end)), by=.(upc, store_code_uc)]
			structures[, week_end:=NULL]
			structures <- unique(structures)
			structures <- structures[, .(week_end=seq(min_date, max_date, 7)), by=.(upc, store_code_uc)]
		},
		error = function(cond){
			print(paste0("ERROR: ", cond))
			next
		},
		warning = function(cond){
			print(paste0("WARNING: ", cond))
			next
		},
		finally = {

		}
	)
	# 2. Rolling join and adjust base prices
	# Split base_prices
	base_prices[base_price==0, base_price:=NA]
	base_prices_non_NA = base_prices[!is.na(base_price)]
	base_prices_NA = base_prices[is.na(base_price)]
	# Set Keys
	setkey(base_prices_NA, upc, store_code_uc, week_end)
	setkey(base_prices_non_NA, upc, store_code_uc, week_end)
	setkey(structures, upc, store_code_uc, week_end)
	# Rolling Join
	base_prices = base_prices_non_NA[structures, roll=TRUE]
	# Adjust wrongly imputed base prices
	base_prices_NA[, `:=`(identifier=1, base_price=NULL)]
	base_prices = base_prices_NA[base_prices]
	base_prices[identifier==1, base_price:=NA]
	base_prices[, identifier:=NULL]
	rm(structures, base_prices_NA, base_prices_non_NA, move_lookup_table)
	# -------------------------------------------------- #
	# Step 3: Imputing Prices and Cleaning
	# -------------------------------------------------- #
	setkey(movement, upc, store_code_uc, week_end)
	move = movement[base_prices]
	move[, imputed_price:=price]
	move[is.na(imputed_price), imputed_price:=base_price] # Impute prices for zero-sale weeks
	move[is.na(units), units:=0]
	move[, `:=`(processed=NULL, price=NULL)]
	move[, `:=`(week_end=as.Date(week_end, origin="1970-01-01"))]
	setnames(move, "prmult", "prmult_no_use")
	# -------------------------------------------------- #
	# Step 4: Adding not_processed to the file
	# -------------------------------------------------- #
	not_processed[, `:=`(processed=NULL, base_price=NA, imputed_price=price)]
	not_processed[, `:=`(week_end=as.Date(week_end, origin="1970-01-01"), price=NULL)]
	setnames(not_processed, "prmult", "prmult_no_use")

	move[, processed:=TRUE]
	not_processed[, processed:=FALSE]

	setcolorder(move, c("upc", "upc_ver_uc", "store_code_uc",
    		"week_end", "base_price", "imputed_price", "units", "feature", "display", 
    		"prmult_no_use", "processed"))

	setcolorder(not_processed, c("upc", "upc_ver_uc", "store_code_uc",
    		"week_end", "base_price", "imputed_price", "units", "feature", "display", 
    		"prmult_no_use", "processed"))

	res <- rbindlist(list(move, not_processed))
	return(res)
}






