// Process-Price-Promotion-Data/Scripts/cpp-Source/Config.h
// Defines the class "Product" for the base price detection algorithm

// Author: Günter J. Hitsch, 2014
// Last Edited: Jihong Song, Sept. 2016

#ifndef PRODUCT_H
#define PRODUCT_H


#include <iostream>
#include <vector>
#include <cmath>
#include <numeric>

#include "Config.h"
//#include "Config.h"

using namespace std;


class Product {
public:
   std::vector<double> ps, bps,bps_Snapshot;
   int id, T, T0, T1;
   
   Product(std::vector<double>);
   
   void BasePriceDetectionAlgorithm();
   void BasePriceDetectionHybridAlgorithm();

   /// Standalone processing steps for bps:
   /// Made public for debugging
   
   void SetBasePriceToModeOfTopPrices();
   void FindLonerBasePrices();
   void ContinueBasePricesWithinTolerance();
   void FindValleys();
   void FindValleysWithBasePrice();
   
   void FindAdjustmentPeriods();
   void FindIsolatedBasePriceClusters();
//   void Telescope();
   void Telescope(bool);
   bool isProm(int, int);
   void FindMistakenBasePrices();
   void FillGapsIfContainsPromotedPrices();
   
   int FindPromotedAndNotPromotedPrices(); // initBpsBySpells();
   bool FindMistakenNotPromotedPrices(); // findPromNotProm();
   
   /// Private utilized methods
   std::vector<double> modes(std::vector<double> &);
   double max_mode(std::vector<double> &);
   double min_mode(std::vector<double> &);
   std::vector<double> rolling_modes(int, int);
   // std::vector<pcut> get_n_pcuts(int, bool);
   std::vector<double> get_n_prom(int, bool);
   double median(std::vector<double>);
   //vector< vector<double> > HybridDebugVerbose();

private:

};

// -----------------------------------------------
// Constructors
// -----------------------------------------------
Product::Product(vector<double> x){
   ps = x;
   T = (int) ps.size();
   bps.resize(T);
   fill(bps.begin(),bps.end(),0);
   T0 = mode_T_before+1;
   T1 = T-mode_T_after;
}

// -----------------------------------------------
// Private methods
// -----------------------------------------------
double Product::median (vector<double> values) {
   vector<double> x = values;
   double res;
   if (x.size() > 0){
      sort(x.begin(),x.end());
      if (x.size()%2 ==1){
         res=x[x.size()/2];
      } else {
         res=(x[x.size()/2]+x[x.size()/2-1])/2;
      }
      
   } else {
      // Error information
      res=-999999;
   }
   return res;
}


vector<double> Product::modes(vector<double> &x){
   int c = 1, cM = 0;
   vector<double> z;
   z.push_back(x.back());
   for (int i = (int) x.size()-2; i >= 0; i--) {
      if (x[i] == x[i+1]) {
         c++;
      } else if (c > cM) {
         z.push_back(x[i+1]);
         cM = c;
         c = 1;
      } else {
         c = 1;
      }
   }
   
   if (c > cM) z.push_back(x[0]);
   return z;
}


double Product::max_mode(vector<double> &x){
   int c = 1, cM = 0;
   double z = x.back();
   for (int i = (int) x.size()-2; i >= 0; i--) {
      if (x[i] == x[i+1]) {
         c++;
      } else if (c > cM) {
         z = x[i+1];
         cM = c;
         c = 1;
      } else {
         c = 1;
      }
   }
   
   if (c > cM) z = x[0];
   return z;
}

double Product::min_mode(vector<double> &x){
   int c=1, cM = 1;
   double z = x.back();
   for (int i = (int) x.size()-2; i >= 0; i--) {
      if (x[i] == x[i+1]) {
         c++;
      } else if (c >= cM) {
         z = x[i+1];
         cM = c;
         c = 1;
      } else {
         c = 1;
      }
   }
   
   if (c >= cM) z = x[0];
   
   return z;
}

// Returns rolling max modes of ps above the cutoff
vector<double> Product::rolling_modes(int winmode_T_before, int winmode_T_after) {
   int l = winmode_T_before;
   int u = winmode_T_after;
   
   int t0 = max(1 - l, 1);
   int t1 = min(T - u, T);
   int cutoff = round(mode_percentile_cutoff*(u-l));
   vector<double> res(T);

   // TODO Costly procedures
   for (int i = t0 - 1; i <= t1-1; i++) {
      vector<double> s(ps.begin()+i+l,ps.begin()+i+u+1);
      sort(s.begin(),s.end());
      double ref = s[cutoff];

      vector<double> z;
      for (int i2 = 0; i2 < s.size(); i2++) {
         if (s[i2] >= ref) {
            z.push_back(s[i2]);
         }
      }
      res[i] = max_mode(z);
   }

   return(res);
}

// -----------------------------------------------
// Private standalone processing steps
// -----------------------------------------------

// -----------------------------------------------
// SetBasePriceToModeOfTopPrices
// Initiate baseprices by max rolling modes
//
// First the max modes of the rolling windows before
// and after.  If the current price euqals either mode,
// let it be the baseprice.
//
void Product::SetBasePriceToModeOfTopPrices() {
   vector<double> modes_b(T),modes_a(T);
   modes_b= rolling_modes(-mode_T_before, -1);
   modes_a= rolling_modes(1, mode_T_after);
   
   for (int i = T0-1; i < T1; i++) {
      if (ps[i]==modes_a[i] || ps[i]==modes_b[i]) bps[i]=ps[i];
   }   
}

// -----------------------------------------------
// FindLonerBasePrices
//
// Remove bps occur at most loner_N times within a
// rolling window
//
void Product::FindLonerBasePrices(){
   int l, u, c;
   for (int t = T0-1; t < T1; t++)
   {
      if (bps[t] > 0)
      {
         l = max(t-loner_T_before, T0-1);
         u = min(t+loner_T_after, T1-1);
         c = (int) count(ps.begin()+l, ps.begin()+u+1, bps[t]);
         
         if (c <= loner_N)
            bps[t] = 0;
      }
   }
}

// -----------------------------------------------
// ContinueBasePricesWithinTolerance
// Fill baseprices by continuation
//
// Step 1: Move forward in time
// If a price is not a base price and at most tol_dollars_lower cheaper or
// tol_dollars_higher more expensive than the previous base price then define
// the current base price = current price
//
// Step 2: Move backwards in time
// If a price is not a base price and at most tol_dollars_lower cheaper or
// tol_dollars_higher more expensive than the next base price then define
// the current base price = current price.
//
void Product::ContinueBasePricesWithinTolerance(){
   for (int t = T0; t < T1; t++) {
      if (bps[t]==0) {
         if (ps[t] >= bps[t-1] - tol_dollars_lower &&
             ps[t] <= bps[t-1] + tol_dollars_higher) {
            bps[t] = ps[t];
         }
      }
   }
   for (int t=T1-2; t >= T0-1; t--) {
      if (bps[t]==0) {
         if (ps[t] >= bps[t+1] - tol_dollars_lower &&
             ps[t] <= bps[t+1] + tol_dollars_higher){
            bps[t] = ps[t];
         }
      }
   }
}

// -----------------------------------------------
// FindValleys
// Fill baseprice valleys
//
// Define valleys to be baseprices starting with a base price and ending
// with a base price at least as large as next_base_price_percentage_factor of the current base price
//
// If there are (i)   no base prices in this valley, and
//              (ii)  if all prices in this valley are smaller or equal to the
//                    max of the base prices at the ends, and
//              (iii) if at least one price in the valley is strictly smaller
//                    than min_base_price_percentage_factor of the min of the base prices at the ends
// percentile_cutoffen set the base prices in the valley to the current base price.
//
// Finally, find "adjustment periods" of prices rising to the new higher base
// price at the end of the promotion period, and set the base prices in this
// period to the coresponding price
//
void Product::FindValleys(){
   for (int t = T0; t < T1-1; t++){
      double ref = bps[t-1];
      if (bps[t] == 0 && ref !=0) {
         int c = 0;
         while (bps[t+c+1] == 0 && c<promotion_window && t+c+1 < T1-1) {c++;}
         
//         cout << t << " "<<c <<"\n";
         
         if (bps[t+c+1] > 0 && c < promotion_window){
//            cout << t << " "<<c <<"\n";
            
//            typedef vector<double>::iterator TI;
//            pair<TI, TI> minmaxP = minmax_element(ps.begin()+t, ps.begin()+t+c+1);
//            double minP = *minmaxP.first;
//            double maxP = *minmaxP.second;
            
            double minP = *min_element(ps.begin()+t, ps.begin()+t+c+1);
            double maxP = *max_element(ps.begin()+t, ps.begin()+t+c+1);
            
            
            double LgBp = max(ref,bps[t+c+1]);
            double SmBp = min(ref,bps[t+c+1]);
            double benBp = SmBp * min_base_price_percentage_factor;
                        
            if (bps[t+c+1] >= next_base_price_percentage_factor*ref
                && maxP <= LgBp && minP <= benBp
                ) {
               fill(bps.begin()+t, bps.begin()+t+c+1, ref);
               
               if (maxP > ref) {
                  int k = 0;
                  while(ps[t+c-k]>ref){bps[t+c-k] = ps[t+c-k]; k++;}
               }
               
            }
         }
         
         t=t+c;
      }
   }
}

// TODO Expensive, need rework
// -----------------------------------------------
// FindValleysWithBasePrice
// Fill baseprice valleys with baseprices
//
// If (i)  all prices in a valley are smaller or equal to the max of the prices
//         at the ends, and
//    (ii) if at least one price in the valley is strictly smaller than min_price_percentage_factor
//         of the min of the prices at the ends
// percentile_cutoffen set the base prices in the valley to the current base price
//
//
// Finally, find "adjustment periods" of prices rising to the new higher base
// price at the end of the promotion period, and set the base prices in this
// period to the coresponding price
//
void Product::FindValleysWithBasePrice(){
   for (int t = T0-1; t < T1-2; t++){
      if (bps[t] > 0) {
         double ref = bps[t];
         
         int c = 0;
         for (int i = 1; i <= promotion_window+1; i++) {
            if (ps[t+i] >= next_price_percentage_factor*ref && t+i<T1) { c=i; break;}
         }
         
         if (c > 1){
            
               // As long as no base price is hit, as long as prices are <= than
               // the initial base price, and as long as the prices are increasing,
               // "widen" the promotion valley
            
            int k = 0;
            while(bps[t+c+k]==0 && ps[t+c+k]<=ref && ps[t+c+k]<=ps[t+c+k+1] && t+c+k < T1-1 && c+k<=promotion_window ) {
               k++;
            }
            
            c = c+k;
            
            double minP = *min_element(ps.begin()+t+1, ps.begin()+t+c);
            double maxP = *max_element(ps.begin()+t+1, ps.begin()+t+c);
            
            
            double LgP = max(ref,ps[t+c]);
            double SmP = min(ref,ps[t+c]);
            double benP = SmP * min_price_percentage_factor;
            
            if (maxP <= LgP  && minP <= benP) {
               fill(bps.begin()+t+1, bps.begin()+t+c, ref);
               
                  // Adjust Bps to meet Ps at end of promotion
               if (maxP > ref) {
                  int k = 0;
                  while(ps[t+c-k]>ref){bps[t+c-k] = ps[t+c-k]; k++;}
               }
               
            }
         }
      }
   }
}

// -----------------------------------------------
// FindAdjustmentPeriods
// Fill baseprices based on adjustment periods
//
// Finds intervals between base prices where prices are either monotonically
// increasing or monontonically decreasing for at most N_adjustment_periods_up or N_adjustment_periods_down periods
//  and records these periods as base prices
//
void Product::FindAdjustmentPeriods(){
   for (int t=T0; t < T1-1; t++){
      int c = 0;
      if (bps[t]==0 && bps[t-1]!=0) {
         while (bps[t+c+1] == 0 && t+c < T1) {c++;}
         
         vector<double> p(ps.begin()+t, ps.begin()+t+c+1);
         p.insert(p.begin(),bps[t-1]);
         p.push_back(bps[t+c+1]);
         int l = (int)p.size();
         vector<double> d(l);
         adjacent_difference(p.begin(), p.end(), d.begin());
         
         if ((*min_element(d.begin()+1, d.end()) >= 0 && l-2 <= N_adjustment_periods_up) ||
             (*max_element(d.begin()+1, d.end()) <= 0 && l-2 <= N_adjustment_periods_down)){
            for (int t2=t; t2 <= t+c; t2++){ bps[t2] = ps[t2]; }
         }
      }
      t=t+c;
   }
}

// -----------------------------------------------
// FindIsolatedBasePriceClusters
// Remove baseprices that are isolated clusters
//
// Detect "isolated" clusters of base prices of at most length cluster_T_max
// If there are no base prices before the cluster within a window of at
// least length min_T_isolated_before and and if there are no base prices after the cluster
// within a window of at least length min_T_isolated_after, then infer that the cluster
// does not represent actual base prices
//
void Product::FindIsolatedBasePriceClusters() {
   vector<bool> isChanged(T);
   
   for (int t = T0-1; t < T1-2; t++){
      // Scan for zeros before a cluster
      int zeros_before = 0;
      if (bps[t]==0 && (bps[t-1] > 0 || isChanged[t-1]==true || t == T0-1) ) {
         
         while (bps[t+zeros_before+1] == 0 && t+zeros_before+1 < T1) {zeros_before++;}
         
//         cout<<t<<" "<<zeros_before<<" "<<T1-1<<"\n";
         
         int size_cluster = 0;
         while (bps[t+zeros_before+size_cluster+1] != 0 && t+zeros_before+size_cluster+1 <= T1-1) {
            size_cluster++;
         }

//         cout<<t<<" "<<size_cluster<<" "<<T1-1<<"\n";

         // Scan for a small cluster
         if (zeros_before + 1 >= min_T_isolated_before
             && size_cluster <= cluster_T_max
             && t+1+zeros_before+size_cluster <=T1-1){
            // Scan for zeros after a cluster
            int zeros_after = 1;
            while (bps[t+zeros_before+size_cluster+zeros_after+1] == 0
                   && t+zeros_before+size_cluster+zeros_after+1 <= T1-1) {
               zeros_after++;
            }
            
            if (bps[t+zeros_before+size_cluster+2] > 0
                || t+zeros_before+size_cluster+zeros_after+1 > T1-1) {
               zeros_after = 0;
            }
            
//            cout<<t<<" "<<zeros_after<<" "<<T1-1<<"\n";

            // A cluster found
            if (1){
               fill(bps.begin()+t+zeros_before+1, bps.begin()+t+zeros_before+size_cluster+zeros_after+1, 0);
               fill(isChanged.begin()+t+zeros_before+1,isChanged.begin()+t+zeros_before+size_cluster+zeros_after+1,true);
            }
         }
      }
   }
}

// -----------------------------------------------
// Telescope
// Fill baseprices based on telescopes
//
// Finds gaps in the base price series ending with the same price as the
// current base price.  Gaps cannot exceed telescope_window periods.
// Set the base price in the gap interval to the current base price
//
//
//void Product::Telescope(){
//   for (int t = T0-1; t < T1-1; t++){
//      if (bps[t]==0 && bps[t-1]!=0) {
//         double ref = bps[t-1];
//
//         int c = 0;
//         while (bps[t+c] == 0 && t+c < T1-1 && c <= telescope_window) {
//            if (ps[t+c+1]==ref) {
//              fill(bps.begin()+t, bps.begin()+t+c+1, ref);
//            }
//            c++;
//         }
//      }
//   }
//}

void Product::Telescope(bool is_direction_forward){
   if (is_direction_forward){
   for (int t = T0; t < T1-1; t++){
      if (bps[t]==0 && bps[t-1]!=0) {
         double ref = bps[t-1];
         
         int c = 0;
         while (bps[t+c] == 0 && t+c < T1-1 && c < telescope_window_forward) {
            if (ps[t+c+1]==ref) {
               fill(bps.begin()+t, bps.begin()+t+c+1, ref);
            }
            c++;
         }
      }
   }
   } else {
      for (int t = T1-2; t >= T0; t--){
         if (bps[t]==0 && bps[t+1]!=0) {
            double ref = bps[t+1];
            
            int c = 0;
            while (bps[t-c] == 0 && t-c > T0-1 && c < telescope_window_backward) {
               if (ps[t-c-1]==ref) {
                  fill(bps.begin()+t-c, bps.begin()+t+1, ref);
               }
               c++;
            }
         }
      }
   }
   
}

//Lagacy code:  use get_n_prom to get promoted price levels and discounts
// ----------------------------------------------------------------------------
// get_n_pcuts
// Get N price cuts
//
// Find all price cuts in the window of width comparison_promotion_window ether forward or backward.
// Among these, choose the N price cuts that are closest to t (all price cuts
// if there are less than N such cuts).
//
// Both price levels and cuts are returned.
// //
// vector<pcut> Product::get_n_pcuts(int timeAt, bool forward=true){
//    bool f = forward;
//    int t = timeAt;
//    vector<pcut> res;
   
//    int ct = 0; int cn=0;
//    if (f) {
//       while(ct<=comparison_promotion_window && cn<= N_most_recent && t+ct <= T1-1){
//          double p = ps[t+ct];
//          double bp= bps[t+ct];
//          double d = bp-p;
//          if (d>0) {
//             pcut pc; pc.p = p; pc.disc=d;
//             res.push_back(pc);
//             cn++;
//          }
//          ct++;
//       }
//    } else {
//       while(ct<=comparison_promotion_window && cn<= N_most_recent && t-ct >= T0-1){
//          double p = ps[t-ct];
//          double bp= bps[t-ct];
//          double d = bp-p;
//          if (d>0) {
//             pcut pc; pc.p = p; pc.disc=d;
//             res.insert(res.begin(), pc);
//             cn++;
//          }
//          ct++;
//       }
//    }
//    return res;
// }

// ----------------------------------------------------------------------------
// get_n_prom
// Get N promoted prices
//
// Find all price cuts in the window of width comparison_promotion_window ether forward or backward.
// Among these, choose the N price cuts that are closest to t (all price cuts
// if there are less than N such cuts).
//
// Only price levels are returned.
//
vector<double> Product::get_n_prom(int timeAt, bool forward=true){
   bool f = forward;
   int t = timeAt;
   vector<double> res;
   
   int ct = 0; int cn=0;
   if (f) {
      while(ct< comparison_promotion_window && cn<  N_most_recent && t+ct <= T1-1){
         double p = ps[t+ct];
         double bp= bps[t+ct];
         if (bp>0 && bp>p) {
            res.push_back(p);
            cn++;
         }
         ct++;
      }
   } else {
      while(ct< comparison_promotion_window && cn<  N_most_recent && t-ct >= T0-1){
         double p = ps[t-ct];
         double bp= bps[t-ct];
         if (bp>0 && bp>p) {
            res.insert(res.begin(), p);
            cn++;
         }
         ct++;
      }
   }
   return res;
}

// ----------------------------------------------------------------------------
// isProm
// is price promoted?
//
// Tests if the prices during t0 and t1 could be promoted prices.  Identify the
// N_most_recent promotions in the comparison window before t0 and after t1.
// percentile_cutoffen compare prices in [t0, t1] to these promotions
//

bool Product::isProm(int timeStart, int timeEnd) {
   int t0 = timeStart; int t1 = timeEnd;

   vector<double> ps0 (ps.begin()+t0, ps.begin()+t1+1);
   sort(ps0.begin(), ps0.end());
   double minP = ps0[0];
   double minM = min_mode(ps0);

   vector<double> pca = get_n_prom(t1+1);
   vector<double> pcb = get_n_prom(t0-1, false);
   
   bool testa = false, testb = false;
   if (pca.size()> 0){
      double minPca = *min_element(pca.begin(),pca.end());
      double diffMa = abs(pca.front()-minM)/minM;
      double diffPa = abs(pca.front()-ps[t1])/ps[t1];
      
      testa = diffMa <= promotion_percent_similarity_closest ||  diffPa <=promotion_percent_similarity_closest ||
         minP <= promotion_comparison_factor * minPca;
   }
   
   
   if (pcb.size()> 0){
      double minPcb = *min_element(pcb.begin(),pcb.end());
      double diffMb = abs(pcb.back()-minM)/minM;
      double diffPb = abs(pcb.back()-ps[t0])/ps[t0];
      
      testb = diffMb <= promotion_percent_similarity_closest ||  diffPb <=promotion_percent_similarity_closest ||
         minP <= promotion_comparison_factor * minPcb;
   }
   
   return (testa || testb);
}

// ----------------------------------------------------------------------------
// FindMistakenBasePrices
// Fix baseprices mistakenly to be not promoted
//
// Check if a base price at time t is below the factor mistaken_initial_threshold_factor * ref. If yes,
// check if within mistaken_window_for_initial_ps_drop+1 periods after the initial cut the base price
// returns to mistaken_end_threshold_factor of the initial price drop.  Skip this criterion if
// t+ mistaken_window_for_initial_ps_drop+1 > T1
//
// Check if the resulting "valley" of prices qualify as promotions.
// If so, conclude that the prices in the "valley" are not base prices, and
// set the corresponding base prices in this period to the base price at t.
//

void Product::FindMistakenBasePrices() {
   for (int t = T0-1; t < T1-1; t++){
      if (bps[t] > 0 ) {
         double ref = bps[t];
         
            // Search for initial_drop_base_price_index
         int c0 = 0;
         for (int i = 1; i <= mistaken_window_for_initial_ps_drop; i++){
            if (bps[t+i] > 0 & t+i < T1) {
               c0 = i; break;
            }
         }
         
         if (c0 > 0) {
            int t0 = t+c0;
            int t1 = t0;
            
            double ref2 = bps[t0];
            
            double threshold_ini = mistaken_initial_threshold_factor * ref;
            double threshold_end = mistaken_end_threshold_factor * ref + (1-mistaken_end_threshold_factor)*ref2;
            
//            cout<<t<<" "<<t0<<" "<<ref2<<" "<<threshold_ini<<" "<<ref<<" "<<threshold_end<<"\n";
            
            if (ref2 < threshold_ini){
               
               int c1= 0;
               for (int j = 2; j <= mistaken_window; j++) {
                  if (bps[t0+j]> 0 && bps[t0+j] >= threshold_end && t0+j <= T1-1) {
                     c1 = j; break;
                  }
               }
               
               if (c1 > 0 || t+mistaken_window > T1-1){
                  if (c1 > 0) {
                     t1 = t0+c1-1;
                  } else {
                     t1 = T1-1;
                  }
               
//                  cout<<t<<" "<<t0<<" "<<t1<<" "<<isProm(t0, t1)<<"\n";
                  
                  if (t1-t > 0 && isProm(t0, t1)) {
                     fill(bps.begin()+t,bps.begin()+t1+1, ref);
                     t = t1;
                  }
               }
            }

         } else {
            t = t+mistaken_window_for_initial_ps_drop;
         }
      }
   }
}

// ----------------------------------------------------------------------------
// FillGapsIfContainsPromotedPrices
// Fill baseprices containing promotion gaps
//
// Check for gaps in the base price series.  Check if there are promotions
// within this gap, based on a comparision with promoted price levels before or
// after the gap.
//
// If we conclude the gap contains promoted prices, set the corresponding
// base prices in this period to the base price at t

void Product::FillGapsIfContainsPromotedPrices(){
   for (int t = T0; t < T1; t++){
      int c = 0, t1 = 0;
      if (bps[t]==0 && bps[t-1] > 0) {
         double ref = bps[t-1];
         for (int i=0; i< gap_window && t+i< T1-1; i++){
            if (bps[t+i+1]>0) {
               c = i;
               break;
            }
         }
         
//         cout << t<<" "<<c<<"\n";
         
         if (c == 0 && bps[t+c+1] == 0 && t+gap_window >= T1){
            t1 = T1-1;
         } else {
            t1 = t+c;
         }
         
//         cout << t<<" "<<c<<" "<<t1<<" "<<isProm(t, t1)<<"\n";
         
         if ((bps[t+c+1] > 0 || t1 == T1-1) && isProm(t, t1)) {
            fill(bps.begin()+t,bps.begin()+t1+1, ref);
         }
      }
   }
}


// TODO percentile_cutoffis is a blunt translation of R code: rewrite for speed
// ----------------------------------------------------------------------------
// FindPromotedAndNotPromotedPrices
// Initiate baseprices by spells of prices
//
// Find spells of prices, defined as sequences within threshold_factor_reference_decrease and threshold_factor_reference_increase
// of the first price (the "reference price") in this spell.  Find local maxima
// of these spells:  percentile_cutoffese are defined as not promoted prices.  Between local
// maxima, prices that are below threshold_factor_promotion of the median of the prices in the
// inital "local maximum spell" are defined to be promoted prices.

int Product::FindPromotedAndNotPromotedPrices(){
   vector<double> psProm(T), psNotProm(T);
   vector<int> timeSpells;  // time of spell
   vector<dir> dirSpells;  // direction of spell

   // Records spell points
   double ref = ps[0];
   
   for (int t = 0; t <= T-1; t++){
      if (ps[t+1] < threshold_factor_reference_decrease * ref) {
         timeSpells.push_back(t+1);
         dirSpells.push_back(dnDir);
         ref = ps[t+1];
      } else if (ps[t+1] > threshold_factor_reference_increase * ref) {
         timeSpells.push_back(t+1);
         dirSpells.push_back(upDir);
         ref = ps[t+1];
      }
   }

   vector<int> TMaxBegins, TMaxEnds;
   
   if (timeSpells.size() >= 3) {
      for (int n = 0; n < timeSpells.size()-2; n++){
         if (dirSpells[n]==upDir && dirSpells[n+1]==dnDir){
            TMaxBegins.push_back(timeSpells[n]);
            TMaxEnds.push_back(timeSpells[n+1]-1);            
            for (int t = TMaxBegins.back(); t <= TMaxEnds.back(); t++){
               psNotProm[t] = ps[t];
            }
         }
      }
   }

   
   if (TMaxBegins.size() > 0) {
      for (int m = 0; m < TMaxBegins.size()-1; m++) {
         int l = TMaxEnds[m] + 1;
         int u = TMaxBegins[m+1]-1;
         
         vector<double> subPs(ps.begin()+TMaxBegins[m], ps.begin()+TMaxEnds[m]+1);
         double pMed=median(subPs);

         if( u >=l && *min_element(ps.begin()+l, ps.begin()+u+1) < threshold_factor_promotion * pMed){
            for (int t = l; t <= u; t++) {
               psProm[t] = ps[t];
            }
         }
      }
      
      // Mark promotion by max
      // BEGIN Legacy FindMistakenNotPromotedPrices
      //
      // For any spell of price maxima, compare to the N (before/after) closest
      // promoted and non-promoted prices in the windows before and after.
      // Calculate the difference of the medians of these promoted and non-
      // promoted prices, called distance. If (pMed-pPMed)/dis <= mistaken_not_promoted_ratio_threshold_promotion,
      // then declare the prices in the current maximum spell to be promoted
      
      int NFixes = 1;
      
      while(NFixes > 0){

         NFixes = 0;
         for (int m = 0; m < TMaxBegins.size(); m++) {
            
            vector<double> psCompP, psCompNP;
            
            int Tb = TMaxBegins[m]; int Te = TMaxEnds[m];
            int t0 = max(0, Tb-mistaken_not_promoted_T_window_before);
            int t1 = min(T-1, Te+mistaken_not_promoted_T_window_after);
            
//            cout<<"["<<m<<"] "<<Tb<<" "<<Te+1<<"\n";

            double pP;
            int c = 0;

            // Sampling before
            while (t0+c < Tb) {
               pP = psProm[t0+c];
               if (pP>0) psCompP.insert(psCompP.begin(), pP);
               c++;
            }
            
            int s0=min((int) psCompP.size(), mistaken_not_promoted_N_before+1);
            psCompP.resize(s0);

            c=1;
            while (Te+c <= t1) {
               pP = psProm[Te+c];
               if(pP>0) psCompP.push_back(pP);
               c++;
            }
            
            int s1=min((int) psCompP.size(), s0+mistaken_not_promoted_N_before+1);
            psCompP.resize(s1);
            
            // Sampling before
            c=0;
            while (t0+c < Tb) {
               pP = psNotProm[t0+c];
               if (pP>0) psCompNP.insert(psCompNP.begin(), pP);
               c++;
            }
            
            s0=min((int) psCompNP.size(), mistaken_not_promoted_N_before+1);
            psCompNP.resize(s0);
            
            c=1;
            while (Te+c <= t1) {
               pP = psNotProm[Te+c];
               if(pP>0) psCompNP.push_back(pP);
               c++;
            }

            s1=min((int) psCompNP.size(), mistaken_not_promoted_N_before+s0+1);
            psCompNP.resize(s1);
            
            if (psCompP.size()>0 && psCompNP.size()>0) {
            vector<double> subPs(ps.begin()+Tb, ps.begin()+Te+1);
               double pMed = median(subPs);
               
               double pPMed =median(psCompP);
               double pNPMed=median(psCompNP);
            
            double dis = pNPMed-pPMed;
            if (dis > 0 && (pMed-pPMed)/dis <= mistaken_not_promoted_ratio_threshold_promotion) {
               
               
//               cout<<Tb<<" "<<Te+1<<"\n";
               
               fill(psNotProm.begin()+Tb,psNotProm.begin()+Te+1,0);
               
               for (int t = Tb; t<=Te; t++) {
                  psProm[t] = ps[t];
               }
               
               NFixes ++;
               
               // Mark changed positions for deletion;
               TMaxBegins[m] = -99;
               TMaxEnds[m] = -99;
               
            }
            }
         }
         
         // Removed changd prices
         vector<int> NewTBegins;
         vector<int> NewTEnds;
         for (int r = 0; r<TMaxBegins.size();r++){
            if (TMaxBegins[r]!= -99) {
               NewTBegins.push_back(TMaxBegins[r]);
            }
         }
         
         for (int r = 0; r<TMaxEnds.size();r++){
            if (TMaxEnds[r]!= -99) {
               NewTEnds.push_back(TMaxEnds[r]);
            }
         }
         
         TMaxBegins=NewTBegins;
         TMaxEnds=NewTEnds;
         
      }
      // END Legacy FindMistakenNotPromotedPrices
      
         bps = psNotProm;
   }
   
   return (int) TMaxBegins.size();
}

// -----------------------------------------------
// Publc methods (Wrappers)
// -----------------------------------------------

void Product::BasePriceDetectionAlgorithm(){
   SetBasePriceToModeOfTopPrices();
   FindLonerBasePrices();
   ContinueBasePricesWithinTolerance();
   FindValleys();
   FindValleysWithBasePrice();
   ContinueBasePricesWithinTolerance();
   FindAdjustmentPeriods();
   FindIsolatedBasePriceClusters();
   Telescope(true);
   Telescope(false);
   ContinueBasePricesWithinTolerance();
   FindMistakenBasePrices();
   FillGapsIfContainsPromotedPrices();
   FindMistakenBasePrices();
}

void Product::BasePriceDetectionHybridAlgorithm(){
   BasePriceDetectionAlgorithm();
   bps_Snapshot = bps;
   
   if (FindPromotedAndNotPromotedPrices()>0) {
      ContinueBasePricesWithinTolerance();
      FindValleys();
      FindValleysWithBasePrice();
      ContinueBasePricesWithinTolerance();
      FindAdjustmentPeriods();
      Telescope(true);
      Telescope(false);
      ContinueBasePricesWithinTolerance();
      FindMistakenBasePrices();
      FillGapsIfContainsPromotedPrices();
      FindMistakenBasePrices();
   }
   
}

#endif


