#----------------------------------------
# Check False UPC Versioning Changes
#
#----------------------------------------

rm(list=ls())
library(data.table)
library(bit64)

# Get UPCs with version changes

dir = "~/RMS/" # Directory where results from 1 to 3 are stored
output_dir = paste0(dir, "Version-Correction/")

if(!dir.exists(output_dir)){
  dir.create(output_dir)
}

meta_data_path = paste0(dir, "Meta-Data.RData")
all_upc_path = paste0(dir, "Products.RData")
versions_path = paste0(dir, "Versions.RData")

load(meta_data_path)
load(all_upc_path)
load(versions_path)

# Get Multi Ver UPC: Versions
meta_upc = unique(meta_data[, .(upc, upc_ver_uc)])
multi_ver_upc_list = meta_upc[, .(count=uniqueN(upc_ver_uc)), by=upc][count>1, upc]
setkey(meta_upc, upc, upc_ver_uc)
multi_ver_upc = products[upc %in% multi_ver_upc_list]
setkey(multi_ver_upc, upc, upc_ver_uc)
multi_ver_upc = meta_upc[multi_ver_upc, nomatch=0L]

print(dim(multi_ver_upc))
save(multi_ver_upc, file=paste0(output_dir, "Multi-Ver-UPC.RData"))

# Loop through all observations; get old and new things

counter = 1
version_change_summary = data.table()
multi_ver_upc[, grp:=.GRP, by=upc]
for(product in 1:uniqueN(multi_ver_upc[,upc])){
  if(counter%%100==0){
    print(counter)
  }
  subset = multi_ver_upc[grp==product]
  versions = subset$upc_ver_uc
  relevant_meta = meta_data[upc==subset[1,upc]]
  for(i in 2:length(versions)){
    version_change = data.table(upc = subset[1, upc],
                                old_version = versions[i-1],
                                new_version = versions[i],
                                old_module = subset[i-1, product_module_code],
                                new_module = subset[i, product_module_code],
                                old_brand_descr = subset[i-1, brand_descr],
                                new_brand_descr = subset[i, brand_descr],
                                old_multi = subset[i-1, multi],
                                new_multi = subset[i, multi],
                                old_size1_code_uc = subset[i-1, size1_code_uc],
                                new_size1_code_uc = subset[i, size1_code_uc],
                                old_upc_descr = subset[i-1, upc_descr],
                                new_upc_descr = subset[i, upc_descr]
                                #new_upc_descr = subset[i, upc_descr],
                                #old_revenue_store = relevant_meta[upc_ver_uc==versions[i-1]][year==max(year)][, revenue_RMS/N_stores_RMS],
                                #new_revenue_store = relevant_meta[upc_ver_uc==versions[i]][year==min(year)][, revenue_RMS/N_stores_RMS]
                                )
    # Deal with the problem of versioning
    # In some cases, version number *decreases* as year increases
    year_old = relevant_meta[upc_ver_uc==versions[i-1], year][1]
    year_new = relevant_meta[upc_ver_uc==versions[i], year][1]
    if(year_old<year_new){
      version_change[, `:=`(old_revenue_store = relevant_meta[upc_ver_uc==versions[i-1]][year==max(year)][, revenue_RMS/N_stores_RMS],
                            new_revenue_store = relevant_meta[upc_ver_uc==versions[i]][year==min(year)][, revenue_RMS/N_stores_RMS])]
    } else{
      version_change[, `:=`(old_revenue_store = relevant_meta[upc_ver_uc==versions[i]][year==max(year)][, revenue_RMS/N_stores_RMS],
                            new_revenue_store = relevant_meta[upc_ver_uc==versions[i-1]][year==min(year)][, revenue_RMS/N_stores_RMS])]
    }

    version_change_summary = rbindlist(list(version_change, version_change_summary))
  }
  counter = counter + 1
}

version_change_summary[, `:=`(module_change=(old_module!=new_module),
                              brand_change=(old_brand_descr!=new_brand_descr),
                              multi_change = (old_multi!=new_multi),
                              size_change = (old_size1_code_uc!=new_size1_code_uc),
                              revenue_change_pct = 100*(new_revenue_store-old_revenue_store)/old_revenue_store)]


only_brand_change_cases = version_change_summary[brand_change==TRUE & module_change==FALSE & multi_change==FALSE
                                            & size_change==FALSE][,.(upc, old_version, new_version, old_brand_descr, new_brand_descr, 
                                              old_upc_descr, new_upc_descr, revenue_change_pct, old_revenue_store, new_revenue_store)]

# Ranks
setkey(meta_data, upc, upc_ver_uc)
setkey(products, upc, upc_ver_uc)
meta_data = products[meta_data][department_code!=9]

meta_data = meta_data[, .(revenue=sum(revenue_RMS)), by=.(upc, year)]
meta_data[, revenue_rank:=rank(-revenue, ties.method="min"), by=year]
meta_data = meta_data[upc %in% multi_ver_upc_list]
max_ranks = meta_data[, .(max_rank=min(revenue_rank)), by=upc]

setkey(version_change_summary, upc)
setkey(only_brand_change_cases, upc)
setkey(max_ranks, upc)

version_change_summary = max_ranks[version_change_summary, nomatch=0L]
only_brand_change_cases = max_ranks[only_brand_change_cases, nomatch=0L]

only_brand_change_cases[, false_positive:=""]
setcolorder(only_brand_change_cases, c('upc', 'max_rank' ,'old_version', 'new_version', 'old_brand_descr', 'new_brand_descr',
                                        'false_positive', 'revenue_change_pct', 'old_revenue_store', 'new_revenue_store', 'old_upc_descr', 'new_upc_descr'))                                  

potential_in_top_50000 = only_brand_change_cases[max_rank<=50000]

save(version_change_summary, only_brand_change_cases, potential_in_top_50000, file=paste0(output_dir, "Version-Change-Tables.RData"))
write.csv(potential_in_top_50000, file=paste0(output_dir, "Potential-FP-Top-50000.csv"))

# Manually fill false_positive to get Labeled-Potential-FP-Top-50000.csv
# Then save .csv file into Corrections.RData

