rm(list=ls())
library(data.table)
library(bit64)
library(Rcpp)
library(RcppArmadillo)
options(scipen=999)
setNumericRounding(0)

# Parallel computing implemented by skipping mechanisms in for loops (for now)
# This scripts processes products in All-Other.RData

run_on_linux = TRUE

data_dir = "/grpshares/hitsch_shapiro_ads/data/RMS/" # Directory where results from 1 to 3 are stored
product_path = paste0(data_dir, "Products.RData")
meta_data_path = paste0(data_dir, "Meta-Data.RData")
movement_path_start = paste0(data_dir, "Modules/")

dir = "~/Price-Promotion-Analysis/Process-New-Build/"
Rcpp::sourceCpp("../Base-Price-Imputation/Scripts/cpp-Source/Price-Algorithm.cpp")

correct_meta = TRUE

# load version correction reference data
labeled_file_path = "./Processed-Data/Corrections.RData"
load(labeled_file_path)
load(meta_data_path)
load(product_path)
corrections = corrections[false_positive==1, .(upc, old_version, new_version)]

# What we will do: Change new version back to old version

# Pre-Process: upc with multiple versions (count can only be 1 or 2)
corrections[, count:=.N, by=upc]
corrections[count>1, old_version:=min(old_version), by=upc]
corrections[, count:=NULL]
setnames(corrections, c("new_version", "old_version"), c("upc_ver_uc", "upc_ver_uc_corrected"))
setkey(corrections, upc, upc_ver_uc)

if(correct_meta){
	# Deal with meta-data
	setkey(meta_data, upc, upc_ver_uc)
	meta_data = corrections[meta_data]
	meta_data[is.na(upc_ver_uc_corrected), upc_ver_uc_corrected:=upc_ver_uc]
	save(meta_data, file=paste0(data_dir, "Meta-Data-Corrected.RData"))

	# Deal with product data
	setkey(products, upc, upc_ver_uc)
	products = corrections[products]
	#upc_module_list = unique(products[is.na(upc_ver_uc_corrected) & upc_ver_uc!=upc_ver_uc_corrected, .(upc, product_module_code)])
	products[is.na(upc_ver_uc_corrected), upc_ver_uc_corrected:=upc_ver_uc]
	save(products, file=paste0(data_dir, "Products-Corrected.RData"))
}


# Correct and Process
output_dir = "~/data/RMS-Build-2016/RMS-Processed/Modules/"
if(!dir.exists(output_dir)){
	dir.create(output_dir)
}

module_list = as.numeric(list.files(movement_path_start))

Sys.sleep(runif(1) * 10)

for(module in module_list){
	move_input = paste0(movement_path_start, module, "/All-Other.RData")
	move_output = paste0(output_dir, module, "/All-Other.RData")
	if(file.exists(move_output)){
		print(paste0("Module ", module, " Skipped."))
		next
	} else{
		print(paste0("Module ", module, " Is Being Processed."))
		move = data.table()
		save(move, file=move_output)
	}

	time_start = proc.time()[3]

	if(!file.exists(move_input)){
		print(paste0("All-Other.RData In Module ", module, " DOES NOT EXIST"))
		next
	}

	load(move_input) # movement

	# --------------------------- Core Unit Processing --------------------------- #

	# Prmult
	movement[, `:=`(price=price/prmult, week_end=as.numeric(week_end))]

	# Remove Duplicates
	setkey(movement, upc, upc_ver_uc, store_code_uc, week_end)
	movement = unique(movement)
	movement = corrections[movement]
	movement[is.na(upc_ver_uc_corrected), upc_ver_uc_corrected:=upc_ver_uc]

	# Filter for process
	filter_table = movement[, .(span=(max(week_end)-min(week_end))/7+1, obs=.N), by=.(upc, upc_ver_uc_corrected, store_code_uc)]
	filter_table[, percent_covered:=obs/span]
	filter_table[, processed:=(percent_covered>=0.2)]

	setkey(filter_table, upc, upc_ver_uc_corrected, store_code_uc)
	setkey(movement, upc, upc_ver_uc_corrected, store_code_uc)
	movement = filter_table[,.(upc, upc_ver_uc_corrected, store_code_uc, processed)][movement]

	not_processed = movement[processed==FALSE]
	movement = movement[processed==TRUE]
	# Create move_lookup_table for the CPP algorithm
	move_lookup_table = movement[, .(index_begin=.I[1], index_end=.I[.N]), by=.(upc, upc_ver_uc_corrected, store_code_uc)]

	# Processing
	# -------------------------------------------------- #
	# Step 1: Calculating Base Price
	# -------------------------------------------------- #
	structures = movement[, .(upc, upc_ver_uc_corrected, store_code_uc, week_end)]
	base_prices = as.data.table(BasePriceDetectionHybridAlgorithm(movement, move_lookup_table))
	base_prices = cbind(structures, base_prices)
	# -------------------------------------------------- #
	# Step 2: Imputing Base Price
	# -------------------------------------------------- #	
	# 1. Get full week structures for each upc-version-store combination
	# (Equivalent to "ExpandMissingWeeks" in C++)
	structures[, `:=`(min_date=min(week_end), max_date=max(week_end)), by=.(upc, upc_ver_uc_corrected, store_code_uc)]
	structures[, week_end:=NULL]
	structures = unique(structures)
	structures = structures[, .(week_end=seq(min_date, max_date, 7)), by=.(upc, upc_ver_uc_corrected, store_code_uc)]
	# 2. Rolling join and adjust base prices
	# Split base_prices
	base_prices[base_price==0, base_price:=NA]
	base_prices_non_NA = base_prices[!is.na(base_price)]
	base_prices_NA = base_prices[is.na(base_price)]
	# Set Keys
	setkey(base_prices_NA, upc, upc_ver_uc_corrected, store_code_uc, week_end)
	setkey(base_prices_non_NA, upc, upc_ver_uc_corrected, store_code_uc, week_end)
	setkey(structures, upc, upc_ver_uc_corrected, store_code_uc, week_end)
	# Rolling Join
	base_prices = base_prices_non_NA[structures, roll=TRUE]
	# Adjust wrongly imputed base prices
	base_prices_NA[, `:=`(identifier=1, base_price=NULL)]
	base_prices = base_prices_NA[base_prices]
	base_prices[identifier==1, base_price:=NA]
	base_prices[, identifier:=NULL]
	rm(structures, base_prices_NA, base_prices_non_NA, move_lookup_table)
	# -------------------------------------------------- #
	# Step 3: Imputing Prices and Cleaning
	# -------------------------------------------------- #
	setkey(movement, upc, upc_ver_uc_corrected, store_code_uc, week_end)
	move = movement[base_prices]
	move[, imputed_price:=price]
	move[is.na(imputed_price), imputed_price:=base_price] # Impute prices for zero-sale weeks
	move[is.na(units), units:=0]
	move[, `:=`(processed=NULL, price=NULL)]
	move[, `:=`(week_end=as.Date(week_end, origin="1970-01-01"))]
	setnames(move, "prmult", "prmult_for_reference_only_do_not_use")

	# -------------------------------------------------- #
	# Step 4: Adding not_processed to the file, save
	# -------------------------------------------------- #
	not_processed[, `:=`(processed=NULL, base_price=NA, imputed_price=price)]
	not_processed[, `:=`(week_end=as.Date(week_end, origin="1970-01-01"), price=NULL)]
	setnames(not_processed, "prmult", "prmult_for_reference_only_do_not_use")

	setcolorder(move, c("upc", "upc_ver_uc_corrected", "store_code_uc",
		"week_end", "base_price", "imputed_price", "units", "feature", "display", 
		"upc_ver_uc", "prmult_for_reference_only_do_not_use"))
	setcolorder(not_processed, c("upc", "upc_ver_uc_corrected", "store_code_uc",
		"week_end", "base_price", "imputed_price", "units", "feature", "display", 
		"upc_ver_uc", "prmult_for_reference_only_do_not_use"))

	# --------------------------- Core Unit Processing Finished --------------------------- #

	save(move, not_processed, filter_table, file=move_output)

	print(paste0("Module ", module, " Finished. Time Elapsed: ", (proc.time()[3]-time_start)/60, " Minutes."))
}












