rm(list=ls())
library(data.table)
library(bit64)
library(Rcpp)
library(RcppArmadillo)
options(scipen=999)
setNumericRounding(0)

# Parallel computing implemented by skipping mechanisms in for loops (for now)
# This scripts processes products outside All-Other.RData

run_on_linux = TRUE

data_dir = "/grpshares/hitsch_shapiro_ads/data/RMS/" # Directory where results from 1 to 3 are stored
movement_path_start = paste0(data_dir, "Modules/")

Rcpp::sourceCpp("../Base-Price-Imputation/Scripts/cpp-Source/Price-Algorithm.cpp")

correct_meta = FALSE
test = FALSE

# load version correction reference data
labeled_file_path = "./Processed-Data/Corrections.RData"
load(labeled_file_path)
corrections = corrections[false_positive==1, .(upc, old_version, new_version)]

# What we will do: Change new version back to old version

# Pre-Process: upc with multiple versions (count can only be 1 or 2)
corrections[, count:=.N, by=upc]
corrections[count>1, old_version:=min(old_version), by=upc]
corrections[, count:=NULL]
setnames(corrections, c("new_version", "old_version"), c("upc_ver_uc", "upc_ver_uc_corrected"))
setkey(corrections, upc, upc_ver_uc)

# Correct and Process
output_dir = "~/data/RMS-Build-2016/RMS-Processed/Modules/" # Change appropriately
if(!dir.exists(output_dir)){
	dir.create(output_dir)
}

module_list = as.numeric(list.files(movement_path_start))

for(module in module_list){
	module_output = paste0(output_dir, module, "/")
	if(!dir.exists(module_output)){
		dir.create(module_output)
	}
}

# Get upcs that are individually stored
indv_upc_path = paste0(dir, "Individually-Stored-UPCs.RData")
if(!file.exists(indv_upc_path)){
	products_to_process = data.table()
	for(module in module_list){
		upcs = list.files(paste0(movement_path_start, module, "/"))
		upcs = as.numeric(sub(".RData", "", upcs))
		to_process = data.table(upc=upcs, product_module_code = module)
		products_to_process = rbindlist(list(to_process, products_to_process))
	}
	products_to_process = products_to_process[!is.na(upc)]
	save(products_to_process, file=indv_upc_path)
} else{
	load(indv_upc_path)
}

if(test){
	products_to_process = products_to_process[product_module_code==1484]
}

print(nrow(products_to_process))

split_length = round(nrow(products_to_process)/50) + 1

Sys.sleep(runif(1) * 10)

for(i in 1:50){

	spaceholder_path = paste0(output_dir, "Spaceholder-", i, ".RData")
	if(file.exists(spaceholder_path)){
		print(paste0("Block ", i, " Skipped."))
		next
	} else{
		print(paste0("Block ", i, " Is Being Processed."))
		save(i, file=spaceholder_path)
	}

	first_index = (i-1) * split_length + 1
	last_index = min(i * split_length, nrow(products_to_process))

	block = products_to_process[first_index:last_index]
	print(nrow(block))

	time_start = proc.time()[3]

	missing_upc_list = c()

	for(j in 1:nrow(block)){
		UPC = block[j, upc]
		module = block[j, product_module_code]

		move_output = paste0(output_dir, module, "/", UPC, ".RData")
		if(file.exists(move_output)){
			next
		}

		move_input = paste0(movement_path_start, module, "/", UPC, ".RData")
		if(!file.exists(move_input)){
			print(paste0("UPC ", UPC, " From Module ", module, " Does Not Exist."))
			next
		} else{
			load(move_input) # Movement
		}

		# Prmult
		movement[, `:=`(price=price/prmult, week_end=as.numeric(week_end))]

		# Remove Duplicates
		setkey(movement, upc, upc_ver_uc, store_code_uc, week_end)
		movement = unique(movement)
		movement = corrections[movement]
		movement[is.na(upc_ver_uc_corrected), upc_ver_uc_corrected:=upc_ver_uc]

		# Filter for process
		filter_table = movement[, .(span=(max(week_end)-min(week_end))/7+1, obs=.N), by=.(upc, upc_ver_uc_corrected, store_code_uc)]
		filter_table[, percent_covered:=obs/span]
		filter_table[, processed:=(percent_covered>=0.2)]

		setkey(filter_table, upc, upc_ver_uc_corrected, store_code_uc)
		setkey(movement, upc, upc_ver_uc_corrected, store_code_uc)
		movement = filter_table[,.(upc, upc_ver_uc_corrected, store_code_uc, processed)][movement]

		not_processed = movement[processed==FALSE]
		movement = movement[processed==TRUE]
		# Create move_lookup_table for the CPP algorithm
		move_lookup_table = movement[, .(index_begin=.I[1], index_end=.I[.N]), by=.(upc, upc_ver_uc_corrected, store_code_uc)]

		# Processing
		# -------------------------------------------------- #
		# Step 1: Calculating Base Price
		# -------------------------------------------------- #
		structures = movement[, .(upc, upc_ver_uc_corrected, store_code_uc, week_end)]
		base_prices = as.data.table(BasePriceDetectionHybridAlgorithm(movement, move_lookup_table))
		base_prices = cbind(structures, base_prices)
		# -------------------------------------------------- #
		# Step 2: Imputing Base Price
		# -------------------------------------------------- #	
		# 1. Get full week structures for each upc-version-store combination
		# (Equivalent to "ExpandMissingWeeks" in C++)
		tryCatch(
			{
				structures[, `:=`(min_date=min(week_end), max_date=max(week_end)), by=.(upc, upc_ver_uc_corrected, store_code_uc)]
				structures[, week_end:=NULL]
				structures = unique(structures)
				structures = structures[, .(week_end=seq(min_date, max_date, 7)), by=.(upc, upc_ver_uc_corrected, store_code_uc)]
			},
			error = function(cond){
				print(paste0("ERROR: ", cond))
				next
			},
			warning = function(cond){
				print(paste0("WARNING: ", cond))
				next
			},
			finally = {

			}
		)
		# 2. Rolling join and adjust base prices
		# Split base_prices
		base_prices[base_price==0, base_price:=NA]
		base_prices_non_NA = base_prices[!is.na(base_price)]
		base_prices_NA = base_prices[is.na(base_price)]
		# Set Keys
		setkey(base_prices_NA, upc, upc_ver_uc_corrected, store_code_uc, week_end)
		setkey(base_prices_non_NA, upc, upc_ver_uc_corrected, store_code_uc, week_end)
		setkey(structures, upc, upc_ver_uc_corrected, store_code_uc, week_end)
		# Rolling Join
		base_prices = base_prices_non_NA[structures, roll=TRUE]
		# Adjust wrongly imputed base prices
		base_prices_NA[, `:=`(identifier=1, base_price=NULL)]
		base_prices = base_prices_NA[base_prices]
		base_prices[identifier==1, base_price:=NA]
		base_prices[, identifier:=NULL]
		rm(structures, base_prices_NA, base_prices_non_NA, move_lookup_table)
		# -------------------------------------------------- #
		# Step 3: Imputing Prices and Cleaning
		# -------------------------------------------------- #
		setkey(movement, upc, upc_ver_uc_corrected, store_code_uc, week_end)
		move = movement[base_prices]
		move[, imputed_price:=price]
		move[is.na(imputed_price), imputed_price:=base_price] # Impute prices for zero-sale weeks
    	move[is.na(units), units:=0]
    	move[, `:=`(processed=NULL, price=NULL)]
    	move[, `:=`(week_end=as.Date(week_end, origin="1970-01-01"))]
    	setnames(move, "prmult", "prmult_for_reference_only_do_not_use")
    	
		# -------------------------------------------------- #
		# Step 4: Adding not_processed to the file, save
		# -------------------------------------------------- #
		not_processed[, `:=`(processed=NULL, base_price=NA, imputed_price=price)]
		not_processed[, `:=`(week_end=as.Date(week_end, origin="1970-01-01"), price=NULL)]
		setnames(not_processed, "prmult", "prmult_for_reference_only_do_not_use")

		setcolorder(move, c("upc", "upc_ver_uc_corrected", "store_code_uc",
    		"week_end", "base_price", "imputed_price", "units", "feature", "display", 
    		"upc_ver_uc", "prmult_for_reference_only_do_not_use"))
		setcolorder(not_processed, c("upc", "upc_ver_uc_corrected", "store_code_uc",
    		"week_end", "base_price", "imputed_price", "units", "feature", "display", 
    		"upc_ver_uc", "prmult_for_reference_only_do_not_use"))

		save(move, not_processed, filter_table, file=move_output)

		if(j%%200 == 0 | j==nrow(block)){
			print(paste0("Processed ", j, " out of ", nrow(block), " cases."))
			print(paste0("Time Elapsed: ", (proc.time()[3]-time_start)/60, " Minutes."))
		}
	}

}






