# ------------------------------------------------------------------------------
# Ad-Intel/Build-Regular.R 
#
# ------------------------------------------------------------------------------
# This script builds the all regular parts of the Ad-Intel Data:
# National TV, Spot TV, Local Clearances
#
# Prerequisite: None
# ------------------------------------------------------------------------------
rm(list = ls())

library(data.table)
library(bit64)
library(foreach)
library(doParallel)
library(lubridate)

source_dir <- "/kilts/adintel"
output_dir <- "/project/ghitsch/data/Ad-Intel-2020"

YEAR_START <- 2010L
YEAR_END <- 2018L

Init.Impute.Table <- function(){
  impute_table <- data.table(use.month = c(1,3,4,6,8,9,10,12),
                             month1 = c(2,2,2,5,7,7,7,11),
                             month2 = c(11,5,5,7,11,11,11,2),
                             weight1 = c(2/3,2/3,1/3,1/2,3/4,1/2,1/4,2/3),
                             weight2 = c(1/3,1/3,2/3,1/2,1/4,1/2,3/4,1/3))
  impute_table <- impute_table[rep(1:.N, each=YEAR_END-YEAR_START+1)]
  impute_table[, use.year := rep(YEAR_START:YEAR_END, 8)]
  impute_table[, year1 := use.year]
  impute_table[, year2 := use.year]
  impute_table[use.month==1, year2 := use.year - 1L]
  impute_table[use.month==12, year2 := use.year + 1L]
  impute_table[year2 == YEAR_START - 1L, `:=`(year2=NA, weight1=1, weight2=0, month2=NA)]
  impute_table[year2 == YEAR_END + 1L, `:=`(year2=NA, weight1=1, weight2=0, month2=NA)]
  return(impute_table)
}

Read.National.Occurrences <- function(year, month){
  
  dir <- paste0(source_dir, "/",year, "/Occurrences")
  
  occ <- fread(paste0(dir, '/NetworkTV.tsv'), showProgress=F)
  
  # Extract month level data
  
  occ <- occ[as.numeric(substr(AdDate,6,7)) == month]
  
  occ[,`:=`(AdDate=as.Date(AdDate),
            PrimBrandCode=as.integer(PrimBrandCode),
            ScndBrandCode=as.integer(ScndBrandCode),
            TerBrandCode=as.integer(TerBrandCode),
            ImpressionType=`Impression Type`)]
  
  cols_to_keep <- c('AdDate', 'AdTime', 'MediaTypeID', 
                    'PrimBrandCode', 'ScndBrandCode',
                    'TerBrandCode', 'Duration', 'Units', 'Spend',
                    "DistributorCode", "TVDaypartCode", 
                    "ImpressionType", "NielsenProgramCode", "TelecastNumber",
                    "DistributorID", "TimeIntervalNumber")
  
  occ <- occ[, cols_to_keep, with=FALSE]
  return(occ)
}

Read.Spot.Occurrences <- function(year, month){
  
  dir <- paste0(source_dir, "/",year, "/Occurrences")
  
  # Extract month level data
  
  ym_start <- as.Date(paste0(year,"-",month, "-01"))
  
  if(month != 12){
    ym_end <- as.Date(paste0(year,"-",month+1, "-01")) - 1
  } else{
    ym_end <- as.Date(paste0(year,"-12-31"))
  }
  
  ym_start <- as.Date(cut(ym_start, "week"))
  
  if(wday(ym_end) != 1){
    ym_end <- as.Date(cut(ym_end, "week")) - 1
  } 
  
  occ <- fread(paste0(dir, '/SpotTV.tsv'), showProgress=F)
  
  occ[,`:=`(AdDate=as.Date(AdDate))]
  
  occ <- occ[AdDate >= ym_start & AdDate <= ym_end]
  
  occ[,`:=`(PrimBrandCode=as.integer(PrimBrandCode),
            ScndBrandCode=as.integer(ScndBrandCode),
            TerBrandCode=as.integer(TerBrandCode))]
  
  occ[MediaTypeID != 5, Spend := NA]
  
  cols_to_keep <- c('AdDate', 'AdTime', 'MediaTypeID', 'MarketCode',
                    'PrimBrandCode', 'ScndBrandCode',
                    'TerBrandCode', 'Duration', 'Units', 'Spend',
                    "DistributorCode", "TVDaypartCode", 
                    "PeriodYearMonth","DistributorID","DayOfWeek","TimeIntervalNumber")
  occ <- occ[, cols_to_keep, with=FALSE]
  
  setkey(occ, AdDate)
  
  return(occ)
}

Read.Impressions <- function(year, month, type){
  if(!(type %in% c("National", "Spot", "SpotRadio"))){
    error <- paste0("In function Read.Impressions: unknown input type ", type,
                    "\nTypes allowed are: National, Spot, SpotRadio")
    stop(error)
  }
  
  dir <- paste0(source_dir, "/",year, "/Impressions")
 
  imp <- fread(paste0(dir, '/Imp', type, "TV.tsv"), showProgress=F)
  
  if(type == "National"){
    imp.p <- imp[ImpressionType == "P"]
    imp.t <- imp[ImpressionType == "T"]
    imp.t <- imp.t[as.numeric(substr(ImpressionDate,6,7)) == month]
    
    imp <- rbind(imp.p, imp.t)
    rm(imp.p, imp.t)
  }
  if(type == "Spot"){
    imp <- imp[as.numeric(substr(PeriodYearMonth,5,6)) == month]
  }
  
  remove.cols = grep("male|Male|ww|children",names(imp))
  imp[, (remove.cols) := NULL]
  
  setnames(imp, 'TV_HH', 'imp_total')
  
  if(type == 'National'){
    # For spanish stuff, use hispanic flag = "Y"; otherwise, use "N"
    imp <- imp[DataStreamID==2]
    imp1 <- imp[MediaTypeID %in% c(4,22) & HispanicFlag=="Y"]
    imp2 <- imp[MediaTypeID %in% 1:3]
    imp <- rbind(imp1, imp2, fill=TRUE)
    imp[, HispanicFlag := NULL]
  }
  if(type == 'Spot'){
    imp <- imp[HispanicFlag == "N"]
    imp[, c('MediaTypeID', 'HispanicFlag', 'UC_dim_Bridge_occ_ImpSpotTV_key') := NULL]
  }
  
  suppressWarnings(imp[, c('DataStreamID', 'ActualFlag', 'PeriodYearMonth') := NULL])
  return(imp)
}

Read.Universe.Estimates <- function(type){
  if(!(type %in% c("National", "Spot", "SpotRadio"))){
    error <- paste0("In function Read.Impressions: unknown input type ", type,
                    "\nTypes allowed are: National, Spot, SpotRadio")
    stop(error)
  }
  
  dir <- paste0(source_dir, "/2018/Universe_Estimates")
  
  ue <- fread(paste0(dir, '/UE', type, "TV.tsv"), showProgress=F)
  
  ue[, `:=`(StartDate=as.Date(StartDate),
            EndDate=as.Date(EndDate))]
  
  remove.cols = grep("male|Male|ww|children",names(ue))
  ue[, (remove.cols):=NULL]
  ue <- ue[HispanicFlag == 'N']
  ue[, HispanicFlag := NULL]
  setnames(ue, 'TV_HH', 'ue_total')
  return(ue)
}

Build.Spot.Occurrence.Level <- function(year, month){
  cat('Processing', year, month, '...', '\n')
  # ------------------------------ #
  # Read Data
  # ------------------------------ #
  
  # Occurrences
  occ <- Read.Spot.Occurrences(year, month)
  # Impressions and Universe
  local_imp <- Read.Impressions(year, month,'Spot')
  local_ue <- Read.Universe.Estimates('Spot')
  
  
  # Compute the start date and the end date of the month
  ym_start <- as.Date(paste0(year,"-",month, "-01"))
  
  if(month != 12){
    ym_end <- as.Date(paste0(year,"-",month+1, "-01")) - 1
  } else{
    ym_end <- as.Date(paste0(year,"-12-31"))
  }
  
  ym_start <- as.Date(cut(ym_start, "week"))
  
  if(wday(ym_end) != 1){
    ym_end <- as.Date(cut(ym_end, "week")) - 1
  } 
  
  # Filter out related UEs
  local_ue <- local_ue[(StartDate <= ym_start & EndDate >= ym_start) | (StartDate <= ym_end & EndDate >= ym_end)]
  
  # Impression Imputation
  
  occ_current <- occ[PeriodYearMonth==as.character(year*100+month)]
  occ_impute <- occ[PeriodYearMonth!=as.character(year*100+month)]
  rm(occ)
    
  occ_current[, PeriodYearMonth := NULL]
  occ_impute[, PeriodYearMonth := NULL]

  
  # (1) Merge Current
  occimp_current <- merge(occ_current, local_imp, 
                          by=c('DistributorID', 'DayOfWeek', 'TimeIntervalNumber'), all.x = T)
  occimp_current[, c('DistributorID', 'DayOfWeek') := NULL]
  
  if(nrow(occ_impute) != 0){
    # (2) Impute
    impute_entry <- impute_table[use.year==year & use.month==month]
    ## 2.1 Month 1
    impute_imp1 <- Read.Impressions(impute_entry$year1, impute_entry$month1, 'Spot')
    setnames(impute_imp1, 'imp_total', 'imp_total1')
    occimp_impute1 <- merge(occ_impute, impute_imp1,
                            by=c('DistributorID', 'DayOfWeek', "TimeIntervalNumber"), all.x=T)
    ## 2.2 Month 2
    if(impute_entry$weight2 > 0){
      impute_imp2 <- Read.Impressions(impute_entry$year2, impute_entry$month2, 'Spot')
      setnames(impute_imp2, 'imp_total', 'imp_total2')
      occimp_impute <- merge(occimp_impute1, impute_imp2,
                             by=c('DistributorID', 'DayOfWeek', "TimeIntervalNumber"), all.x=T)
      occimp_impute[, imp_total := imp_total1 * impute_entry$weight1 + imp_total2 * impute_entry$weight2]
      occimp_impute[, c('DistributorID', 'DayOfWeek', 'imp_total1', 'imp_total2') := NULL]
    } else{
      occimp_impute <- occimp_impute1
      occimp_impute[, c('DistributorID', 'DayOfWeek') := NULL]
      setnames(occimp_impute, 'imp_total1', 'imp_total')
    }
    # (3) Combine
    occimp <- rbind(occimp_current, occimp_impute, fill=TRUE)
  } else{
    occimp <- occimp_current
  }
  
  # (4) Merge UE
  
  if(local_ue[,.N] != length(unique(local_ue[,MarketCode]))){
    
    local_ue1 <- local_ue[,.SD[1], by = MarketCode]
    local_ue2 <- local_ue[,.SD[.N], by = MarketCode]
    
    setkey(local_ue1, MarketCode)
    setkey(local_ue2, MarketCode)
    setkey(occimp, MarketCode)
    
    occimpue1 <- local_ue1[occimp, roll=TRUE]
    occimpue2 <- local_ue2[occimp, roll=TRUE]
    
    occimpue1 <- occimpue1[StartDate <= AdDate & EndDate >= AdDate]
    occimpue2 <- occimpue2[StartDate <= AdDate & EndDate >= AdDate]
    
    occimpue <- rbindlist(list(occimpue1, occimpue2))
    
    occimpue <- occimpue[order(MarketCode)]
    
  } else{
    
    setkey(local_ue, MarketCode)
    setkey(occimp, MarketCode)
    occimpue <- local_ue[occimp, roll=TRUE]
    
  }

  # (5) Re-Arrange
  setnames(occimpue, c("imp_total", "ue_total"), c("Impression", "Universe"))
  cols_to_keep <- c("AdDate", "AdTime", "MediaTypeID", "TVDaypartCode", 
                    "MarketCode", "DistributorCode", 
                    "PrimBrandCode", "ScndBrandCode", "TerBrandCode",
                    "Duration", "Spend",
                    "Impression", "Universe")
  occimpue <- occimpue[, cols_to_keep, with=FALSE]
  
  # To match the previous data build 
  occimpue[, `:=`(DistributorCode = as.numeric(DistributorCode),
                  AdTime = gsub("\\..*", "", AdTime))]
  
  occimpue[PrimBrandCode == 0, PrimBrandCode := NA]
  occimpue[ScndBrandCode == 0, ScndBrandCode := NA]
  occimpue[TerBrandCode == 0, TerBrandCode := NA]
  
  setindex(occimpue, NULL)
  setkey(occimpue, MarketCode, AdDate)
  
  cat("Done\n")
  return(occimpue)
}

Build.National.Occurrence.Level <- function(year, month){
  cat('Processing', year, month, 'National...', '\n')
  occ <- Read.National.Occurrences(year, month)
  imp <- Read.Impressions(year, month, "National")
  ue <- Read.Universe.Estimates("National")
  ue <- ue[order(StartDate)]
  
  occ.P <- occ[ImpressionType=="P", -c("DistributorID", "TimeIntervalNumber", "ImpressionType"), with=F]
  occ.T <- occ[ImpressionType=="T", -c("NielsenProgramCode", "TelecastNumber", "ImpressionType"), with=F]
  imp.P <- imp[ImpressionType=="P", .(MediaTypeID, NielsenProgramCode, TelecastNumber, Impression=imp_total)]
  imp.T <- imp[ImpressionType=="T", .(MediaTypeID, DistributorID, AdDate=as.Date(ImpressionDate), TimeIntervalNumber, Impression=imp_total)]
  
  # Type P Impressions (99.99 percent of the data...)
  occimp.P <- merge(occ.P, imp.P, by=c("MediaTypeID", "NielsenProgramCode", "TelecastNumber"), all.x=T)
  occimp.P[, c("NielsenProgramCode", "TelecastNumber") := NULL]
  # Type T impressions
  occimp.T <- merge(occ.T, imp.T, by=c("MediaTypeID", "AdDate", "DistributorID", "TimeIntervalNumber"), all.x=T)
  occimp.T[, c("DistributorID", "TimeIntervalNumber") := NULL]
  # Combine
  occimp <- rbindlist(list(occimp.P, occimp.T), fill=TRUE)
  rm(occ.P, occ.T, imp.P, imp.T, occimp.P, occimp.T)
  
  # UE
  
  # Compute the start date and the end date of the month
  ym_start <- as.Date(paste0(year,"-",month, "-01"))
  
  if(month != 12){
    ym_end <- as.Date(paste0(year,"-",month+1, "-01")) - 1
  } else{
    ym_end <- as.Date(paste0(year,"-12-31"))
  }
  
  # Filter out related UEs
  ue <- ue[(StartDate <= ym_start & EndDate >= ym_start) | (StartDate <= ym_end & EndDate >= ym_end)]
  
  if(ue[,.N] != 1){
    date1 <- ue[1,.(StartDate, EndDate)]
    date2 <- ue[2,.(StartDate, EndDate)]
    
    occimpue1 <- occimp[date1[,StartDate] <= AdDate & date1[,EndDate] >= AdDate][,
      Universe := ue[1, ue_total]
    ]
    occimpue2 <- occimp[date2[,StartDate] <= AdDate & date2[,EndDate] >= AdDate][,
      Universe := ue[2, ue_total]
    ]
        
    occimpue <- rbindlist(list(occimpue1, occimpue2))
  } else{
    ue <- ue[,ue_total]
    occimpue <- occimp[,Universe := ue]
  }
  
  rm(occimp)
  
  cols_to_keep <- c("AdDate", "AdTime", "MediaTypeID", "TVDaypartCode", 
                    "DistributorCode", "PrimBrandCode", "ScndBrandCode", "TerBrandCode",
                    "Duration", "Spend",
                    "Impression", "Universe")
  occimpue <- occimpue[, cols_to_keep, with=F]
  
  # To match the previous data build 
  occimpue[,AdTime := gsub("\\..*", "", AdTime)]
  occimpue[PrimBrandCode == 0, PrimBrandCode := NA]
  occimpue[ScndBrandCode == 0, ScndBrandCode := NA]
  occimpue[TerBrandCode == 0, TerBrandCode := NA]
  
  setindex(occimpue, NULL)
  setkey(occimpue, AdDate)
  
  cat("Done\n")
  return(occimpue)
}

Read.Brand.Info <- function(year){
  
  dir <- paste0(source_dir, "/",year, "/References")

  pcc <- fread(paste0(dir, "/ProductCategories.tsv"), showProgress = F)
  brd <- fread(paste0(dir, "/Brand.tsv"), showProgress = F)
  adv <- fread(paste0(dir, "/Advertiser.tsv"), showProgress = F)
  
  prod_cols <- c("BrandCode","BrandDesc","BrandVariant",
                 "AdvParentCode","AdvParentDesc","AdvSubsidCode","AdvSubsidDesc",
                 "ProductID","ProductDesc","PCCSubCode","PCCSubDesc","PCCMajCode",
                 "PCCMajDesc","PCCIndusCode","PCCIndusDesc")
  
  pccbrd <- merge(pcc, brd, by="ProductID", allow.cartesian = TRUE)
  pccbrdadv <- merge(pccbrd, adv, by=c("AdvParentCode","AdvSubsidCode"))
  
  pccbrdadv[, PCCSubCode.y:=NULL]
  setnames(pccbrdadv, "PCCSubCode.x", "PCCSubCode")
  
  pccbrdadv <- pccbrdadv[, prod_cols, with=F]
  return(pccbrdadv)
}	

impute_table <- Init.Impute.Table()

Build.All <- function(year, month){
  cat(year, month, "\n")
  national <- Build.National.Occurrence.Level(year, month)
  spot <- Build.Spot.Occurrence.Level(year, month)
  brands <- Read.Brand.Info(year)
  
  clearance  <- spot[MediaTypeID %in% c(13, 14)]
  spot <- spot[MediaTypeID == 5]
  
  saveRDS(national, file=paste0(output_dir, "/Occurrence-Level/National-TV/", year, "-", month, ".rds"))
  saveRDS(spot, file=paste0(output_dir, "/Occurrence-Level/Spot-TV/", year, "-", month, ".rds"))
  saveRDS(clearance, file=paste0(output_dir, "/Occurrence-Level/Local-Clearance/", year, "-", month, ".rds"))
  saveRDS(brands, file=paste0(output_dir, "/Dynamic-References/Brands/", year, "-", month, ".rds"))

  rm(national, spot, clearance, brands)
  
  return(0)
}

# Build
year_list <- rep(YEAR_START:YEAR_END, each=12)
month_list <- rep(1:12, YEAR_END-YEAR_START+1)
n <- length(year_list)

cl <- makeCluster(2, type="FORK", outfile="")
registerDoParallel(cl)

status <- foreach(i=1:n) %do% {
  x <- "Failed"
  try(x <- Build.All(year_list[i], month_list[i]))
  x
}

stopCluster(cl)

print("Done!")
