library(data.table)
library(bit64)
library(lubridate)
library(foreach)
library(doParallel)

# Functions
Process.No.Broadcast.Delay <- function(STATION, NETWORK, TIMEZONE, start_date=date("2010-01-01"), end_date=date("2018-12-30")){
  cat("----Processing station", STATION, "in network", NETWORK, "in time zone", TIMEZONE, "\n")
  
  # ---------------------------------------- #
  # 0. Configure Parameters
  # ---------------------------------------- #
  date_range <- c(start_date, end_date)
  MOVED_AROUND_CLEARANCE <- 11100
  if(!(TIMEZONE %in% c("US/Eastern", "US/Central"))){
    stop("Process.No.Broadcast.Delay only accepts eastern and central time")
  }
  
  # ---------------------------------------- #
  # 1. Load Station-Level Data
  # ---------------------------------------- #
  
  cat("Loading data...")
  
  # 1.1 National
  # Files are in Eastern time. First convert using eastern time, then change to market time
  nat <- readRDS(paste0(national_dir, NETWORK, ".rds"))
  nat[, time := ymd_hms(paste(as.character(AdDate), AdTime), tz="US/Eastern")]
  nat[, time:=with_tz(time, tz=TIMEZONE)]
  nat_brands <- nat[!is.na(time), .(time, PrimBrandCode, ScndBrandCode, TerBrandCode)]
  nat <- nat[!is.na(time), .(time, type="Network", PrimBrandCode, Duration)]
  
  # 1.2 Network Clearance
  # Files are in local time
  clear <- readRDS(paste0(clearance_dir, STATION, ".rds"))
  clear[, time := ymd_hms(paste(as.character(AdDate), AdTime), tz=TIMEZONE)]
  clear_brands <- clear[!is.na(time), .(time, PrimBrandCode, ScndBrandCode, TerBrandCode)]
  clear <- clear[!is.na(time), .(time, type="Clear", PrimBrandCode, Duration)]
  
  # 1.3 Spot (Can be empty)
  spot_file <- paste0(spot_dir, STATION, ".rds")
  if(!file.exists(spot_file)){
    spot <- NULL
  } else{
    spot <- readRDS(spot_file)
    spot[, time := ymd_hms(paste(as.character(AdDate), AdTime), tz=TIMEZONE)]
    spot <- spot[!is.na(time), .(time, type="Other", PrimBrandCode, Duration)]
  }
  
  # 1.4 Syndicated (Can be empty)
  synd_file <- paste0(synd_dir, STATION, ".rds")
  if(!file.exists(synd_file)){
    synd <- NULL
  } else{
    synd <- readRDS(synd_file)
    synd[, time := ymd_hms(paste(as.character(AdDate), AdTime), tz=TIMEZONE)]
    synd <- synd[!is.na(time), .(time, type="Other", PrimBrandCode, Duration)]
  }
  
  # 1.5 Follow date range of clearance
  clear_range <- range(date(clear$time))
  start <- max(clear_range[1], date_range[1])
  end <- min(clear_range[2], date_range[2])
  use_range <- c(start, end)
  
  nat <- nat[date(time) %between% use_range]
  clear <- clear[date(time) %between% use_range]
  if(!is.null(spot)){
    spot <- spot[date(time) %between% use_range]
  }
  if(!is.null(synd)){
    synd <- synd[date(time) %between% use_range]
  }
  
  # 1.6 Stack all occurrences
  setorder(nat, time)
  setorder(clear, time)
  nat[, id := 1:.N]
  clear[, id := nrow(nat) + (1:.N)]
  
  if(is.null(spot) & is.null(synd)){
    other <- data.table()
  } else{
    other <- rbind(spot, synd, fill=TRUE)
    setorder(other, time)
    other[, id := nrow(nat) + nrow(clear) + (1:.N)]
  }
  
  stacked <- rbind(nat, clear, other, fill=T)
  setorder(stacked, time)
  
  cat("Done\n")
  
  # ---------------------------------------- #
  # 2. Find Direct Matches
  # ---------------------------------------- #
  
  cat("Finding direct matches...")
  
  # 2.1 Find forward and backward matches (no time limit)
  stacked[, `:=`(forward=0, backward=0)]
  stacked[, time_from_last := as.integer(time - shift(time,1,type="lag"))]
  stacked[, time_to_next := as.integer(shift(time,1,type="lead")-time)]
  
  stacked[type=="Network" & 
            shift(type,1,type="lead")=="Clear" & 
            shift(PrimBrandCode,1,type="lead")==PrimBrandCode, forward := 1]
  stacked[type=="Clear" & shift(forward,1,type="lag")==1, forward := 1]
  
  stacked[type=="Network" & 
            shift(type,1,type="lag")=="Clear" & 
            shift(PrimBrandCode,1,type="lag")==PrimBrandCode, backward := 1]
  stacked[type=="Clear" & shift(backward,1,type="lead")==1, backward := 1]
  
  # 2.2 Clean up conflicts: if both forward and backward match, choose the closest and remove the orphans
  stacked[type=="Network" & forward==1 & backward==1 & time_from_last < time_to_next, forward := 0]
  stacked[type=="Network" & forward==1 & backward==1 & time_from_last >= time_to_next, backward := 0]
  stacked[type=="Clear" & forward==1 & backward==1 & time_from_last >= time_to_next, forward := 0]
  stacked[type=="Clear" & forward==1 & backward==1 & time_from_last < time_to_next, backward := 0]
  
  stacked[type=="Network" & backward==1 & shift(backward,1,type="lag")!=1, backward := 0]
  stacked[type=="Network" & forward==1 & shift(forward,1,type="lead")!=1, forward := 0]
  stacked[type=="Clear" & backward==1 & shift(backward,1,type="lead")!=1, backward := 0]
  stacked[type=="Clear" & forward==1 & shift(forward,1,type="lag")!=1, forward := 0]
  
  # 2.3 Extract matches and non_recordings; keep others
  direct_matches <- stacked[forward==1 | backward==1, .(id, time, type, PrimBrandCode, Duration, forward, backward)]
  not_recorded <- stacked[forward==0 & backward==0 & type=="Network" & (hour(time) %in% 2:4), .(id, time, type, PrimBrandCode)]
  s2 <- stacked[forward==0 & backward==0 & !(type=="Network" & (hour(time) %in% 2:4))]
  s2[, c("forward", "backward", "time_to_next", "time_from_last") := NULL]
  
  cat("Done\n")
  
  # ---------------------------------------- #
  # 3. Find "Moved Around" Matches
  # ---------------------------------------- #
  
  cat("Finding moved around matches...")
  
  # 3.1 Match the leftover network brands with the leftover clearance brands
  un_net <- s2[type=="Network", .(id, time, PrimBrandCode, Duration)]
  un_clear <- s2[type=="Clear", .(id, time, PrimBrandCode, Duration)]
  merged <- merge(un_net, un_clear, by="PrimBrandCode", allow.cartesian=T)
  
  # 3.2 For each unexplained CLEARANCE ad, find the closest network ad
  merged[, diff := as.integer(time.y - time.x)]
  merged <- merged[order(id.y, abs(diff))]
  merged <- merged[, .(id.x=id.x[1], time.x=time.x[1], diff=diff[1]), .(id.y, time.y, PrimBrandCode)]
  
  # 3.3 For each network ad involved, find the closest match
  merged <- merged[order(id.x, abs(diff))]
  merged <- merged[, .(id.y=id.y[1], time.y=time.y[1], diff=diff[1]), .(id.x, time.x, PrimBrandCode)]
  
  # 3.5 Extract
  moved_around <- merged[abs(diff) <= MOVED_AROUND_CLEARANCE]
  s3 <- s2[!(id %in% c(moved_around$id.x, moved_around$id.y))]
  
  cat("Done\n")
  
  # ---------------------------------------- #
  # 4. Replacement by Spot
  # ---------------------------------------- #
  
  cat("Finding spot replacements...")
  
  # s3[, c("time_from_last", "time_to_next") := NULL]
  s3[, `:=`(forward=0, backward=0)]
  s3[, last_end_to_this_start := as.integer(time - shift(time,1,type="lag") - shift(Duration,1,type="lag"))]
  s3[, this_end_to_next_start := as.integer(shift(time,1,type="lead") - time - Duration)]
  
  # (1) The previous other ad runs into the network ad's territory
  s3[type=="Network" & 
       shift(type,1,type="lag")=="Other" & 
       last_end_to_this_start < 0, backward := 1]
  # (2) The next other ad covers the schedule time of the network ad
  s3[type=="Network" &
       shift(type,1,type="lead")=="Other" &
       this_end_to_next_start < 0, forward := 1]
  
  # Remove the overran networks
  spot_replaced <- s3[forward==1 | backward==1, .(id, time, type, PrimBrandCode, Duration, forward, backward)]
  s4 <- s3[type=="Clear" | (type=="Network" & forward==0 & backward==0)]
  
  cat("Done\n")
  
  # ---------------------------------------- #
  # 5. Replacement by other network clearance
  # ---------------------------------------- #
  
  cat("Finding clearance replacements...")
  
  s4[, `:=`(forward=0, backward=0)]
  s4[, last_end_to_this_start := as.integer(time - shift(time,1,type="lag") - shift(Duration,1,type="lag"))]
  s4[, this_end_to_next_start := as.integer(shift(time,1,type="lead") - time - Duration)]
  
  s4[type=="Network" & 
       shift(type,1,type="lag")=="Clear" & 
       last_end_to_this_start < 0, backward := 1]
  s4[type=="Clear" & shift(backward,1,type="lead")==1, backward := 1]
  
  s4[type=="Network" & 
       shift(type,1,type="lead")=="Clear" & 
       this_end_to_next_start < 0, forward := 1]
  s4[type=="Clear" & shift(forward,1,type="lag")==1, forward := 1]
  
  s4[, time_from_last := as.integer(time - shift(time,1,type="lag"))]
  s4[, time_to_next := as.integer(shift(time,1,type="lead")-time)]
  
  s4[type=="Network" & forward==1 & backward==1 & time_from_last < time_to_next, forward := 0]
  s4[type=="Network" & forward==1 & backward==1 & time_from_last >= time_to_next, backward := 0]
  s4[type=="Clear" & forward==1 & backward==1 & time_from_last >= time_to_next, forward := 0]
  s4[type=="Clear" & forward==1 & backward==1 & time_from_last < time_to_next, backward := 0]
  
  s4[type=="Network" & backward==1 & shift(backward,1,type="lag")!=1, backward := 0]
  s4[type=="Network" & forward==1 & shift(forward,1,type="lead")!=1, forward := 0]
  s4[type=="Clear" & backward==1 & shift(backward,1,type="lead")!=1, backward := 0]
  s4[type=="Clear" & forward==1 & shift(forward,1,type="lag")!=1, forward := 0]
  
  # We finally get missings and ghosts
  clear_replaced <- s4[forward==1 | backward==1, .(id, time, type, PrimBrandCode, Duration, forward, backward)]
  
  cat("Done\n")
  
  # ---------------------------------------- #
  # 6. Final and Diagnostics
  # ---------------------------------------- #
  final <- s4[forward==0 & backward==0, .(id, time, type, PrimBrandCode, Duration)]
  d1 <- clear[, .(Duration=sum(Duration), type="Clear"), .(AdDate=date(time))]
  d2 <- other[, .(Duration=sum(Duration), type="Other"), .(AdDate=date(time))]
  d3 <- final[type=="Network", .(Duration=sum(Duration), type="Missing"), .(AdDate=date(time))]
  d4 <- final[type=="Clear", .(Duration=sum(Duration), type="Ghost"), .(AdDate=date(time))]
  diag <- rbind(d1,d2,d3,d4,fill=TRUE)
  diag <- dcast(diag, AdDate~type, value.var="Duration")
  diag[is.na(diag)] <- 0
  print(diag[, quantile(Clear+Other)/3600])
  print(diag[, quantile(Clear+Other+Missing)/3600])
  print(diag[, quantile(Clear+Other+Missing-Ghost)/3600])
  
  # ---------------------------------------- #
  # 7. Merge Back Brands
  # ---------------------------------------- #
  missing_network <- final[type=="Network", .(time, PrimBrandCode, Duration)]
  missing_network <- merge(missing_network, nat_brands, by=c("time", "PrimBrandCode"), all.x=T)
  ghost_clearance <- final[type=="Clear", .(time, PrimBrandCode, Duration)]
  ghost_clearance <- merge(ghost_clearance, clear_brands, by=c("time", "PrimBrandCode"), all.x=T)
  
  output <- list(missing_network=missing_network, ghost_clearance=ghost_clearance, diagnostics=diag)
  
  cat("----Finishing station", STATION, "in network", NETWORK, "in time zone", TIMEZONE, "\n")
  return(output)
}

Process.With.Broadcast.Delay <- function(STATION, NETWORK, TIMEZONE, start_date=date("2010-01-01"), end_date=date("2018-12-25")){
  cat("----Processing station", STATION, "in network", NETWORK, "in time zone", TIMEZONE, "\n")
  
  # ---------------------------------------- #
  # 0. Configure Parameters
  # ---------------------------------------- #
  date_range <- c(start_date, end_date)
  if(TIMEZONE == "US/Mountain"){
    DELAY_LIST <- 0:2
    DELAY_ORDER <- c(1,0,2)
    MOVED_AROUND_CLEARANCE <- 21900
  } else if(TIMEZONE == "US/Pacific"){
    DELAY_LIST <- 0:3
    DELAY_ORDER <- c(3,2,0,1)
    MOVED_AROUND_CLEARANCE <- 25500
  } else{
    stop("Process.With.Broadcast.Delay only accepts mountain and pacific time")
  }
  
  # ---------------------------------------- #
  # 1. Load Station-Level Data
  # ---------------------------------------- #
  cat("Loading data...")
  
  # 1.1 National
  # Files are in Eastern time. First convert using eastern time, then change to market time
  nat <- readRDS(paste0(national_dir, NETWORK, ".rds"))
  nat[, time := ymd_hms(paste(as.character(AdDate), AdTime), tz="US/Eastern")]
  nat[, time:=with_tz(time, tz=TIMEZONE)]
  nat_brands <- nat[!is.na(time), .(original_time=time, PrimBrandCode, ScndBrandCode, TerBrandCode)]
  nat <- nat[!is.na(time), .(time, type="Network", PrimBrandCode, Duration)]
  
  # 1.2 Network Clearance
  # Files are in local time
  clear <- readRDS(paste0(clearance_dir, STATION, ".rds"))
  clear[, time := ymd_hms(paste(as.character(AdDate), AdTime), tz=TIMEZONE)]
  clear_brands <- clear[!is.na(time), .(original_time=time, PrimBrandCode, ScndBrandCode, TerBrandCode)]
  clear <- clear[!is.na(time), .(time, type="Clear", PrimBrandCode, Duration)]
  
  # 1.3 Spot (Can be empty)
  spot_file <- paste0(spot_dir, STATION, ".rds")
  if(!file.exists(spot_file)){
    spot <- NULL
  } else{
    spot <- readRDS(spot_file)
    spot[, time := ymd_hms(paste(as.character(AdDate), AdTime), tz=TIMEZONE)]
    spot <- spot[!is.na(time), .(time, type="Other", PrimBrandCode, Duration)]
  }
  
  # 1.4 Syndicated (Can be empty)
  synd_file <- paste0(synd_dir, STATION, ".rds")
  if(!file.exists(synd_file)){
    synd <- NULL
  } else{
    synd <- readRDS(synd_file)
    synd[, time := ymd_hms(paste(as.character(AdDate), AdTime), tz=TIMEZONE)]
    synd <- synd[!is.na(time), .(time, type="Other", PrimBrandCode, Duration)]
  }
  
  # 1.5 Follow date range of clearance
  clear_range <- range(date(clear$time))
  start <- max(clear_range[1], date_range[1])
  end <- min(clear_range[2], date_range[2])
  use_range <- c(start, end)
  
  nat <- nat[date(time) %between% use_range]
  clear <- clear[date(time) %between% use_range]
  if(!is.null(spot)){
    spot <- spot[date(time) %between% use_range]
  }
  if(!is.null(synd)){
    synd <- synd[date(time) %between% use_range]
  }
  
  setorder(nat, time)
  setorder(clear, time)
  nat[, id := 1:.N]
  clear[, id := nrow(nat) + (1:.N)]
  
  if(is.null(spot) & is.null(synd)){
    other <- data.table()
  } else{
    other <- rbind(spot, synd, fill=TRUE)
    setorder(other, time)
    other[, id := nrow(nat) + nrow(clear) + (1:.N)]
  }
  
  cat("Done\n")
  
  # ---------------------------------------- #
  # 2. Find Direct Matches
  # ---------------------------------------- #
  cat("Finding direct matches...")
  
  stacked <- rbind(nat, clear, other, fill=TRUE)
  stacked[, original_time := time]
  
  direct_matches <- vector("list", length(DELAY_LIST))
  
  for(i in 1:length(DELAY_LIST)){
    # Initialize
    delay <- DELAY_LIST[i]
    stacked[type=="Network", time := original_time + delay*3600]
    setkey(stacked, time, type)
    suppressWarnings(stacked[, c("forward", "backward", "time_from_last", "time_to_next") := NULL])
    stacked[, `:=`(forward=0, backward=0)]
    stacked[, time_from_last := as.integer(time - shift(time,1,type="lag"))]
    stacked[, time_to_next := as.integer(shift(time,1,type="lead")-time)]
    
    # Forward Matches
    stacked[type=="Network" & 
              shift(type,1,type="lead")=="Clear" & 
              shift(PrimBrandCode,1,type="lead")==PrimBrandCode, forward := 1]
    stacked[type=="Clear" & shift(forward,1,type="lag")==1, forward := 1]
    
    # Backward Matches
    stacked[type=="Network" & 
              shift(type,1,type="lag")=="Clear" & 
              shift(PrimBrandCode,1,type="lag")==PrimBrandCode, backward := 1]
    stacked[type=="Clear" & shift(backward,1,type="lead")==1, backward := 1]
    
    # Resolve Conflicts
    stacked[type=="Network" & forward==1 & backward==1 & time_from_last < time_to_next, forward := 0]
    stacked[type=="Network" & forward==1 & backward==1 & time_from_last >= time_to_next, backward := 0]
    stacked[type=="Clear" & forward==1 & backward==1 & time_from_last >= time_to_next, forward := 0]
    stacked[type=="Clear" & forward==1 & backward==1 & time_from_last < time_to_next, backward := 0]
    stacked[type=="Network" & backward==1 & shift(backward,1,type="lag")!=1, backward := 0]
    stacked[type=="Network" & forward==1 & shift(forward,1,type="lead")!=1, forward := 0]
    stacked[type=="Clear" & backward==1 & shift(backward,1,type="lead")!=1, backward := 0]
    stacked[type=="Clear" & forward==1 & shift(forward,1,type="lag")!=1, forward := 0]
    
    # Find Matches
    matched <- stacked[forward==1 | backward==1]
    setorder(matched, time, type, PrimBrandCode)
    
    ## Extract Information: Forward
    matched_forward <- matched[forward==1, .(id, original_time, type, PrimBrandCode, Duration, time_from_last, time_to_next)]
    matched_forward[type=="Network", timediff := time_to_next]
    matched_forward[type=="Clear", timediff := time_from_last]
    matched_forward[, c("time_to_next", "time_from_last") := NULL]
    matched_forward[, timediff := as.integer(timediff)]
    matched_forward[, match_id := as.integer(1:.N)]
    matched_forward[match_id %% 2 == 0, match_id := match_id - 1L]
    matched_forward[, match_id := as.integer((match_id+1L)/2L)]
    net_forward <- matched_forward[type=="Network", -c("type"), with=F]
    clear_forward <- matched_forward[type=="Clear", -c("type"), with=F]
    matched_forward <- merge(net_forward, clear_forward, by=c("match_id", "PrimBrandCode", "timediff"))
    
    ## Extract Information: Backward
    matched_backward <- matched[backward==1, .(id, original_time, type, PrimBrandCode, Duration, time_from_last, time_to_next)]
    matched_backward[type=="Network", timediff := time_from_last]
    matched_backward[type=="Clear", timediff := time_to_next]
    matched_backward[, c("time_to_next", "time_from_last") := NULL]
    matched_backward[, timediff := as.integer(timediff)]
    matched_backward[, match_id := as.integer(1:.N)]
    matched_backward[match_id %% 2 == 0, match_id := match_id - 1L]
    matched_backward[, match_id := as.integer((match_id+1L)/2L)]
    net_backward <- matched_backward[type=="Network", -c("type"), with=F]
    clear_backward <- matched_backward[type=="Clear", -c("type"), with=F]
    matched_backward <- merge(net_backward, clear_backward, by=c("match_id", "PrimBrandCode", "timediff"))
    
    ## Combine Matches
    matched_forward[, `:=`(delay_level=delay, match_type="Forward", match_id=NULL)]
    matched_backward[, `:=`(delay_level=delay, match_type="Backward", match_id=NULL)]
    matched_all <- rbind(matched_forward, matched_backward, fill=TRUE)
    
    direct_matches[[i]] <- matched_all
  }
  direct_matches <- rbindlist(direct_matches, fill=T)
  
  # Matches with no conflict
  direct_matches[, occ.x := .N, id.x]
  direct_matches[, occ.y := .N, id.y]
  good <- direct_matches[occ.x==1 & occ.y==1]
  problems <- direct_matches[occ.x>1 | occ.y>1]
  good[, c("occ.x", "occ.y") := NULL]
  problems[, c("occ.x", "occ.y") := NULL]
  
  ## For each network, find closest clearance
  ## If time diff same, follow precendence
  problems[, delay_precedence := factor(delay_level, levels=DELAY_ORDER)]
  problems[, min_timediff.x := min(timediff), id.x]
  problems <- problems[timediff == min_timediff.x]
  problems[, r := rank(delay_precedence), id.x]
  problems <- problems[r==1, -c("r", "min_timediff.x")]
  ## For each clearance, find closest network
  ## If time diff same, follow precendence
  problems[, min_timediff.y := min(timediff), id.y]
  problems <- problems[timediff == min_timediff.y]
  problems[, r := rank(delay_precedence), id.y]
  problems <- problems[r==1, -c("r", "min_timediff.y")]
  
  # Combine
  problems[, delay_precedence := NULL]
  direct_matches <- rbind(problems, good, fill=T)
  
  cat("Done\n")
  
  # ---------------------------------------- #
  # 3. Find "Moved Around" matches
  # ---------------------------------------- #
  cat("Finding moved around matches...")
  
  # Find Leftovers
  stacked <- rbind(nat, clear, other, fill=TRUE)
  stacked[, original_time := time]
  s2 <- stacked[!(id %in% c(direct_matches$id.x, direct_matches$id.y))]
  
  # 3.1 Match the leftover network brands with the leftover clearance brands
  un_net <- s2[type=="Network", .(id, original_time, PrimBrandCode, Duration)]
  un_clear <- s2[type=="Clear", .(id, original_time, PrimBrandCode, Duration)]
  merged <- merge(un_net, un_clear, by="PrimBrandCode", allow.cartesian=T)
  
  # 3.2 For each unexplained CLEARANCE ad, find the closest network ad
  merged[, diff := as.integer(original_time.y - original_time.x)]
  merged <- merged[order(id.y, abs(diff))]
  merged <- merged[, .(id.x=id.x[1], original_time.x=original_time.x[1], diff=diff[1]), .(id.y, original_time.y, PrimBrandCode)]
  
  # 3.3 For each network ad involved, find the closest match
  merged <- merged[order(id.x, abs(diff))]
  merged <- merged[, .(id.y=id.y[1], original_time.y=original_time.y[1], diff=diff[1]), .(id.x, original_time.x, PrimBrandCode)]
  
  # 3.4 Limit and Extract
  moved_around <- merged[abs(diff) <= MOVED_AROUND_CLEARANCE]
  
  cat("Done\n")
  
  # ---------------------------------------- #
  # 4. Replacement by Spot
  # ---------------------------------------- #
  # Rule: if a network ad is overran or replaced by spot/synd at ANY delay levels, then it is "explained missing"
  # This is of course conservative!!
  
  cat("Finding spot replacements...")
  
  # Find Leftovers
  s2[, time := NULL]
  s3 <- s2[!(id %in% c(moved_around$id.x, moved_around$id.y))]
  
  # Spot Replacement
  spot_replaced <- vector("list", length(DELAY_LIST))
  for(i in 1:length(DELAY_LIST)){
    delay <- DELAY_LIST[i]
    s3[, time := original_time]
    s3[type=="Network", time := original_time + delay*3600]
    setkey(s3, time, type)
    suppressWarnings(s3[, c("forward", "backward", "last_end_to_this_start", "this_end_to_next_start") := NULL])
    s3[, `:=`(forward=0, backward=0)]
    s3[, last_end_to_this_start := as.integer(time - shift(time,1,type="lag") - shift(Duration,1,type="lag"))]
    s3[, this_end_to_next_start := as.integer(shift(time,1,type="lead") - time - Duration)]
    
    # (1) The previous other ad runs into the network ad's territory
    s3[type=="Network" & 
         shift(type,1,type="lag")=="Other" & 
         last_end_to_this_start < 0, backward := 1]
    # (2) The next other ad covers the schedule time of the network ad
    s3[type=="Network" &
         shift(type,1,type="lead")=="Other" &
         this_end_to_next_start < 0, forward := 1]
    
    out <- s3[forward==1 | backward==1]
    out <- out[, .(id, original_time, PrimBrandCode, Duration, delay_level=delay, forward, backward, last_end_to_this_start, this_end_to_next_start)]
    spot_replaced[[i]] <- out
  }
  spot_replaced <- rbindlist(spot_replaced, fill=T)
  
  cat("Done\n")
  
  # ---------------------------------------- #
  # 5. Replacement by other network clearance
  # ---------------------------------------- #
  
  cat("Finding clearance replacements...")
  
  s3 <- s2[!(id %in% c(moved_around$id.x, moved_around$id.y))]
  s4 <- s3[type=="Clear" | (type=="Network" & !(id %in% unique(spot_replaced$id)))]
  
  clear_replace_master <- vector("list", length(DELAY_LIST))
  for(i in 1:length(DELAY_LIST)){
    delay <- DELAY_LIST[i]
    s4[, time := original_time]
    s4[type=="Network", time := original_time + delay*3600]
    setkey(s4, time, type)
    suppressWarnings(s4[, c("forward", "backward", "time_from_last", "time_to_next", "last_end_to_this_start", "this_end_to_next_start") := NULL])
    s4[, `:=`(forward=0, backward=0)]
    s4[, time_from_last := as.integer(time - shift(time,1,type="lag"))]
    s4[, time_to_next := as.integer(shift(time,1,type="lead")-time)]
    s4[, last_end_to_this_start := as.integer(time - shift(time,1,type="lag") - shift(Duration,1,type="lag"))]
    s4[, this_end_to_next_start := as.integer(shift(time,1,type="lead") - time - Duration)]
    
    # Forward and backward overrun
    s4[type=="Network" & 
         shift(type,1,type="lag")=="Clear" & 
         last_end_to_this_start < 0, backward := 1]
    s4[type=="Clear" & shift(backward,1,type="lead")==1, backward := 1]
    s4[type=="Network" & 
         shift(type,1,type="lead")=="Clear" & 
         this_end_to_next_start < 0, forward := 1]
    s4[type=="Clear" & shift(forward,1,type="lag")==1, forward := 1]
    
    # Resolve Conflict
    s4[type=="Network" & forward==1 & backward==1 & time_from_last < time_to_next, forward := 0]
    s4[type=="Network" & forward==1 & backward==1 & time_from_last >= time_to_next, backward := 0]
    s4[type=="Clear" & forward==1 & backward==1 & time_from_last >= time_to_next, forward := 0]
    s4[type=="Clear" & forward==1 & backward==1 & time_from_last < time_to_next, backward := 0]
    s4[type=="Network" & backward==1 & shift(backward,1,type="lag")!=1, backward := 0]
    s4[type=="Network" & forward==1 & shift(forward,1,type="lead")!=1, forward := 0]
    s4[type=="Clear" & backward==1 & shift(backward,1,type="lead")!=1, backward := 0]
    s4[type=="Clear" & forward==1 & shift(forward,1,type="lag")!=1, forward := 0]
    
    # Find Matches
    clear_replaced <- s4[forward==1 | backward==1]
    setorder(clear_replaced, time, type, PrimBrandCode)
    
    ## Extract Information: Forward
    replaced_forward <- clear_replaced[forward==1, .(id, original_time, type, PrimBrandCode, Duration, time_from_last, time_to_next)]
    replaced_forward[type=="Network", timediff := time_to_next]
    replaced_forward[type=="Clear", timediff := time_from_last]
    replaced_forward[, c("time_to_next", "time_from_last") := NULL]
    replaced_forward[, timediff := as.integer(timediff)]
    replaced_forward[, match_id := as.integer(1:.N)]
    replaced_forward[match_id %% 2 == 0, match_id := match_id - 1L]
    replaced_forward[, match_id := as.integer((match_id+1L)/2L)]
    net_forward <- replaced_forward[type=="Network", -c("type"), with=F]
    clear_forward <- replaced_forward[type=="Clear", -c("type"), with=F]
    replaced_forward <- merge(net_forward, clear_forward, by=c("match_id", "timediff"))
    
    ## Extract Information: Backward
    replaced_backward <- clear_replaced[backward==1, .(id, original_time, type, PrimBrandCode, Duration, time_from_last, time_to_next)]
    replaced_backward[type=="Network", timediff := time_from_last]
    replaced_backward[type=="Clear", timediff := time_to_next]
    replaced_backward[, c("time_to_next", "time_from_last") := NULL]
    replaced_backward[, timediff := as.integer(timediff)]
    replaced_backward[, match_id := as.integer(1:.N)]
    replaced_backward[match_id %% 2 == 0, match_id := match_id - 1L]
    replaced_backward[, match_id := as.integer((match_id+1L)/2L)]
    net_backward <- replaced_backward[type=="Network", -c("type"), with=F]
    clear_backward <- replaced_backward[type=="Clear", -c("type"), with=F]
    replaced_backward <- merge(net_backward, clear_backward, by=c("match_id", "timediff"))
    
    ## Combine Matches
    replaced_forward[, `:=`(delay_level=delay, match_type="Forward", match_id=NULL)]
    replaced_backward[, `:=`(delay_level=delay, match_type="Backward", match_id=NULL)]
    replaced_all <- rbind(replaced_forward, replaced_backward, fill=TRUE)
    
    clear_replace_master[[i]] <- replaced_all
  }
  clear_replace_master <- rbindlist(clear_replace_master, fill=T)
  
  # Matches with no conflict
  clear_replace_master[, occ.x := .N, id.x]
  clear_replace_master[, occ.y := .N, id.y]
  good <- clear_replace_master[occ.x==1 & occ.y==1]
  problems <- clear_replace_master[occ.x>1 | occ.y>1]
  good[, c("occ.x", "occ.y") := NULL]
  problems[, c("occ.x", "occ.y") := NULL]
  
  ## For each network, find closest clearance
  ## If time diff same, follow precendence
  problems[, delay_precedence := factor(delay_level, levels=DELAY_ORDER)]
  problems[, min_timediff.x := min(timediff), id.x]
  problems <- problems[timediff == min_timediff.x]
  problems[, r := rank(delay_precedence), id.x]
  problems <- problems[r==1, -c("r", "min_timediff.x")]
  ## For each clearance, find closest network
  ## If time diff same, follow precendence
  problems[, min_timediff.y := min(timediff), id.y]
  problems <- problems[timediff == min_timediff.y]
  problems[, r := rank(delay_precedence), id.y]
  problems <- problems[r==1, -c("r", "min_timediff.y")]
  
  # Combine
  problems[, delay_precedence := NULL]
  clear_replaced <- rbind(problems, good, fill=T)
  
  cat("Done\n")
  
  # ---------------------------------------- #
  # 6. 2am to 5am
  # Conservative rule here AGAIN!
  # ---------------------------------------- #
  s4 <- s3[type=="Clear" | (type=="Network" & !(id %in% unique(spot_replaced$id)))]
  s5 <- s4[!(id %in% c(clear_replaced$id.x, clear_replaced$id.y))]
  s5[, d0 := hour(original_time) %in% 2:4]
  s5[, d1 := hour(original_time+3600) %in% 2:4]
  s5[, d2 := hour(original_time+7200) %in% 2:4]
  if(TIMEZONE == "US/Pacific"){
    s5[, d3 := hour(original_time+10800) %in% 2:4]
    s5[, nr := d0 | d1 | d2 | d3]
  } else{
    s5[, nr := d0 | d1 | d2]
  }
  not_recorded <- s5[nr==T & type=="Network", .(type, PrimBrandCode, Duration, id, original_time)]
  final <- s5[type=="Clear" | (type=="Network" & nr==F), .(type, PrimBrandCode, Duration, id, original_time)]
  
  # ---------------------------------------- #
  # 7. Diagnostics
  # ---------------------------------------- #
  d1 <- clear[, .(Duration=sum(Duration), type="Clear"), .(AdDate=date(time))]
  d2 <- other[, .(Duration=sum(Duration), type="Other"), .(AdDate=date(time))]
  d3 <- final[type=="Network", .(Duration=sum(Duration), type="Missing"), .(AdDate=date(original_time))]
  d4 <- final[type=="Clear", .(Duration=sum(Duration), type="Ghost"), .(AdDate=date(original_time))]
  diag <- rbind(d1,d2,d3,d4,fill=TRUE)
  diag <- dcast(diag, AdDate~type, value.var="Duration")
  diag[is.na(diag)] <- 0
  print(diag[, quantile(Clear+Other)/3600])
  print(diag[, quantile(Clear+Other+Missing)/3600])
  print(diag[, quantile(Clear+Other+Missing-Ghost)/3600])
  
  # ---------------------------------------- #
  # 8. Merge Back Brands
  # ---------------------------------------- #
  missing_network <- final[type=="Network", .(original_time, PrimBrandCode, Duration)]
  missing_network <- merge(missing_network, nat_brands, by=c("original_time", "PrimBrandCode"), all.x=T)
  ghost_clearance <- final[type=="Clear", .(original_time, PrimBrandCode, Duration)]
  ghost_clearance <- merge(ghost_clearance, clear_brands, by=c("original_time", "PrimBrandCode"), all.x=T)
  
  setnames(missing_network, "original_time", "time")
  setnames(ghost_clearance, "original_time", "time")
  
  output <- list(missing_network=missing_network, ghost_clearance=ghost_clearance, diagnostics=diag)
  
  cat("----Finishing station", STATION, "in network", NETWORK, "in time zone", TIMEZONE, "\n")
  return(output)
}

national_dir <- "/project/ghitsch/data/Ad-Intel-2020/Data-By-Station/National/"
clearance_dir <- "/project/ghitsch/data/Ad-Intel-2020/Data-By-Station/Clearance/"
spot_dir <- "/project/ghitsch/data/Ad-Intel-2020/Data-By-Station/Spot/"
synd_dir <- "/project/ghitsch/data/Ad-Intel-2020/Data-By-Station/Syndicated/"

station_info_path <- "/project/ghitsch/data/Ad-Intel-2020/Auxiliary-Data/Stations-For-Missing-Network-Imputation.csv"
stations <- fread(station_info_path)
setorder(stations, group)
stations[, `:=`(start_date=date(start_date), end_date=date(end_date))]
stations[start_date < date("2010-01-01"), start_date := date("2010-01-01")]
stations[end_date > date("2018-12-30"), end_date := date("2018-12-30")]

output_dir <- "/project/ghitsch/data/Ad-Intel-2020/Data-By-Station/Missing-Network-And-Ghost-Clearance/"
dir.create(output_dir, showWarnings=FALSE)

# Process Wrapper
Process.Wrapper <- function(case, skip_existing=TRUE){
  ## Find Information
  dis_code <- stations[case, DistributorCode]
  dis_id <- stations[case, DistributorID]
  network <- stations[case, Network]
  market <- stations[case, MarketCode]
  timezone <- stations[case, time_zone]
  start <- stations[case, start_date]
  end <- stations[case, end_date]
  
  outfile <- paste0(output_dir, dis_code, "-", network, ".RData")
  if(skip_existing & file.exists(outfile)){
    return("Skipped")
  }
  
  ## Process
  if(timezone %in% c("US/Eastern", "US/Central")){
    proc <- Process.No.Broadcast.Delay(dis_code, network, timezone, start, end)
  } else{
    proc <- Process.With.Broadcast.Delay(dis_code, network, timezone, start, end)
  }
  
  ## Tidy up
  missing_network <- copy(proc$missing_network)
  ghost_clearance <- copy(proc$ghost_clearance)
  diagnostics <- copy(proc$diagnostics)
  
  ## Add Information
  missing_network[, `:=`(DistributorCode=dis_code, DistributorID=dis_id, MarketCode=market,
                         time_zone=timezone, network=network)]
  ghost_clearance[, `:=`(DistributorCode=dis_code, DistributorID=dis_id, MarketCode=market,
                         time_zone=timezone, network=network)]
  diagnostics[, `:=`(DistributorCode=dis_code, MarketCode=market,
                     time_zone=timezone, network=network)]
  
  ## Save File
  save(missing_network, ghost_clearance, diagnostics, file=outfile)
  return("Success")
}

cl <- makeCluster(2, type="FORK", outfile="")
registerDoParallel(cl)

ncases <- nrow(stations)
status <- foreach(i=1:ncases) %dopar% {
  x <- "Failed"
  try(x <- Process.Wrapper(i))
  x
}

stopCluster(cl)

# Test Loading and stack all diagnostics
diagnostics_master <- vector("list", nrow(stations))
for(i in 1:nrow(stations)){
  dis_code <- stations[i, DistributorCode]
  network <- stations[i, Network]
  outfile <- paste0(output_dir, dis_code, "-", network, ".RData")
  cat("------", i, dis_code, network, "\n")
  load(outfile)
  print(dim(missing_network))
  print(dim(ghost_clearance))
  print(dim(diagnostics))
  diagnostics_master[[i]] <- diagnostics
  rm(missing_network, ghost_clearance, diagnostics)
}
diagnostics_master <- rbindlist(diagnostics_master, fill=T)
fwrite(diagnostics_master, file="~/Missing-Network-Diagnostics.csv")
