
library(data.table)
library(bit64)
library(foreach)
library(doParallel)
library(lubridate)
library(chron)

YEAR_START <- 2010L
YEAR_END <- 2014L

Init.Impute.Table <- function(){
  impute_table <- data.table(use.month = c(1,3,4,6,8,9,10,12),
                             month1 = c(2,2,2,5,7,7,7,11),
                             month2 = c(11,5,5,7,11,11,11,2),
                             weight1 = c(2/3,2/3,1/3,1/2,3/4,1/2,1/4,2/3),
                             weight2 = c(1/3,1/3,2/3,1/2,1/4,1/2,3/4,1/3))
  impute_table <- impute_table[rep(1:.N, each=YEAR_END-YEAR_START+1)]
  impute_table[, use.year := rep(YEAR_START:YEAR_END, 8)]
  impute_table[, year1 := use.year]
  impute_table[, year2 := use.year]
  impute_table[use.month==1, year2 := use.year - 1L]
  impute_table[use.month==12, year2 := use.year + 1L]
  impute_table[year2 == YEAR_START - 1L, `:=`(year2=NA, weight1=1, weight2=0, month2=NA)]
  impute_table[year2 == YEAR_END + 1L, `:=`(year2=NA, weight1=1, weight2=0, month2=NA)]
  return(impute_table)
}


Read.Impressions <- function(year, month, type){
  if(!(type %in% c("National", "Spot", "SpotRadio"))){
    error <- paste0("In function Read.Impressions: unknown input type ", type,
                    "\nTypes allowed are: National, Spot, SpotRadio")
    stop(error)
  }
  
  dir <- paste0(source_dir, "/",year, "/Impressions")
  
  imp <- fread(paste0(dir, '/Imp', type, "TV.tsv"), showProgress=F)
  
  if(type == "National"){
    imp.p <- imp[ImpressionType == "P"]
    imp.t <- imp[ImpressionType == "T"]
    imp.t <- imp.t[as.numeric(substr(ImpressionDate,6,7)) == month]
    
    imp <- rbind(imp.p, imp.t)
    rm(imp.p, imp.t)
  }
  if(type == "Spot"){
    imp <- imp[as.numeric(substr(PeriodYearMonth,5,6)) == month]
  }
  
  remove.cols = grep("male|Male|ww|children",names(imp))
  imp[, (remove.cols) := NULL]
  
  setnames(imp, 'TV_HH', 'imp_total')
  
  if(type == 'National'){
    # For spanish stuff, use hispanic flag = "Y"; otherwise, use "N"
    imp <- imp[DataStreamID==2]
    imp1 <- imp[MediaTypeID %in% c(4,22) & HispanicFlag=="Y"]
    imp2 <- imp[MediaTypeID %in% 1:3]
    imp <- rbind(imp1, imp2, fill=TRUE)
    imp[, HispanicFlag := NULL]
  }
  if(type == 'Spot'){
    imp <- imp[HispanicFlag == "N"]
    imp[, c('MediaTypeID', 'HispanicFlag', 'UC_dim_Bridge_occ_ImpSpotTV_key') := NULL]
  }
  
  suppressWarnings(imp[, c('DataStreamID', 'ActualFlag', 'PeriodYearMonth') := NULL])
  return(imp)
}


Read.Universe.Estimates <- function(type){
  if(!(type %in% c("National", "Spot", "SpotRadio"))){
    error <- paste0("In function Read.Impressions: unknown input type ", type,
                    "\nTypes allowed are: National, Spot, SpotRadio")
    stop(error)
  }
  
  dir <- paste0(source_dir, "/2018/Universe_Estimates")
  
  ue <- fread(paste0(dir, '/UE', type, "TV.tsv"), showProgress=F)
  
  ue[, `:=`(StartDate=as.Date(StartDate),
            EndDate=as.Date(EndDate))]
  
  remove.cols = grep("male|Male|ww|children",names(ue))
  ue[, (remove.cols):=NULL]
  ue <- ue[HispanicFlag == 'N']
  ue[, HispanicFlag := NULL]
  setnames(ue, 'TV_HH', 'ue_total')
  return(ue)
}

source_dir <- "/kilts/adintel"
output_dir <- "/project/ghitsch/data/Ad-Intel-2020/Occurrence-Level/Missing-Network-Full/"
missing_network_dir <- "/project/ghitsch/data/Ad-Intel-2020/Occurrence-Level/Missing-Network-Occurrences/"

station_info_path <- "/project/ghitsch/data/Ad-Intel-2020/Dynamic-References/Local-Distributors-Stacked.csv"
stations <- fread(station_info_path)
setnames(stations, c("year", "month"), c("Year", "Month"))
impute_table <- Init.Impute.Table()

Build.Missing.Network.Occurrence.Level <- function(year, month){
  cat(year, month, 'Missing Network\n')
  
  ## Read Occurrences
  occ <- readRDS(paste0(missing_network_dir, year, "-", month, ".rds"))
  
  ## 965 is a ghost time interval for whatever reason...
  ## We just change it to 960 for convenience
  occ[TimeIntervalNumber==965, TimeIntervalNumber := 960]
  
  ## Read current period impressions
  imp <- Read.Impressions(year, month, "Spot")
  
  ## Impute Rule:
  ## If current period impression covers the distributor, then use current
  ## If current period impression does not cover the distributor, then impute
  if(month %in% impute_table$use.month){
    
    occ_current <- occ[DistributorID %in% unique(imp$DistributorID)]
    occ_impute <- occ[!(DistributorID %in% unique(imp$DistributorID))]
    
    ## Merge Current
    occimp_current <- merge(occ_current, imp, 
                            by=c('DistributorID', 'DayOfWeek', 'TimeIntervalNumber'), all.x=T)
    
    ## Impute
    if(nrow(occ_impute) != 0){
      # (2) Impute
      impute_entry <- impute_table[use.year==year & use.month==month]
      ## 2.1 Month 1
      impute_imp1 <- Read.Impressions(impute_entry$year1, impute_entry$month1, 'Spot')
      setnames(impute_imp1, 'imp_total', 'imp_total1')
      occimp_impute1 <- merge(occ_impute, impute_imp1,
                              by=c('DistributorID', 'DayOfWeek', "TimeIntervalNumber"), all.x=T)
      ## 2.2 Month 2
      if(impute_entry$weight2 > 0){
        impute_imp2 <- Read.Impressions(impute_entry$year2, impute_entry$month2, 'Spot')
        setnames(impute_imp2, 'imp_total', 'imp_total2')
        occimp_impute <- merge(occimp_impute1, impute_imp2,
                               by=c('DistributorID', 'DayOfWeek', "TimeIntervalNumber"), all.x=T)
        occimp_impute[, imp_total := imp_total1 * impute_entry$weight1 + imp_total2 * impute_entry$weight2]
        occimp_impute[, c('imp_total1', 'imp_total2') := NULL]
      } else{
        occimp_impute <- occimp_impute1
        setnames(occimp_impute, 'imp_total1', 'imp_total')
      }
      # (3) Combine
      occimp <- rbind(occimp_current, occimp_impute, fill=TRUE)
    } else{
      occimp <- occimp_current
    }
  } else{ # If sweep month, just merge everything
    occimp <- merge(occ, imp, 
                    by=c('DistributorID', 'DayOfWeek', 'TimeIntervalNumber'), all.x=T)
  }
  
  ## Merge in station information
  curr_stations <- stations[Year==year & Month==month]
  curr_stations <- curr_stations[, .(DistributorID, DistributorCode, MarketCode, Affiliation)]
  occimp <- merge(occimp, curr_stations, by = c("DistributorID"), allow.cartesian = TRUE)
  occimp[, c("DistributorID", "DayOfWeek", "TimeIntervalNumber") := NULL]
  
  ## Merge Universe
  
  ue <- Read.Universe.Estimates("Spot")
  
  # Compute the start date and the end date of the month
  ym_start <- as.Date(paste0(year,"-",month, "-01"))
  
  if(month != 12){
    ym_end <- as.Date(paste0(year,"-",month+1, "-01")) - 1
  } else{
    ym_end <- as.Date(paste0(year,"-12-31"))
  }
  
  ym_start <- as.Date(cut(ym_start, "week"))
  
  if(wday(ym_end) != 1){
    ym_end <- as.Date(cut(ym_end, "week")) - 1
  } 
  
  # Filter out related UEs
  ue <- ue[(StartDate <= ym_start & EndDate >= ym_start) | (StartDate <= ym_end & EndDate >= ym_end)]
  
  # (4) Merge UE
  
  if(ue[,.N] != length(unique(ue[,MarketCode]))){
    
    ue1 <- ue[,.SD[1], by = MarketCode]
    ue2 <- ue[,.SD[.N], by = MarketCode]
    
    setkey(ue1, MarketCode)
    setkey(ue2, MarketCode)
    setkey(occimp, MarketCode)
    
    occimpue1 <- ue1[occimp, roll=TRUE]
    occimpue2 <- ue2[occimp, roll=TRUE]
    
    occimpue1 <- occimpue1[StartDate <= AdDate & EndDate >= AdDate]
    occimpue2 <- occimpue2[StartDate <= AdDate & EndDate >= AdDate]
    
    occimpue <- rbindlist(list(occimpue1, occimpue2))
    
    occimpue <- occimpue[order(MarketCode)]
    
  } else{
    
    setkey(ue, MarketCode)
    setkey(occimp, MarketCode)
    occimpue <- ue[occimp, roll=TRUE]
    
  }
  
  ## Re-Arrange
  setnames(occimpue, c("imp_total", "ue_total"), c("Impression", "Universe"))
  cols_to_keep <- c("AdDate", "AdTime", 
                    "MarketCode", "DistributorCode", "Affiliation", "type", "delay_weight",
                    "PrimBrandCode", "ScndBrandCode", "TerBrandCode",
                    "Duration", "Impression", "Universe")
  occimpue <- occimpue[, cols_to_keep, with=FALSE]

  return(occimpue)
}

Build.And.Save <- function(year, month){
  cat(year, month, "\n")
  
  out <- Build.Missing.Network.Occurrence.Level(year, month)
  outfile <- paste0(output_dir, year, "-", month, ".rds")
  saveRDS(out, file=outfile)
  
  rm(out, outfile)
  
  return(0)
}

years <- rep(YEAR_START:YEAR_END, each=12)
months <- rep(1:12, YEAR_END-YEAR_START+1)

n <- length(years)

cl <- makeCluster(2, type="FORK", outfile=" ")
registerDoParallel(cl)

foreach(i=1:n) %do%{
  x <- "Failed"
  try(x <- Build.And.Save(years[i], months[i]))
  x
}

stopCluster(cl)
