rm(list=ls())
library(data.table)
library(bit64)
library(Hmisc)
library(lubridate)
setNumericRounding(0)
options(scipen=999)

source_dir <- '/grpshares/ghitsch/data/RMS-Build-2016/RMS-Processed/Modules/'
output_dir <- '/grpshares/hitsch_shapiro_ads/data/Brand-Level-Data/RMS/'

home_dir <- "/grpshares/hitsch_shapiro_ads/Advertising-Generalizability/"
universe <- fread(paste0(home_dir, "2-String-Matching/Exported-Files/Universe-For-Estimation-180508.csv"))
upc_table <- readRDS(paste0(home_dir, "2-String-Matching/Exported-Files/Consolidated-UPC-Dictionary-For-Aggregation-180508.rds"))

upc_table <- upc_table[revenue_RMS > 0]
upc_table[, c("size1_units", "revenue_RMS") := NULL]

RMS.Brand.Aggregation <- function(module, brand, promotion_threshold=0.05){
  cat("Aggregating brand", brand, "in module", module, "\n")

  # Get List of UPC
  upc_list <- upc_table[new_module_code==module & new_brand_code_uc==brand]

  cat("\tStacking individual UPCs...")
  # Load Individual UPC
  infiles <- upc_list[, paste0(source_dir, product_module_code, "/", as.numeric(upc), ".RData")]
  infiles <- infiles[!is.na(infiles)]
  versions <- upc_list$upc_ver_uc_corrected
  modules <- upc_list$product_module_code
  versions <- versions[file.exists(infiles)]
  modules <- modules[file.exists(infiles)]
  infiles <- infiles[file.exists(infiles)]
  nfiles <- length(infiles)
  load.upc <- function(i){
    if(i %% 50 == 0) { cat("\t\t", i, "/", nfiles, "\n") }
    load(infiles[i])
    move <- move[processed==T & year(week_end) >= 2010 & upc_ver_uc_corrected==versions[i],
           .(upc, upc_ver_uc_corrected,
             store_code_uc, week_end, 
             base_price, imputed_price, units,
             feature, display, product_module_code=modules[i])]
    return(move)
  }
  main <- rbindlist(lapply(1:length(infiles), load.upc))

  # Load Separate UPCs
  misc_modules <- unique(upc_list$product_module_code)
  misc <- data.table()
  for(m in misc_modules){
    # print(m)
    infile <- paste0(output_dir, "Extract-From-All-Other/", m, ".rds")
    move <- readRDS(infile)
    relevant <- unique(upc_list[product_module_code==m, .(upc, upc_ver_uc_corrected)])
    move <- merge(move, relevant, by=c("upc", "upc_ver_uc_corrected"))
    move[, product_module_code := m]
    misc <- rbindlist(list(misc, move))
  }

  # Stack and house-keeping
  stacked <- unique(rbind(main, misc, fill=T))
  rm(move, main, misc); gc()
  cat("Done\n")

  cat("\tFilling NA prices...")
  # Fill NA Prices
  setkey(stacked, product_module_code, upc, upc_ver_uc_corrected, store_code_uc, week_end)
  bps <- stacked[, .(product_module_code, upc, upc_ver_uc_corrected, store_code_uc, week_end, base_price)]
  str <- bps[, .(product_module_code, upc, upc_ver_uc_corrected, store_code_uc, week_end)]
  bps <- bps[!is.na(base_price)]
  setkey(str, product_module_code, upc, upc_ver_uc_corrected, store_code_uc, week_end)
  setkey(bps, product_module_code, upc, upc_ver_uc_corrected, store_code_uc, week_end)
  bps <- bps[str, roll=TRUE, rollends=c(TRUE, TRUE)]
  if(nrow(bps) != nrow(stacked)){ stop("ERROR: BPS N-ROWS NOT MATCHED") }
  stacked[, base_price := bps$base_price]
  stacked[is.na(imputed_price), imputed_price:=base_price]
  stacked <- stacked[!is.na(base_price)]
  rm(str, bps); gc()
  cat('Done\n')

  cat("\tPreparing for aggregation...")
  # Prepare data for aggregation
  ## Merge in units
  units_list <- unique(upc_list[, .(product_module_code, upc, upc_ver_uc_corrected, size1_amount, multi)])
  stacked <- merge(stacked, units_list, by=c("product_module_code", "upc", "upc_ver_uc_corrected"))
  ## Quantity Aggregation
  stacked[, volume:=size1_amount*multi]
  stacked[, quantity:=units*volume]
  stacked[, c("size1_amount","multi"):=NULL]
  ## Calculate Weight
  stacked[, revenue := imputed_price*units]
  stacked[, weight  := sum(revenue, na.rm=T)/.N, .(upc, upc_ver_uc_corrected, store_code_uc)]
  stacked[, revenue := NULL]
  ## Promotion Indicator
  stacked[, promotion:=(base_price-imputed_price)/base_price > promotion_threshold]
  cat("Done\n")

  # Aggregate at store-week level
      # Sum of quantity
      # Weighted mean equivalent price
      # Weighted mean equivalent base price
      # Weighted mean promotion indicator
      # Weighted mean feature indicator (remove NA)
      # Weighted mean display indicator (remove NA)
  ngroups <- uniqueN(stacked[, .(store_code_uc, week_end)])
  cat("\tAggregating", ngroups, "groups...\n")
  aggregated <- stacked[, { if(.GRP%%100000==0){cat("\t\t", .GRP, "/", ngroups, "\n")}
                              list(units=sum(quantity, na.rm=T),
                              price=wtd.mean(imputed_price/volume, weight, na.rm=T),
                              base_price=wtd.mean(base_price/volume, weight, na.rm=T),
                              promo=wtd.mean(promotion, weight, na.rm=T),
                              feature=wtd.mean(feature, weight, na.rm=T),
                              display=wtd.mean(display, weight, na.rm=T))
                            }, .(store_code_uc, week_end)]
  cat("Done\n")
  cat("Returning brand", brand, "in module", module, "\n")
  return(aggregated)
}

case_table <- unique(upc_table[, .(new_module_code, new_brand_code_uc)])
ncases <- nrow(case_table)
set.seed(as.integer(now()))
case_table <- case_table[sample(ncases, ncases)]

for(i in 1:ncases){
  module <- case_table[i, new_module_code]
  brand <- case_table[i, new_brand_code_uc]
  dir.create(paste0(output_dir, module), showWarnings=FALSE)
  outfile <- paste0(output_dir, module, "/", brand, ".rds")
  if(file.exists(outfile)){
    cat("SKIPPING brand", brand, "in module", module, "\n")
    next
  } else{
    cat("PROCESSING brand", brand, "in module", module, "\n")
    placeholder <- NULL
    saveRDS(placeholder, outfile)
  }
  aggregated <- NULL
  try(aggregated <- RMS.Brand.Aggregation(module, brand))
  if(is.null(aggregated)){
    print("Remove failed file")
    file.remove(outfile)
  } else{
    saveRDS(aggregated, outfile)
  } 
}
