# ------------------------------------ #
# FIXED SPECIFICATIONS
# Data Level: Market (90% cumulative revenue)
# Regression Strategy: Border
# Feature-Display: No
# Advertising: Own 1+3, Own 2+4, Top Comp 1-4
# ------------------------------------ #
# VARIABLE SPECIFICATIONS
# Price-Promotion Control: {None, 0, 1, 2, 3} Competitors
# Delta: seq(0,1,0.01)
# Time Fixed Effect: Quarter / Month
# Advertising measure: GRP / Duration
# Goodwill: log-sum or sum-log
# ------------------------------------ #
# VARIABLE SPECIFICATIONS COUNT: 5*101*2*2*2 = 4040
# ------------------------------------ #

rm(list=ls())
library(data.table)
library(bit64)
library(lfe)
library(RcppRoll)
library(lubridate)
library(Hmisc)

options(lfe.eps = 1e-4)
options(lfe.threads = 1)

home_dir <- "/grpshares/hitsch_shapiro_ads/Advertising-Generalizability/"
ad_dir <- "/grpshares/hitsch_shapiro_ads/data/Brand-Level-Data/AdIntel/"
rms_dir <- "/grpshares/hitsch_shapiro_ads/data/Brand-Level-Data/RMS/"

output_dir <- "/grpshares/hitsch_shapiro_ads/data/Regression-Results/Market-Border-No-Feature-Display/"
dir.create(output_dir, showWarnings=FALSE)

universe <- fread(paste0(home_dir, "2-String-Matching/Exported-Files/Universe-For-Estimation-180508.csv"))
universe <- universe[, .(new_module_code, new_brand_code_uc, adcomp_code, comp1_code, comp2_code, comp3_code)]

screening <- readRDS("/grpshares/hitsch_shapiro_ads/data/Descriptive-Statistics/Screening.rds")
screening <- screening[fracpos > 0.05 & cmean > 10]
universe <- merge(universe, screening[, .(new_module_code, new_brand_code_uc)], by=c("new_module_code", "new_brand_code_uc"))

load(paste0(home_dir, "4-Store-Panelist-Selection/Processed-Data/Selected-Stores.RData"))
border.stores <- border.stores[cumshare<=0.9, .(store_code_uc, dma_code, border_id, total_revenue)]

N_LAGS <- 52
DELTA_LIST <- seq(0,1,0.01)
SPEC_INFO <- data.table(spec=1:8,
						goodwill=rep(c("logsum", "logsum", "sumlog", "sumlog"), 2),
						measure=rep(c("GRP", "Duration"), 4),
						time=rep(c("Quarter", "Month"), each=4))

# ------------------------------------ #
# 1. Process RMS Data
# ------------------------------------ #
Load.RMS.Data <- function(module, brand){
	cat("Loading RMS Data...")

	# 1.1: Own Brand
	infile <- paste0(rms_dir, module, "/", brand, ".rds")
	move <- readRDS(infile)
	move[, base_price := NULL]
	move <- merge(move, border.stores, by="store_code_uc", allow.cartesian=TRUE)
	## Aggregate to market-border
	move <- move[, .(units=sum(units, na.rm=T), 
					 price=wtd.mean(price, total_revenue, na.rm=T),
					 promo=wtd.mean(promo, total_revenue, na.rm=T)), .(dma_code, border_id, week_end)]
	## Remove one-sided borders
	bad_borders <- move[, uniqueN(dma_code), border_id][V1<2, border_id]
	move <- move[!(border_id %in% bad_borders)]
	move[, brand_code := 0]

	# 1.2: Top Competitor
	comp1 <- universe[new_module_code==module & new_brand_code_uc==brand, comp1_code]
	if(is.na(comp1)){
		move1 <- data.table()
	} else{
		infile <- paste0(rms_dir, module, "/", comp1, ".rds")
		move1 <- readRDS(infile)
		move1[, base_price := NULL]
		move1 <- merge(move1, border.stores, by="store_code_uc", allow.cartesian=TRUE)
		## Aggregate to market-border
		move1 <- move1[, .(units=sum(units, na.rm=T), 
						 price=wtd.mean(price, total_revenue, na.rm=T),
						 promo=wtd.mean(promo, total_revenue, na.rm=T)), .(dma_code, border_id, week_end)]
		## Remove one-sided borders
		bad_borders <- move1[, uniqueN(dma_code), border_id][V1<2, border_id]
		move1 <- move1[!(border_id %in% bad_borders)]
		move1[, brand_code := 1]
	}

	# 1.3: Second Competitor
	comp2 <- universe[new_module_code==module & new_brand_code_uc==brand, comp2_code]
	if(is.na(comp2)){
		move2 <- data.table()
	} else{
		infile <- paste0(rms_dir, module, "/", comp2, ".rds")
		move2 <- readRDS(infile)
		move2[, base_price := NULL]
		move2 <- merge(move2, border.stores, by="store_code_uc", allow.cartesian=TRUE)
		## Aggregate to market-border
		move2 <- move2[, .(units=sum(units, na.rm=T), 
						 price=wtd.mean(price, total_revenue, na.rm=T),
						 promo=wtd.mean(promo, total_revenue, na.rm=T)), .(dma_code, border_id, week_end)]
		## Remove one-sided borders
		bad_borders <- move2[, uniqueN(dma_code), border_id][V1<2, border_id]
		move2 <- move2[!(border_id %in% bad_borders)]
		move2[, brand_code := 2]
	}

	# 1.4: Third Competitor
	comp3 <- universe[new_module_code==module & new_brand_code_uc==brand, comp3_code]
	if(is.na(comp3)){
		move3 <- data.table()
	} else{
		infile <- paste0(rms_dir, module, "/", comp3, ".rds")
		move3 <- readRDS(infile)
		move3[, base_price := NULL]
		move3 <- merge(move3, border.stores, by="store_code_uc", allow.cartesian=TRUE)
		## Aggregate to market-border
		move3 <- move3[, .(units=sum(units, na.rm=T), 
						 price=wtd.mean(price, total_revenue, na.rm=T),
						 promo=wtd.mean(promo, total_revenue, na.rm=T)), .(dma_code, border_id, week_end)]
		## Remove one-sided borders
		bad_borders <- move3[, uniqueN(dma_code), border_id][V1<2, border_id]
		move3 <- move3[!(border_id %in% bad_borders)]
		move3[, brand_code := 3]
	}

	# 1.5: Combine and dcast
	move_combined <- rbindlist(list(move, move1, move2, move3))
	move_combined <- dcast(move_combined, dma_code+border_id+week_end~brand_code,
						   value.var=c("units", "price", "promo"))
	cols_to_remove <- c(paste0("units_", 1:3))
	suppressWarnings(move_combined[, (cols_to_remove) := NULL])
	move_combined <- move_combined[!is.na(units_0)]

	cat("Done\n")
	return(move_combined)
}

# ------------------------------------ #
# 2. Process Advertising Data
# ------------------------------------ #
Load.AdIntel.Data <- function(module, brand){
	cat("Loading Ad-Intel Data...")

	# 2.1 Own Brand
	infile <- paste0(ad_dir, module, "-", brand, ".rds")
	ad <- readRDS(infile)
	ad <- ad[Class != "Ghost Clearance"]
	ad[match_tier %in% c(1,3), match_tier := 13]
	ad[match_tier %in% c(2,4), match_tier := 24]
	ad[, Duration := Duration / 60]
	ad <- ad[, .(Duration=sum(Duration, na.rm=T), GRP=sum(GRP, na.rm=T)), .(dma_code, week_end, match_tier)]
	ad <- ad[year(week_end) <= 2014 & dma_code %in% unique(border.stores$dma_code)]
	ad <- dcast(ad, dma_code+week_end~match_tier, value.var=c("GRP", "Duration"))
	str <- data.table(expand.grid(dma_code = unique(border.stores$dma_code),
								  week_end = seq(date("2010-01-09"), date("2014-12-27"), 7)))
	ad <- merge(str, ad, by=names(str), all.x=T)
	ad[is.na(ad)] <- 0
	if(!("GRP_13" %in% names(ad))){
		ad[, `:=`(GRP_13=0.0, Duration_13=0.0)]
	}
	if(!("GRP_24" %in% names(ad))){
		ad[, `:=`(GRP_24=0.0, Duration_24=0.0)]
	}

	# 2.2 Advertising Competitor
	adcomp <- universe[new_module_code==module & new_brand_code_uc==brand, adcomp_code]
	infile <- paste0(ad_dir, module, "-", adcomp, ".rds")
	if(!file.exists(infile)){
		ad1 <- NULL
	} else{
		ad1 <- readRDS(infile)
		ad1 <- ad1[Class != "Ghost Clearance"]
		ad1[, Duration := Duration / 60]
		ad1 <- ad1[, .(Duration_Comp=sum(Duration, na.rm=T), GRP_Comp=sum(GRP, na.rm=T)), .(dma_code, week_end)]
		ad1 <- ad1[year(week_end) <= 2014 & dma_code %in% unique(border.stores$dma_code)]
		str <- data.table(expand.grid(dma_code = unique(border.stores$dma_code),
								  	  week_end = seq(date("2010-01-09"), date("2014-12-27"), 7)))
		ad1 <- merge(str, ad1, by=names(str), all.x=T)
		ad1[is.na(ad1)] <- 0
	}

	# 2.3: Combine
	if(is.null(ad1)){
		ad[, `:=`(Duration_Comp=0.0, GRP_Comp=0.0)]
	} else{
		ad <- merge(ad, ad1, by=c("dma_code", "week_end"))
	}

	cat("Done\n")
	return(ad)
}

# ------------------------------------ #
# 3. Calculate Goodwill Given Delta
# ------------------------------------ #
Calculate.Goodwill.Given.Delta <- function(delta, DT){
	cat("Calculating Goodwill...")
	ad <- copy(DT)

	setkey(ad, dma_code, week_end)
	geom_weights <- sort(cumprod(c(1.0, rep(delta, times = N_LAGS))))

	ad[, ls_GRP_13  := log(roll_sum(GRP_13, n = N_LAGS+1, weights = geom_weights,
	                                normalize = FALSE, align = "right",  fill = NA)+1), dma_code]
	ad[, ls_GRP_24  := log(roll_sum(GRP_24, n = N_LAGS+1, weights = geom_weights,
	                                normalize = FALSE, align = "right",  fill = NA)+1), dma_code]
	ad[, ls_GRP_Comp  := log(roll_sum(GRP_Comp, n = N_LAGS+1, weights = geom_weights,
	                                normalize = FALSE, align = "right",  fill = NA)+1), dma_code]
	ad[, ls_Duration_13  := log(roll_sum(Duration_13, n = N_LAGS+1, weights = geom_weights,
	                                normalize = FALSE, align = "right",  fill = NA)+1), dma_code]
	ad[, ls_Duration_24  := log(roll_sum(Duration_24, n = N_LAGS+1, weights = geom_weights,
	                                normalize = FALSE, align = "right",  fill = NA)+1), dma_code]
	ad[, ls_Duration_Comp  := log(roll_sum(Duration_Comp, n = N_LAGS+1, weights = geom_weights,
	                                normalize = FALSE, align = "right",  fill = NA)+1), dma_code]

	ad[, sl_GRP_13  := roll_sum(log(1+GRP_13), n = N_LAGS+1, weights = geom_weights,
	                                normalize = FALSE, align = "right",  fill = NA), dma_code]
	ad[, sl_GRP_24  := roll_sum(log(1+GRP_24), n = N_LAGS+1, weights = geom_weights,
	                                normalize = FALSE, align = "right",  fill = NA), dma_code]
	ad[, sl_GRP_Comp  := roll_sum(log(1+GRP_Comp), n = N_LAGS+1, weights = geom_weights,
	                                normalize = FALSE, align = "right",  fill = NA), dma_code]
	ad[, sl_Duration_13  := roll_sum(log(1+Duration_13), n = N_LAGS+1, weights = geom_weights,
	                                normalize = FALSE, align = "right",  fill = NA), dma_code]
	ad[, sl_Duration_24  := roll_sum(log(1+Duration_24), n = N_LAGS+1, weights = geom_weights,
	                                normalize = FALSE, align = "right",  fill = NA), dma_code]
	ad[, sl_Duration_Comp  := roll_sum(log(1+Duration_Comp), n = N_LAGS+1, weights = geom_weights,
	                                normalize = FALSE, align = "right",  fill = NA), dma_code]

	ad <- ad[complete.cases(ad)]
	ad[, c("GRP_13", "GRP_24", "GRP_Comp", "Duration_13", "Duration_24", "Duration_Comp") := NULL]

	cat("Done\n")
	return(ad)
}

# ------------------------------------ #
# 4. Run Barebone Regressions Given Delta
# NCOMP := -1
# ------------------------------------ #
Run.Barebone.Regressions <- function(delta, move, adintel){
	cat("Running Barebone Regressions for", delta, "\n")

	ad <- Calculate.Goodwill.Given.Delta(delta, adintel)

	regdata <- move[week_end > date("2011-01-01"), .(dma_code, border_id, week_end,
						units_0=log(1+units_0))]
	regdata <- regdata[complete.cases(regdata)]
	bad_borders <- regdata[, uniqueN(dma_code), border_id][V1<2, border_id]
	regdata <- regdata[!(border_id %in% bad_borders)]

	regdata <- merge(regdata, ad, by=c("dma_code", "week_end"), allow.cartesian=TRUE)

	regdata[, Quarter := floor_date(week_end, "quarter")]
	regdata[, Month := floor_date(week_end, "month")]
	regdata[, Border_Quarter := .GRP, .(border_id, Quarter)]
	regdata[, Border_Month := .GRP, .(border_id, Month)]
	regdata[, Border_DMA := .GRP, .(border_id, dma_code)]
	regdata[, WOY := pmin(week(week_end), 52)]

	# Formulas:
	# 	Time Fixed Effect: Quarter / Month
	# 	Advertising measure: GRP / Duration
	# 	Goodwill: log-sum or sum-log
	flist <- c("units_0~ls_GRP_13+ls_GRP_24+ls_GRP_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA", 
			   "units_0~ls_Duration_13+ls_Duration_24+ls_Duration_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA",
			   "units_0~sl_GRP_13+sl_GRP_24+sl_GRP_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA", 
			   "units_0~sl_Duration_13+sl_Duration_24+sl_Duration_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA",
			   "units_0~ls_GRP_13+ls_GRP_24+ls_GRP_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA", 
			   "units_0~ls_Duration_13+ls_Duration_24+ls_Duration_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA",
			   "units_0~sl_GRP_13+sl_GRP_24+sl_GRP_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA", 
			   "units_0~sl_Duration_13+sl_Duration_24+sl_Duration_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA")

	# Things to collect:
	# 	ALL Coefficients
	#	N, MSE, R2, Partial R2
	run_regression <- function(j){
		cat("\tRunning Specification", j, "...")
		form <- flist[j]
		fit <- summary(felm(as.formula(form), regdata, exactDOF=TRUE))
		coef <- coef(fit)
		names <- rownames(coef)
		coef <- as.data.table(coef)[, c(1,2,4), with=F]
		names(coef) <- c("coef", "se", "pval")
		coef[, variable := names]
		coef[, nobs := fit$N]
		coef[, r2 := fit$r2]
		coef[, r2p := as.numeric(fit$P.r.squared)]
		coef[, mse := mean(resid(fit)^2)]
		coef[, spec := j]
		cat("Done\n")
		return(coef)
	}

	results <- rbindlist(lapply(1:length(flist), run_regression), fill=TRUE)	
	results[, decay := delta]
	results[, ncomp := -1]

	cat("Finishing Barebone Regressions for", delta, "\n")
	return(results)
}

# ------------------------------------ #
# 5. Run Own-Brand Regressions Given Delta
# NCOMP := 0
# ------------------------------------ #
Run.C0.Regressions <- function(delta, move, adintel){
	cat("Running C0 Regressions for", delta, "\n")

	ad <- Calculate.Goodwill.Given.Delta(delta, adintel)

	regdata <- move[week_end > date("2011-01-01"), .(dma_code, border_id, week_end,
						units_0=log(1+units_0), price_0=log(price_0), promo_0)]
	regdata <- regdata[complete.cases(regdata)]
	bad_borders <- regdata[, uniqueN(dma_code), border_id][V1<2, border_id]
	regdata <- regdata[!(border_id %in% bad_borders)]

	regdata <- merge(regdata, ad, by=c("dma_code", "week_end"), allow.cartesian=TRUE)

	regdata[, Quarter := floor_date(week_end, "quarter")]
	regdata[, Month := floor_date(week_end, "month")]
	regdata[, Border_Quarter := .GRP, .(border_id, Quarter)]
	regdata[, Border_Month := .GRP, .(border_id, Month)]
	regdata[, Border_DMA := .GRP, .(border_id, dma_code)]
	regdata[, WOY := pmin(week(week_end), 52)]

	# Formulas:
	# 	Time Fixed Effect: Quarter / Month
	# 	Advertising measure: GRP / Duration
	# 	Goodwill: log-sum or sum-log
	flist <- c("units_0~price_0+promo_0+ls_GRP_13+ls_GRP_24+ls_GRP_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+ls_Duration_13+ls_Duration_24+ls_Duration_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA",
			   "units_0~price_0+promo_0+sl_GRP_13+sl_GRP_24+sl_GRP_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+sl_Duration_13+sl_Duration_24+sl_Duration_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA",
			   "units_0~price_0+promo_0+ls_GRP_13+ls_GRP_24+ls_GRP_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+ls_Duration_13+ls_Duration_24+ls_Duration_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA",
			   "units_0~price_0+promo_0+sl_GRP_13+sl_GRP_24+sl_GRP_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+sl_Duration_13+sl_Duration_24+sl_Duration_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA")

	# Things to collect:
	# 	ALL Coefficients
	#	N, MSE, R2, Partial R2
	run_regression <- function(j){
		cat("\tRunning Specification", j, "...")
		form <- flist[j]
		fit <- summary(felm(as.formula(form), regdata, exactDOF=TRUE))
		coef <- coef(fit)
		names <- rownames(coef)
		coef <- as.data.table(coef)[, c(1,2,4), with=F]
		names(coef) <- c("coef", "se", "pval")
		coef[, variable := names]
		coef[, nobs := fit$N]
		coef[, r2 := fit$r2]
		coef[, r2p := as.numeric(fit$P.r.squared)]
		coef[, mse := mean(resid(fit)^2)]
		coef[, spec := j]
		cat("Done\n")
		return(coef)
	}

	results <- rbindlist(lapply(1:length(flist), run_regression), fill=TRUE)	
	results[, decay := delta]
	results[, ncomp := 0]

	cat("Finishing C0 Regressions for", delta, "\n")
	return(results)
}

# ------------------------------------ #
# 6. Run One-Brand Regressions Given Delta
# NCOMP := 1
# ------------------------------------ #
Run.C1.Regressions <- function(delta, move, adintel){
	cat("Running C1 Regressions for", delta, "\n")

	if(!("price_1" %in% names(move))){
		cat("Skipping C1 Regressions for", delta, "\n")
		return(data.table())
	}

	regdata <- move[week_end > date("2011-01-01"), .(dma_code, border_id, week_end,
						units_0=log(1+units_0), price_0=log(price_0), promo_0, 
						price_1=log(price_1), promo_1)]

	regdata <- regdata[complete.cases(regdata)]
	if(nrow(regdata) < 1000){
		cat("Skipping C1 Regressions for", delta, "\n")
		return(data.table())
	}

	bad_borders <- regdata[, uniqueN(dma_code), border_id][V1<2, border_id]
	regdata <- regdata[!(border_id %in% bad_borders)]

	ad <- Calculate.Goodwill.Given.Delta(delta, adintel)
	regdata <- merge(regdata, ad, by=c("dma_code", "week_end"), allow.cartesian=TRUE)

	regdata[, Quarter := floor_date(week_end, "quarter")]
	regdata[, Month := floor_date(week_end, "month")]
	regdata[, Border_Quarter := .GRP, .(border_id, Quarter)]
	regdata[, Border_Month := .GRP, .(border_id, Month)]
	regdata[, Border_DMA := .GRP, .(border_id, dma_code)]
	regdata[, WOY := pmin(week(week_end), 52)]

	# Formulas:
	# 	Time Fixed Effect: Quarter / Month
	# 	Advertising measure: GRP / Duration
	# 	Goodwill: log-sum or sum-log
	flist <- c("units_0~price_0+promo_0+price_1+promo_1+ls_GRP_13+ls_GRP_24+ls_GRP_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+price_1+promo_1+ls_Duration_13+ls_Duration_24+ls_Duration_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA",
			   "units_0~price_0+promo_0+price_1+promo_1+sl_GRP_13+sl_GRP_24+sl_GRP_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+price_1+promo_1+sl_Duration_13+sl_Duration_24+sl_Duration_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA",
			   "units_0~price_0+promo_0+price_1+promo_1+ls_GRP_13+ls_GRP_24+ls_GRP_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+price_1+promo_1+ls_Duration_13+ls_Duration_24+ls_Duration_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA",
			   "units_0~price_0+promo_0+price_1+promo_1+sl_GRP_13+sl_GRP_24+sl_GRP_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+price_1+promo_1+sl_Duration_13+sl_Duration_24+sl_Duration_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA")

	# Things to collect:
	# 	ALL Coefficients
	#	N, MSE, R2, Partial R2
	run_regression <- function(j){
		cat("\tRunning Specification", j, "...")
		form <- flist[j]
		fit <- summary(felm(as.formula(form), regdata, exactDOF=TRUE))
		coef <- coef(fit)
		names <- rownames(coef)
		coef <- as.data.table(coef)[, c(1,2,4), with=F]
		names(coef) <- c("coef", "se", "pval")
		coef[, variable := names]
		coef[, nobs := fit$N]
		coef[, r2 := fit$r2]
		coef[, r2p := as.numeric(fit$P.r.squared)]
		coef[, mse := mean(resid(fit)^2)]
		coef[, spec := j]
		cat("Done\n")
		return(coef)
	}

	results <- rbindlist(lapply(1:length(flist), run_regression), fill=TRUE)	
	results[, decay := delta]
	results[, ncomp := 1]

	cat("Finishing C1 Regressions for", delta, "\n")
	return(results)
}

# ------------------------------------ #
# 7. Run Two-Brand Regressions Given Delta
# NCOMP := 2
# ------------------------------------ #
Run.C2.Regressions <- function(delta, move, adintel){
	cat("Running C2 Regressions for", delta, "\n")

	if(!("price_2" %in% names(move))){
		cat("Skipping C2 Regressions for", delta, "\n")
		return(data.table())
	}

	regdata <- move[week_end > date("2011-01-01"), .(dma_code, border_id, week_end,
						units_0=log(1+units_0), price_0=log(price_0), promo_0, 
						price_1=log(price_1), promo_1, 
						price_2=log(price_2), promo_2)]

	regdata <- regdata[complete.cases(regdata)]
	if(nrow(regdata) < 1000){
		cat("Skipping C2 Regressions for", delta, "\n")
		return(data.table())
	}

	bad_borders <- regdata[, uniqueN(dma_code), border_id][V1<2, border_id]
	regdata <- regdata[!(border_id %in% bad_borders)]

	ad <- Calculate.Goodwill.Given.Delta(delta, adintel)
	regdata <- merge(regdata, ad, by=c("dma_code", "week_end"), allow.cartesian=TRUE)

	regdata[, Quarter := floor_date(week_end, "quarter")]
	regdata[, Month := floor_date(week_end, "month")]
	regdata[, Border_Quarter := .GRP, .(border_id, Quarter)]
	regdata[, Border_Month := .GRP, .(border_id, Month)]
	regdata[, Border_DMA := .GRP, .(border_id, dma_code)]
	regdata[, WOY := pmin(week(week_end), 52)]

	# Formulas:
	# 	Time Fixed Effect: Quarter / Month
	# 	Advertising measure: GRP / Duration
	# 	Goodwill: log-sum or sum-log
	flist <- c("units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+ls_GRP_13+ls_GRP_24+ls_GRP_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+ls_Duration_13+ls_Duration_24+ls_Duration_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA",
			   "units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+sl_GRP_13+sl_GRP_24+sl_GRP_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+sl_Duration_13+sl_Duration_24+sl_Duration_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA",
			   "units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+ls_GRP_13+ls_GRP_24+ls_GRP_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+ls_Duration_13+ls_Duration_24+ls_Duration_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA",
			   "units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+sl_GRP_13+sl_GRP_24+sl_GRP_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+sl_Duration_13+sl_Duration_24+sl_Duration_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA")

	# Things to collect:
	# 	ALL Coefficients
	#	N, MSE, R2, Partial R2
	run_regression <- function(j){
		cat("\tRunning Specification", j, "...")
		form <- flist[j]
		fit <- summary(felm(as.formula(form), regdata, exactDOF=TRUE))
		coef <- coef(fit)
		names <- rownames(coef)
		coef <- as.data.table(coef)[, c(1,2,4), with=F]
		names(coef) <- c("coef", "se", "pval")
		coef[, variable := names]
		coef[, nobs := fit$N]
		coef[, r2 := fit$r2]
		coef[, r2p := as.numeric(fit$P.r.squared)]
		coef[, mse := mean(resid(fit)^2)]
		coef[, spec := j]
		cat("Done\n")
		return(coef)
	}

	results <- rbindlist(lapply(1:length(flist), run_regression), fill=TRUE)	
	results[, decay := delta]
	results[, ncomp := 2]

	cat("Finishing C2 Regressions for", delta, "\n")
	return(results)
}

# ------------------------------------ #
# 8. Run Three-Brand Regressions Given Delta
# NCOMP := 3
# ------------------------------------ #
Run.C3.Regressions <- function(delta, move, adintel){
	cat("Running C3 Regressions for", delta, "\n")

	if(!("price_3" %in% names(move))){
		cat("Skipping C3 Regressions for", delta, "\n")
		return(data.table())
	}

	regdata <- move[week_end > date("2011-01-01"), .(dma_code, border_id, week_end,
						units_0=log(1+units_0), price_0=log(price_0), promo_0, 
						price_1=log(price_1), promo_1, 
						price_2=log(price_2), promo_2, 
						price_3=log(price_3), promo_3)]

	regdata <- regdata[complete.cases(regdata)]
	if(nrow(regdata) < 1000){
		cat("Skipping C3 Regressions for", delta, "\n")
		return(data.table())
	}

	bad_borders <- regdata[, uniqueN(dma_code), border_id][V1<2, border_id]
	regdata <- regdata[!(border_id %in% bad_borders)]

	ad <- Calculate.Goodwill.Given.Delta(delta, adintel)
	regdata <- merge(regdata, ad, by=c("dma_code", "week_end"), allow.cartesian=TRUE)

	regdata[, Quarter := floor_date(week_end, "quarter")]
	regdata[, Month := floor_date(week_end, "month")]
	regdata[, Border_Quarter := .GRP, .(border_id, Quarter)]
	regdata[, Border_Month := .GRP, .(border_id, Month)]
	regdata[, Border_DMA := .GRP, .(border_id, dma_code)]
	regdata[, WOY := pmin(week(week_end), 52)]

	# Formulas:
	# 	Time Fixed Effect: Quarter / Month
	# 	Advertising measure: GRP / Duration
	# 	Goodwill: log-sum or sum-log
	flist <- c("units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+price_3+promo_3+ls_GRP_13+ls_GRP_24+ls_GRP_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+price_3+promo_3+ls_Duration_13+ls_Duration_24+ls_Duration_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA",
			   "units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+price_3+promo_3+sl_GRP_13+sl_GRP_24+sl_GRP_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+price_3+promo_3+sl_Duration_13+sl_Duration_24+sl_Duration_Comp|Border_DMA+Border_Quarter+WOY|0|Border_DMA",
			   "units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+price_3+promo_3+ls_GRP_13+ls_GRP_24+ls_GRP_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+price_3+promo_3+ls_Duration_13+ls_Duration_24+ls_Duration_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA",
			   "units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+price_3+promo_3+sl_GRP_13+sl_GRP_24+sl_GRP_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA", 
			   "units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+price_3+promo_3+sl_Duration_13+sl_Duration_24+sl_Duration_Comp|Border_DMA+Border_Month+WOY|0|Border_DMA")

	# Things to collect:
	# 	ALL Coefficients
	#	N, MSE, R2, Partial R2
	run_regression <- function(j){
		cat("\tRunning Specification", j, "...")
		form <- flist[j]
		fit <- summary(felm(as.formula(form), regdata, exactDOF=TRUE))
		coef <- coef(fit)
		names <- rownames(coef)
		coef <- as.data.table(coef)[, c(1,2,4), with=F]
		names(coef) <- c("coef", "se", "pval")
		coef[, variable := names]
		coef[, nobs := fit$N]
		coef[, r2 := fit$r2]
		coef[, r2p := as.numeric(fit$P.r.squared)]
		coef[, mse := mean(resid(fit)^2)]
		coef[, spec := j]
		cat("Done\n")
		return(coef)
	}

	results <- rbindlist(lapply(1:length(flist), run_regression), fill=TRUE)	
	results[, decay := delta]
	results[, ncomp := 3]

	cat("Finishing C3 Regressions for", delta, "\n")
	return(results)
}

# ------------------------------------ #
# 9. Combined Wrapper
# ------------------------------------ #
Combined.Wrapper <- function(module, brand, delta_list){
	cat("----------COMBINED PROCESSOR STARTS FOR", module, brand, "----------\n")

	# 9.1: Data Creation
	move <- Load.RMS.Data(module, brand)
	adintel <- Load.AdIntel.Data(module, brand)

	# 9.2: Delta Loop
	master <- vector("list", length(delta_list))
	for(i in 1:length(delta_list)){
		delta <- delta_list[i]

		# 9.2.1: Barebone
		tb <- NULL
		try(suppressWarnings(tb <- Run.Barebone.Regressions(delta, move, adintel)))
		if(is.null(tb)){
			cat("FATAL WARNING: Barebone Failure at", module, brand, delta, "\n")
		}

		# 9.2.2: C0
		t0 <- NULL
		try(suppressWarnings(t0 <- Run.C0.Regressions(delta, move, adintel)))
		if(is.null(t0)){
			cat("FATAL WARNING: C0 Failure at", module, brand, delta, "\n")
		}

		# 9.2.3: C1
		t1 <- NULL
		try(suppressWarnings(t1 <- Run.C1.Regressions(delta, move, adintel)))
		if(is.null(t1)){
			cat("FATAL WARNING: C1 Failure at", module, brand, delta, "\n")
		}

		# 9.2.4: C2
		t2 <- NULL
		try(suppressWarnings(t2 <- Run.C2.Regressions(delta, move, adintel)))
		if(is.null(t2)){
			cat("FATAL WARNING: C2 Failure at", module, brand, delta, "\n")
		}

		# 9.2.5: C3
		t3 <- NULL
		try(suppressWarnings(t3 <- Run.C3.Regressions(delta, move, adintel)))
		if(is.null(t3)){
			cat("FATAL WARNING: C3 Failure at", module, brand, delta, "\n")
		}

		# 9.2.6: Combine
		combined_results <- rbindlist(list(tb,t0,t1,t2,t3), fill=TRUE)
		master[[i]] <- combined_results
	}
	master <- rbindlist(master, fill=TRUE)

	# 9.3: Adding necessary information
	master[, new_module_code := module]
	master[, new_brand_code_uc := brand]
	master <- merge(master, SPEC_INFO, by="spec")

	cat("----------COMBINED PROCESSOR ENDS FOR", module, brand, "----------\n")
	return(master)
}

# ------------------------------------ #
# 10. PSEUDO-PARALLELIZATION
# ------------------------------------ #
case_table <- universe[, .(new_module_code, new_brand_code_uc)]
ncases <- nrow(case_table)
seed <- as.integer(as.integer64(as.numeric(Sys.time())*10000000) %% 1000000)
set.seed(seed)
Sys.sleep(runif(1)*10)
case_table <- case_table[sample(ncases, ncases)]

for(i in 1:ncases){
  module <- case_table[i, new_module_code]
  brand <- case_table[i, new_brand_code_uc]
  outfile <- paste0(output_dir, module, "-", brand, ".rds")
  if(file.exists(outfile)){
    cat("SKIPPING brand", brand, "in module", module, "\n")
    next
  } else{
    cat("PROCESSING brand", brand, "in module", module, "\n")
    placeholder <- NULL
    saveRDS(placeholder, outfile)
  }
  aggregated <- NULL
  try(aggregated <- Combined.Wrapper(module, brand, DELTA_LIST))
  if(is.null(aggregated)){
    print("Remove failed file")
    file.remove(outfile)
  } else{
    saveRDS(aggregated, outfile)
  } 
}

