# ------------------------------------ #
# PREFERRED SPECIFICATIONS
# Data Level: Store (90% cumulative revenue)
# Regression Strategy: NAIVE
# Feature-Display: No
# Advertising: Own 1+3, Own 2+4, Top Comp 1-4
# Price-Promotion Control: 20% observation loss
# Delta: 0/0.9
# Time Fixed Effect: NAIVE
# Advertising measure: GRP
# Goodwill: log-sum
# ------------------------------------ #

rm(list=ls())
library(data.table)
library(bit64)
library(lfe)
library(RcppRoll)
library(lubridate)

options(lfe.eps = 1e-4)
options(lfe.threads = 8)

home_dir <- "/grpshares/hitsch_shapiro_ads/Advertising-Generalizability/"
ad_dir <- "/grpshares/hitsch_shapiro_ads/data/Brand-Level-Data/AdIntel/"
rms_dir <- "/grpshares/hitsch_shapiro_ads/data/Brand-Level-Data/RMS/"

output_dir <- "/grpshares/hitsch_shapiro_ads/data/Regression-Preferred/Naive-No-Feature/"
dir.create(output_dir, showWarnings=FALSE)

universe <- fread(paste0(home_dir, "2-String-Matching/Exported-Files/Universe-For-Estimation-180508.csv"))
universe <- universe[, .(new_module_code, new_brand_code_uc, adcomp_code, comp1_code, comp2_code, comp3_code)]

screening <- readRDS("/grpshares/hitsch_shapiro_ads/data/Descriptive-Statistics/Screening.rds")
screening <- screening[fracpos > 0.05 & cmean > 10]
universe <- merge(universe, screening[, .(new_module_code, new_brand_code_uc)], by=c("new_module_code", "new_brand_code_uc"))

load(paste0(home_dir, "4-Store-Panelist-Selection/Processed-Data/Selected-Stores.RData"))
use.stores <- use.stores[cumshare<=0.9, .(store_code_uc, dma_code)]

N_LAGS <- 52
MIN_OBS_RETAIN <- 0.8

# ------------------------------------ #
# 1. Process RMS Data
# ------------------------------------ #
Load.RMS.Data <- function(module, brand){
	cat("Loading RMS Data...")

	# 1.1: Own Brand
	infile <- paste0(rms_dir, module, "/", brand, ".rds")
	move <- readRDS(infile)
	move[, base_price := NULL]
	move <- merge(move, use.stores, by="store_code_uc")
	move[, brand_code := 0]

	# 1.2: Top Competitor
	comp1 <- universe[new_module_code==module & new_brand_code_uc==brand, comp1_code]
	if(is.na(comp1)){
		move1 <- data.table()
	} else{
		infile <- paste0(rms_dir, module, "/", comp1, ".rds")
		move1 <- readRDS(infile)
		move1[, base_price := NULL]
		move1 <- merge(move1, use.stores, by="store_code_uc")
		move1[, brand_code := 1]
	}

	# 1.3: Second Competitor
	comp2 <- universe[new_module_code==module & new_brand_code_uc==brand, comp2_code]
	if(is.na(comp2)){
		move2 <- data.table()
	} else{
		infile <- paste0(rms_dir, module, "/", comp2, ".rds")
		move2 <- readRDS(infile)
		move2[, base_price := NULL]
		move2 <- merge(move2, use.stores, by="store_code_uc")
		move2[, brand_code := 2]
	}

	# 1.4: Third Competitor
	comp3 <- universe[new_module_code==module & new_brand_code_uc==brand, comp3_code]
	if(is.na(comp3)){
		move3 <- data.table()
	} else{
		infile <- paste0(rms_dir, module, "/", comp3, ".rds")
		move3 <- readRDS(infile)
		move3[, base_price := NULL]
		move3 <- merge(move3, use.stores, by="store_code_uc")
		move3[, brand_code := 3]
	}

	# 1.5: Combine and dcast
	move_combined <- rbindlist(list(move, move1, move2, move3))
	move_combined <- dcast(move_combined, store_code_uc+dma_code+week_end~brand_code,
						   value.var=c("units", "price", "promo"))
	cols_to_remove <- c(paste0("units_", 1:3))
	suppressWarnings(move_combined[, (cols_to_remove) := NULL])
	move_combined <- move_combined[!is.na(units_0)]

	cat("Done\n")
	return(move_combined)
}

# ------------------------------------ #
# 2. Process Advertising Data
# ------------------------------------ #
Load.AdIntel.Data <- function(module, brand){
	cat("Loading Ad-Intel Data...")

	# 2.1 Own Brand
	infile <- paste0(ad_dir, module, "-", brand, ".rds")
	ad <- readRDS(infile)
	ad <- ad[Class != "Ghost Clearance"]
	ad[match_tier %in% c(1,3), match_tier := 13]
	ad[match_tier %in% c(2,4), match_tier := 24]
	ad[, Duration := Duration / 60]
	ad <- ad[, .(Duration=sum(Duration, na.rm=T), GRP=sum(GRP, na.rm=T)), .(dma_code, week_end, match_tier)]
	ad <- ad[year(week_end) <= 2014 & dma_code %in% unique(use.stores$dma_code)]
	ad <- dcast(ad, dma_code+week_end~match_tier, value.var=c("GRP", "Duration"))
	str <- data.table(expand.grid(dma_code = unique(use.stores$dma_code),
								  week_end = seq(date("2010-01-09"), date("2014-12-27"), 7)))
	ad <- merge(str, ad, by=names(str), all.x=T)
	ad[is.na(ad)] <- 0
	if(!("GRP_13" %in% names(ad))){
		ad[, `:=`(GRP_13=0.0, Duration_13=0.0)]
	}
	if(!("GRP_24" %in% names(ad))){
		ad[, `:=`(GRP_24=0.0, Duration_24=0.0)]
	}

	# 2.2 Advertising Competitor
	adcomp <- universe[new_module_code==module & new_brand_code_uc==brand, adcomp_code]
	infile <- paste0(ad_dir, module, "-", adcomp, ".rds")
	if(!file.exists(infile)){
		ad1 <- NULL
	} else{
		ad1 <- readRDS(infile)
		ad1 <- ad1[Class != "Ghost Clearance"]
		ad1[, Duration := Duration / 60]
		ad1 <- ad1[, .(Duration_Comp=sum(Duration, na.rm=T), GRP_Comp=sum(GRP, na.rm=T)), .(dma_code, week_end)]
		ad1 <- ad1[year(week_end) <= 2014 & dma_code %in% unique(use.stores$dma_code)]
		str <- data.table(expand.grid(dma_code = unique(use.stores$dma_code),
								  	  week_end = seq(date("2010-01-09"), date("2014-12-27"), 7)))
		ad1 <- merge(str, ad1, by=names(str), all.x=T)
		ad1[is.na(ad1)] <- 0
	}

	# 2.3: Combine
	if(is.null(ad1)){
		ad[, `:=`(Duration_Comp=0.0, GRP_Comp=0.0)]
	} else{
		ad <- merge(ad, ad1, by=c("dma_code", "week_end"))
	}

	cat("Done\n")
	return(ad)
}

# ------------------------------------ #
# 3. Calculate Goodwill Given Delta
# ------------------------------------ #
Calculate.Goodwill.Given.Delta <- function(delta, DT){
	cat("Calculating Goodwill...")
	ad <- copy(DT)

	setkey(ad, dma_code, week_end)
	geom_weights <- sort(cumprod(c(1.0, rep(delta, times = N_LAGS))))

	ad[, ls_GRP_13  := log(roll_sum(GRP_13, n = N_LAGS+1, weights = geom_weights,
	                                normalize = FALSE, align = "right",  fill = NA)+1), dma_code]
	ad[, ls_GRP_24  := log(roll_sum(GRP_24, n = N_LAGS+1, weights = geom_weights,
	                                normalize = FALSE, align = "right",  fill = NA)+1), dma_code]
	ad[, ls_GRP_Comp  := log(roll_sum(GRP_Comp, n = N_LAGS+1, weights = geom_weights,
	                                normalize = FALSE, align = "right",  fill = NA)+1), dma_code]
	ad <- ad[complete.cases(ad)]
	ad[, c("GRP_13", "GRP_24", "GRP_Comp", "Duration_13", "Duration_24", "Duration_Comp") := NULL]

	cat("Done\n")
	return(ad)
}

# ------------------------------------ #
# 4. Observation Calculation
# ------------------------------------ #
Build.Regression.Data <- function(move, ad){
	# 0 Competitor
	regdata <- move[week_end > date("2011-01-01"), .(store_code_uc, dma_code, week_end,
						units_0=log(1+units_0), price_0=log(price_0), promo_0)]
	regdata <- regdata[complete.cases(regdata)]
	regdata <- merge(regdata, ad, by=c("dma_code", "week_end"), allow.cartesian=TRUE)
	d0 <- copy(regdata)
	n0 <- nrow(regdata)

	# 1 Competitor
	if(!("price_1" %in% names(move))){
		n1 <- 0
	} else{
		regdata <- move[week_end > date("2011-01-01"), .(store_code_uc, dma_code, week_end,
							units_0=log(1+units_0), price_0=log(price_0), promo_0, 
							price_1=log(price_1), promo_1)]
		regdata <- regdata[complete.cases(regdata)]
		regdata <- merge(regdata, ad, by=c("dma_code", "week_end"), allow.cartesian=TRUE)
		d1 <- copy(regdata)
		n1 <- nrow(regdata)
	}

	# 2 Competitor
	if(!("price_2" %in% names(move))){
		n2 <- 0
	} else{
		regdata <- move[week_end > date("2011-01-01"), .(store_code_uc, dma_code, week_end,
							units_0=log(1+units_0), price_0=log(price_0), promo_0, 
							price_1=log(price_1), promo_1, 
							price_2=log(price_2), promo_2)]
		regdata <- regdata[complete.cases(regdata)]
		regdata <- merge(regdata, ad, by=c("dma_code", "week_end"), allow.cartesian=TRUE)
		d2 <- copy(regdata)
		n2 <- nrow(regdata)
	}

	# 3 Competitor
	if(!("price_3" %in% names(move))){
		n3 <- 0
	} else{
		regdata <- move[week_end > date("2011-01-01"), .(store_code_uc, dma_code, week_end,
							units_0=log(1+units_0), price_0=log(price_0), promo_0, 
							price_1=log(price_1), promo_1, 
							price_2=log(price_2), promo_2, 
							price_3=log(price_3), promo_3)]
		regdata <- regdata[complete.cases(regdata)]
		regdata <- merge(regdata, ad, by=c("dma_code", "week_end"), allow.cartesian=TRUE)
		d3 <- copy(regdata)
		n3 <- nrow(regdata)
	}

	r1 <- n1 / n0
	r2 <- n2 / n0
	r3 <- n3 / n0

	if(r1 < MIN_OBS_RETAIN){
		return(d0)
	} else{
		if(r2 < MIN_OBS_RETAIN){
			return(d1)
		} else{
			if(r3 < MIN_OBS_RETAIN){
				return(d2)
			} else{
				return(d3)
			}
		}
	}
}

# ------------------------------------ #
# 5. Run Regression
# ------------------------------------ #
Run.Single.Regression <- function(move, adintel, delta){
	ad <- Calculate.Goodwill.Given.Delta(delta, adintel)

	regdata <- Build.Regression.Data(move, ad)
	formula <- "units_0~price_0+promo_0+price_1+promo_1+price_2+promo_2+price_3+promo_3+ls_GRP_13+ls_GRP_24+ls_GRP_Comp|0|0|dma_code"

	nc <- 3
	if(!("price_3" %in% names(regdata))){
		formula <- gsub("+price_3+promo_3", "", formula, fixed=TRUE)
		nc <- 2
	}
	if(!("price_2" %in% names(regdata))){
		formula <- gsub("+price_2+promo_2", "", formula, fixed=TRUE)
		nc <- 1
	}
	if(!("price_1" %in% names(regdata))){
		formula <- gsub("+price_1+promo_1", "", formula, fixed=TRUE)
		nc <- 0
	}
	if(var(regdata$ls_GRP_24) == 0){
		formula <- gsub("+ls_GRP_24", "", formula, fixed=TRUE)
	}
	if(var(regdata$ls_GRP_Comp) == 0){
		formula <- gsub("+ls_GRP_Comp", "", formula, fixed=TRUE)
	}

	regdata[, Month := floor_date(week_end, "month")]
	regdata[, WOY := pmin(week(week_end), 52)]

	fit <- summary(felm(as.formula(formula), regdata, exactDOF=TRUE))
	coef <- coef(fit)
	names <- rownames(coef)
	coef <- as.data.table(coef)[, c(1,2,4), with=F]
	names(coef) <- c("coef", "se", "pval")
	coef[, variable := names]
	coef[, nobs := fit$N]
	coef[, r2 := fit$r2]
	coef[, r2p := as.numeric(fit$P.r.squared)]
	coef[, mse := mean(resid(fit)^2)]

	coef[, decay := delta]
	coef[, ncomp := nc]

	return(coef)
}

# ------------------------------------ #
# 6. Combined Wrapper
# ------------------------------------ #
Combined.Wrapper <- function(module, brand){
	move <- Load.RMS.Data(module, brand)
	adintel <- Load.AdIntel.Data(module, brand)

	short <- Run.Single.Regression(move, adintel, 0)
	long <- Run.Single.Regression(move, adintel, 0.9)
	out <- rbind(short, long, fill=TRUE)
	return(out)
}

# ------------------------------------ #
# 7. PSEUDO-PARALLELIZATION
# ------------------------------------ #
case_table <- universe[, .(new_module_code, new_brand_code_uc)]
ncases <- nrow(case_table)
seed <- as.integer(as.integer64(as.numeric(Sys.time())*10000000) %% 1000000)
set.seed(seed)
Sys.sleep(runif(1)*10)
case_table <- case_table[sample(ncases, ncases)]

for(i in 1:ncases){
  module <- case_table[i, new_module_code]
  brand <- case_table[i, new_brand_code_uc]
  outfile <- paste0(output_dir, module, "-", brand, ".rds")
  if(file.exists(outfile)){
    cat("SKIPPING brand", brand, "in module", module, "\n")
    next
  } else{
    cat("PROCESSING brand", brand, "in module", module, "\n")
    placeholder <- NULL
    saveRDS(placeholder, outfile)
  }
  aggregated <- NULL
  try(aggregated <- Combined.Wrapper(module, brand))
  if(is.null(aggregated)){
    print("Remove failed file")
    file.remove(outfile)
  } else{
    saveRDS(aggregated, outfile)
  } 
}


