# -------------------------------
# Run Esitmation and Elasticities
# 
# -------------------------------
rm(list = ls())
library(data.table)
library(bit64)
library(gridExtra)
library(ggplot2)
library(dplyr)
library(glmnet)
library(splines2)

home_dir <- "/project/hitsch_shapiro_ads/Advertising-Generalizability/"
ad_dir <- "/project/hitsch_shapiro_ads/data/Brand-Level-Data/AdIntel-wo-Anomaly/"
rms_dir <- "/project/hitsch_shapiro_ads/data/Brand-Level-Data/RMS/"
input_dir = "/project/hitsch_shapiro_ads/scripts/Workspace-Ningyin/RNR-edits/Splines/Regular-Store-No-splines-dma_code+week_end-Month-GRP/"
resid_input_dir = paste0(input_dir, "Resid/")
adstk_input_dir = paste0(input_dir, "Adstock/")

knots_quantile = seq(0.1, 0.9, 0.1)

Get.Elasticities <- function(module,
                             brand,
                             lambda = "lambda.min",
                             knots_quantile,
                             splines_degree = 3) {
  set.seed(splines_degree)
  
  resid.DT <- readRDS(paste0(resid_input_dir, module,"-",brand,".rds"))
  newAdstock <- readRDS(paste0(adstk_input_dir, module, "-", brand, ".rds"))
  
  columns <- names(newAdstock)[grepl("spline|sum_GRP_13|ls_GRP_13", names(newAdstock))]
  
  resid.fit <- cv.glmnet(x = as.matrix(resid.DT[, columns, with = F]),
                         y = resid.DT$units_0)
  fit.coef <- coef(resid.fit, s = lambda)
  
  fit.predict <- predict(resid.fit, 
                         as.matrix(newAdstock[, columns, with = F]),
                         s = lambda)
  
  fit.coef.DT <- as.data.table(as.matrix(fit.coef), keep.rownames = T)
  names(fit.coef.DT) <- c("coefficient", "value")
  fit.coef.DT[round(value, 100) != 0, type := "kept"]
  fit.coef.DT[is.na(type), type := "discarded"]
  
  coef_num <- fit.coef.DT[coefficient != "(Intercept)", .N]
  dropped_num <- fit.coef.DT[coefficient != "(Intercept)" & type == "discarded", .N]
  
  fit.coef.1se <- coef(resid.fit, s = "lambda.1se")
  fit.coef.1se.DT <- as.data.table(as.matrix(fit.coef.1se), keep.rownames = T)
  names(fit.coef.1se.DT) <- c("coefficient", "value")
  fit.coef.1se.DT[round(value, 100) != 0, type := "kept"]
  fit.coef.1se.DT[is.na(type), type := "discarded"]
  
  coef_num_1se <- fit.coef.1se.DT[coefficient != "(Intercept)", .N]
  dropped_num_1se <- fit.coef.1se.DT[coefficient != "(Intercept)" & type == "discarded", .N]
  
  # Calculate Elasticities
  Adstock.Elas <- newAdstock$Adstock
  spline_knots_13 <- newAdstock[quantile %in% as.character(knots_quantile),
                                Adstock]
  spline_columns <- columns[grepl("spline", columns)]
  boundary <- range(Adstock.Elas, na.rm = TRUE)
  aKnots <- c(rep(boundary[1L], splines_degree + 1), spline_knots_13, rep(boundary[2L], splines_degree + 1))
  
  dbsOut <- splines::splineDesign(aKnots, x = Adstock.Elas, derivs = 1L)
  
  spline.deriv <- dbsOut[,-1] %*% fit.coef.DT[coefficient %in% spline_columns, value]
  
  deriv <- as.data.table(spline.deriv)
  names(deriv) <- "spline"
  
  deriv[, A := Adstock.Elas]
  
  deriv[, ls_GRP_13 := fit.coef.DT[coefficient == "ls_GRP_13", value] / (1 + A)]
  deriv[, sum_GRP_13 := fit.coef.DT[coefficient == "sum_GRP_13", value]]
  
  poly_DT <- data.table(
    poly_power = c(1/2, 2, 3, 4, 5, 6, 7, 8, 9, 10),
    poly_name  = c("square_root",
                   "squared",
                   "cubed",
                   "power4",
                   "power5",
                   "power6",
                   "power7",
                   "power8",
                   "power9",
                   "power10")
  )
  
  for (i in 1:nrow(poly_DT)) {
    poly_power <- poly_DT[i, poly_power]
    poly_name  <- poly_DT[i, poly_name]
    poly_name  <- paste0("sum_GRP_13_", poly_name)
    
    deriv[, eval(poly_name) := fit.coef.DT[coefficient == poly_name, value] * poly_power * A ^ (poly_power - 1)]
    
    poly_name  <- poly_DT[i, poly_name]
    poly_name  <- paste0("ls_GRP_13_", poly_name)
    
    deriv[, eval(poly_name) := fit.coef.DT[coefficient == poly_name, 
                                       value] * poly_power * log(1+A) ^ (poly_power - 1) / (1 + A)]
    
  }
  
  deriv_sum <- rowSums(deriv[, -c("A")], na.rm = T)
  
  deriv_sum <- deriv$A * deriv_sum
  
  Adstock.Elas.DT <- data.table(
    Quantile = newAdstock$quantile,
    Adstock = Adstock.Elas,
    Elasticity = deriv_sum
  )
  
  Adstock.Elas.DT <- Adstock.Elas.DT[Quantile != "NA"]
  Adstock.Elas.DT[Quantile != "Mean", quantile := as.numeric(Quantile)]
  
  mean.Elas <- Adstock.Elas.DT[Quantile == "Mean", Elasticity]
  median.Elas <- Adstock.Elas.DT[quantile == 0.5, Elasticity]
  
  mean.25.75 <- mean(Adstock.Elas.DT[quantile %between% c(0.25, 0.75), Elasticity],
                     na.rm = T)
  median.25.75 <- median(Adstock.Elas.DT[quantile %between% c(0.25, 0.75), Elasticity],
                         na.rm = T)
  q10.25.75 <- quantile(Adstock.Elas.DT[quantile %between% c(0.25, 0.75), Elasticity],
                         probs = 0.1, na.rm = T)
  q25.25.75 <- quantile(Adstock.Elas.DT[quantile %between% c(0.25, 0.75), Elasticity],
                         probs = 0.25, na.rm = T)
  q75.25.75 <- quantile(Adstock.Elas.DT[quantile %between% c(0.25, 0.75), Elasticity],
                         probs = 0.75, na.rm = T)
  q90.25.75 <- quantile(Adstock.Elas.DT[quantile %between% c(0.25, 0.75), Elasticity],
                         probs = 0.9, na.rm = T)
  
  Elas.DT <- data.table(
    Type = c("Mean", "Median", "Mean-25-75", "Median-25-75", 
             "Q10-25-75", "Q25-25-75", "Q75-25-75", "Q90-25-75"),
    Value = c(mean.Elas, median.Elas, mean.25.75, median.25.75,
              q10.25.75, q25.25.75, q75.25.75, q90.25.75)
  )
  
  Elas.DT[, `:=` (new_module_code = module,
                  new_brand_code_uc = brand,
                  mean_Adstock = Adstock.Elas.DT[Quantile == "Mean", Adstock],
                  median_Adstock = Adstock.Elas.DT[quantile == 0.5, Adstock],
                  q25_Adstock = Adstock.Elas.DT[quantile == 0.25, Adstock],
                  q75_Adstock = Adstock.Elas.DT[quantile == 0.75, Adstock],
                  coef_num    = coef_num,
                  dropped_num = dropped_num,
                  dropped_num_1se = dropped_num_1se)]
  
  return(Elas.DT)
}

universe <- fread(paste0(home_dir, 
                          "2-String-Matching/Exported-Files/Universe-For-Estimation-180508.csv"))
universe <- universe[, .(new_module_code, new_brand_code_uc, 
                          adcomp_code, comp1_code, 
                          comp2_code, comp3_code)]

screening <- readRDS("/project/hitsch_shapiro_ads/data/Descriptive-Statistics/Screening.rds")
screening <- screening[fracpos > 0.05 & cmean > 10]
universe  <- merge(universe, screening[, .(new_module_code, new_brand_code_uc)], by=c("new_module_code", "new_brand_code_uc"))

case_table <- universe[, .(new_module_code, new_brand_code_uc)]
ncases <- nrow(case_table)

Elas.DT <- data.table()
for (i in 1:ncases) {

  cat(i, "\n")
  sub.Elas.DT <- Get.Elasticities(module = case_table[i, new_module_code],
                                  brand  = case_table[i, new_brand_code_uc],
                                  knots_quantile = knots_quantile)
  
  Elas.DT <- rbind(Elas.DT, sub.Elas.DT)

}

saveRDS(Elas.DT,
        file = "./Processed-Data/Elasticities.rds")


