# ---------------------------------------- #
# Process Homescan Data for ROI analyses
#
# ---------------------------------------- #

library(data.table)
library(bit64)
library(lfe)
library(RcppRoll)
library(lubridate)
library(Hmisc)
library(doParallel)
library(pryr)

home_dir <- "/project/hitsch_shapiro_ads/Advertising-Generalizability/"
ad_dir   <- "/project/hitsch_shapiro_ads/data/Brand-Level-Data/AdIntel/"
rms_dir  <- "/project/hitsch_shapiro_ads/data/Brand-Level-Data/RMS/"
hms_dir  <- "/project/hitsch_shapiro_ads/data/HMS/"

# Load universe data
universe  <- fread(paste0(home_dir, 
                          "2-String-Matching/Exported-Files/Universe-For-Estimation-180508.csv"))
universe  <- universe[, .(new_module_code, new_brand_code_uc)]

screening <- readRDS("/project/hitsch_shapiro_ads/data/Descriptive-Statistics/Screening.rds")
screening <- screening[fracpos > 0.05 & cmean > 10]
universe  <- merge(universe, 
                   screening[, .(new_module_code, new_brand_code_uc)], 
                   by=c("new_module_code", "new_brand_code_uc"))

# Load upc dictionary
upc_list <- readRDS(paste0(home_dir,
                           "2-String-Matching/Exported-Files/Consolidated-UPC-Dictionary-For-Aggregation-180508.rds"))

upc_list <- unique(upc_list[, .(new_module_code, new_brand_code_uc, 
                                product_module_code,
                                upc, upc_ver_uc_corrected, 
                                multi, size1_amount)])
upc_list[, `:=`(new_module_code = as.numeric(new_module_code),
                new_brand_code_uc = as.numeric(new_brand_code_uc))]

upc_list <- merge(upc_list,
                  universe,
                  by = c("new_module_code",
                         "new_brand_code_uc"))

map_product_module <- unique(upc_list[, .(new_module_code,
                                          new_brand_code_uc,
                                          product_module_code)])

# Load selected panelists
load(paste0(home_dir, 
            "4-Store-Panelist-Selection/Processed-Data/Selected-Panelists.RData"))

# Load corrected products data
load("/project/hitsch_shapiro_ads/data/RMS/Meta-Data/Products-Corrected.RData")
products <- unique(products[, .(upc, upc_ver_uc, upc_ver_uc_corrected,
                                department_code, department_descr,
                                product_group_code, product_group_descr,
                                product_module_code, product_module_descr)])

WALMART_CODE <- 6920
CUMSHARE_CUTOFF <- 0.9

Process <- function(year) {
  print(year)
  
  print(mem_used())
  
  load(paste0(hms_dir,
              "purchases.RData"))
  
  print(mem_used())
  
  mem_change(purchases_sub <- purchases[panel_year == year])

  rm(purchases)

  purchases <- copy(purchases_sub)
  rm(purchases_sub)
  
  cat("got purchases ready.\n")
  
  load(paste0(hms_dir,
              "panelists.RData"))
  panelists <- panelists[panel_year == year][,.(household_code, 
                                                dma_code, 
                                                projection_factor)]
  
  mem_change(purchases_all <- merge(purchases,
                                    panelists,
                                    by  = "household_code"))
  
  stopifnot(purchases_all[,.N] == purchases[,.N])
  cat("Passed the 1st condition.\n")
  
  remove(purchases)
  
  purchases_all <- purchases_all[household_code %in% use.hh$household_code]
  
  purchases_all <- merge(purchases_all,
                         products,
                         by = c("upc", "upc_ver_uc"),
                         all.x = T)
  
  stopifnot(purchases_all[is.na(upc_ver_uc_corrected),.N] == 0)
  cat("Passed the 2nd condition.\n")
  
  purchases_all <- merge(purchases_all, 
                         upc_list, 
                         by=c("upc", "upc_ver_uc_corrected"))
  
  purchases_all[, volume := quantity * size1_amount * multi]
  purchases_all[, price  := total_price_paid /(size1_amount * multi)]
  purchases_all[, coupon := coupon_value /(size1_amount * multi)]
  purchases_all[, c("deal_flag_uc", "quantity", "total_price_paid", "coupon_value", "multi", "size1_amount") := NULL]
  purchases_all[, week_end := date(ceiling_date(date(purchase_date), 
                                                "week", 
                                                week_start = 7)-1)]
  
  save(purchases_all,
       file = paste0("/project/hitsch_shapiro_ads/scripts/Workspace-Ningyin/",
                     "HMS_",
                     year,
                     ".RData"))
  
  return(purchases_all)
}

# Run
for (y in 2010:2014) {
  hms <- Process(year = y)
  rm(hms)
}

