# ---------------------------------------- #
# ROI Calculation - Cost Calculation
# ---------------------------------------- #

library(data.table)
library(bit64)
library(lfe)
library(RcppRoll)
library(lubridate)
library(Hmisc)
library(doParallel)

home_dir <- "/project/hitsch_shapiro_ads/Advertising-Generalizability/"
ad_dir <- "/project/hitsch_shapiro_ads/data/Brand-Level-Data/AdIntel-wo-Anomaly/"
rms_dir <- "/project/hitsch_shapiro_ads/data/Brand-Level-Data/RMS/"

N_LAGS <- 52
WALMART_CODE <- 6920
CUMSHARE_CUTOFF <- 0.9

Calculate.Cost <- function(module,
                           brand) {
  infile <- paste0(ad_dir, module, "-", brand, ".rds")
  ad <- readRDS(infile)
  
  ad <- ad[Class != "Ghost Clearance"]
  ad[match_tier %in% c(1,3), match_tier := 13]
  ad[match_tier %in% c(2,4), match_tier := 24]
  ad[, Duration := Duration / 60]
    
  # Aggregate from AdIntel Brand level to RMS Brand level
  ad <- ad[, .(Duration   = sum(Duration, na.rm=T),
               GRP        = sum(GRP, na.rm=T),
               Spend      = sum(Spend, na.rm=T),
               Impression = sum(Impression, na.rm=T),
               Universe   = mean(Universe)),
           .(dma_code, week_end, match_tier, Class)]
  
  # Separate national and local data for use
  ad_local    <- ad[!(Class %in% c("Cable",
                                   "National Network",
                                   "National Syndicated"))]
  ad_national <- ad[Class %in% c("Cable",
                                 "National Network",
                                 "National Syndicated")]
  
  # Aggregate across all DMA's, return national level data
  ad_local <- ad_local[, .(Duration   = sum(Duration, na.rm=T),
                           GRP        = sum(GRP, na.rm=T),
                           Spend      = sum(Spend, na.rm=T),
                           Impression = sum(Impression, na.rm=T),
                           Universe   = sum(Universe)),
                       .(week_end, match_tier, Class)]
  
  # A test condition to check whether national level data is actually nationl
  ad_national[, test := uniqueN(.SD),
              .(week_end, match_tier, Class)]
  national_dma_num = length(unique(ad_national[,dma_code]))
  stopifnot(ad_national[test!=national_dma_num, .N]==0)
  
  ad_national[, test := NULL]
  
  ad_national <- unique(ad_national[, dma_code := NULL])
  
  
  # Separate Cable and National Network/Syndicated
  ad_cable          = ad_national[Class == "Cable"]
  ad_national_clear = ad_national[Class != "Cable"]
  
  # Merge national and local level of Network/Syndicated
  ad_national_clear[Class == "National Network",
                    Class := "Network"]
  ad_national_clear[Class == "National Syndicated",
                    Class := "Syndicated"]
  
  ad_local = merge(ad_local,
                   ad_national_clear[, .(Class, week_end, match_tier,
                                         Spend, Impression, Universe)],
                   by = c("Class", "week_end", "match_tier"),
                   all.x = T,
                   suffixes = c(".local", ".national"))
  
  
  ad_local[!is.na(Spend.national), Spend := Spend.national]
  ad_local[is.na(Spend.national),  Spend := Spend.local]
  ad_local[, c("Spend.local", "Spend.national") := NULL]
  
  ad_local[, Impression := Impression.local]
  ad_local[, Universe   := Universe.local]
  
  ad_local[, c("Impression.local", "Universe.local") := NULL]
  
  
  # Combine Cable and other Media Types
  ad <- rbind(ad_cable, ad_local, fill = T)
  
  ad <- ad[year(week_end) <= 2014]
  
  ad <- dcast(ad,
              Class + week_end ~ match_tier,
              value.var=c("GRP", "Duration",
                          "Spend", "Impression", "Universe"))
  
  setkey(ad, week_end, Class)
  
  ad <- ad[, `:=`(Duration      = Duration_13,
                  GRP           = GRP_13,
                  Spend         = Spend_13,
                  Impression    = Impression_13,
                  Universe      = Universe_13)]
  
  universe_nat <- data.table()
  for (year in 2010:2014) {
    sub_universe <- fread(paste0("/kilts/adintel/",
                                 year,
                                 "/Universe_Estimates/UENationalTV.tsv"))
    sub_universe[, `:=`(week_end = ceiling_date(date(StartDate), "week", week_start=7)-1,
                        EndDate   = NULL)]
    remove.cols = grep("male|Male|ww|children",
                       names(sub_universe))
    sub_universe[, (remove.cols):=NULL]
    sub_universe <- sub_universe[HispanicFlag == 'N']
    sub_universe[, HispanicFlag := NULL]
    setnames(sub_universe, 'TV_HH', 'Universe_nat')
    sub_universe[, StartDate := NULL]
    
    universe_nat <- rbind(universe_nat,
                          sub_universe)
  }
  
  universe_nat <- unique(universe_nat)
  
  ad_week <- unique(ad[, .(week_end)])
  
  setkey(universe_nat, week_end)
  setkey(ad, week_end)
  
  ad <- universe_nat[ad, roll=T]
  
  national.cost <- ad[, .(Spend = sum(Spend, na.rm=T),
                          Impression = sum(Impression, na.rm=T),
                          Universe_nat = mean(Universe_nat, na.rm=T)),
                      .(week_end)]
  
  national.cost[, `:=`(Spend_per_GRP = Spend/(100 * Impression / Universe_nat),
                       Spend_per_Imp = Spend/Impression*1000,
                       Spend_per_Aud = Spend/Universe_nat*1000)]
  
  national.cost[Spend_per_GRP==Inf, Spend_per_GRP:=NA]
  national.cost[Spend_per_Imp==Inf, Spend_per_Imp:=NA]
  national.cost[Spend_per_Aud==Inf, Spend_per_Aud:=NA]
  
  return(national.cost)
}

Cost.Calculation <- function(module,
                             brand,
                             cost_range) {
  
  change_week_seq = seq(date("2011-01-08"), 
                        date("2013-12-28"), 
                        7)
  
  full_week_seq   = seq(date("2011-01-08"), 
                        date("2014-12-27"), 
                        7)
  
  cost_DT  <- Calculate.Cost(module,
                             brand)
  
  annual_cost_DT <- cost_DT[, .(Spend = sum(Spend, 
                                            na.rm = T), 
                                Impression = sum(Impression, 
                                                 na.rm = T), 
                                Universe = mean(Universe_nat, 
                                                na.rm = T)),
                            .(year = year(week_end))]
  
  annual_cost_DT[, `:=` (Spend_per_GRP = Spend / (100 * Impression / Universe))]
  
  cat("cost_DT Done \n")
  
  
  Cost_results = data.table()
  
  for (i in 1:length(change_week_seq)) {
    
    cat(i, "\n")
    
    week_t = change_week_seq[i]
    
    # weekly advertising cost
    if (week_t %in% unique(cost_DT[, week_end]) & !is.na(cost_DT[week_end == week_t, Spend_per_GRP])) {
      # when this week has cost in the data
      Week_Cost <- cost_DT[week_end == week_t, Spend_per_GRP]
    } else {
      # when this week doesn't have cost in the data
      # find the closest week
      closest_i <- which.min(abs(unique(cost_DT[, week_end]) - date(week_t)))
      closest_week <- unique(cost_DT[, week_end])[closest_i]
      Week_Cost <- cost_DT[week_end == closest_week, Spend_per_GRP]
    }
    
    # annual advertising cost
    if (year(week_t) %in% unique(cost_DT[, year(week_end)])) {
      # when this year is in the data
      Annual_Cost <- annual_cost_DT[year == year(week_t), Spend_per_GRP]
    } else {
      # find the closest year
      closest_i <- which.min(abs(unique(annual_cost_DT[, year]) - year(week_t)))
      closest_year <- unique(annual_cost_DT[, year])[closest_i]
      Annual_Cost <- annual_cost_DT[year == closest_year, Spend_per_GRP]
    }
    
    if (is.na(Annual_Cost) | Annual_Cost == 0) {
      # If there's no year that has ad cost data
      # Use the average cost for the entire time period
      annual_cost_DT <- annual_cost_DT[, .(Spend = sum(Spend, 
                                                       na.rm = T), 
                                           Impression = sum(Impression, 
                                                            na.rm = T), 
                                           Universe = mean(Universe_nat, 
                                                           na.rm = T))]
      annual_cost_DT[, `:=` (Spend_per_GRP = Spend / (100 * Impression / Universe))]
      
      Annual_Cost <- annual_cost_DT[, Spend_per_GRP]
    }
    
      
    for (cost_control in cost_range) {
      
      week_cost <- Week_Cost * cost_control
      
      annual_cost <- Annual_Cost * cost_control
              
      sub.Cost.results <- data.table(
        Module = module,
        Brand  = brand,
        Cost   = cost_control,
        start_week = week_t,
        Week_Cost = week_cost,
        Annual_Cost = annual_cost
      )
      
      Cost_results = rbind(Cost_results, sub.Cost.results, fill = T)
    }
  }
  return(Cost_results)
}

