# Author: Jihong Song 
# Adapted from Jacob Dorn's original script
# Purpose: Creating functions for generic histograms
# Last Edited: May 2017

library(ggplot2)
library(Hmisc)
library(gridExtra)
library(grid)

settings <- theme_bw() + theme(plot.title=element_text(hjust=0.5, face="bold", size=12),
                               strip.text=element_text(size=10, face="bold"),
                               axis.text=element_text(size=9, face="bold"),
                               axis.title=element_text(size=10, face="bold"),
                               legend.text=element_text(size=9, face="bold"),
                               legend.title=element_text(size=10, face="bold")) 

theme_classic2 <- function(base_size = 12, base_family = ""){
  theme_bw(base_size = base_size, base_family = base_family) %+replace%
    theme(
      panel.border     = element_blank(),
      axis.line        = element_line(colour = "black"),
      panel.grid.major = element_line(),
      panel.grid.major.x = element_line(),
      panel.grid.major.y = element_line(),
      panel.grid.minor = element_line(),
      panel.grid.minor.x = element_line(),
      panel.grid.minor.y = element_line(),
      strip.background = element_rect(colour = "black", size = 0.5),
      legend.key       = element_blank()
    )
}

grid_settings <- theme_classic2() + 
  theme(panel.grid.major = element_line(colour='grey90'), 
        panel.grid.minor = element_line(colour='grey95'),
        # plot.title = element_text(face="bold"),
        axis.text=element_text(face="bold"),
        axis.title=element_text(face="bold"))

grid_settings_bold <- grid_settings + 
  theme(plot.title = element_text(face="bold"))

graph_par <- list()
graph_par$line_col  <- "grey60"
# graph_par$fill_col  <- "lightblue2"
# graph_par$part_col  <- "grey92"
graph_par$vline_col <- "violetred3"
graph_par$x_title_size     <- 10
graph_par$y_title_size     <- 10
graph_par$main_title_size  <- 12
graph_par$axis_size        <- 10
graph_par$y_title_move     <- 1.8
graph_par$stat_line_width  <- 0.5
graph_par$stat_text_height <- 0.95
graph_par$stat_text_xalign <- 0.25 # Determines whether the text is on the left or the right of the vline
graph_par$stat_move        <- 1.02
graph_par$label_size       <- 9
graph_par$round            <- 3


Custom.Hist <- function(x, weight=1, count=FALSE, xtitle="",
                        ytitle="", xlimits=NULL, 
                        main_title=NULL, binwidth, bins, xbreaks, ybreaks, 
                        color_val=NULL, legend_pos="none", legend_name="",
                        vline=TRUE, stat="median", annotate=TRUE, 
                        fill_col='lightblue2', part_col='grey92'){
    # Remove NA
    x <- x[!is.na(x)]
    if(length(weight) > 1){
        weight <- weight[!is.na(weight)]
        if(length(weight) != length(x)){
            stop("In Custom.Hist, Invalid x and weight dimensions")
        }
    }

    # Axis
    if(missing(xbreaks)){
        scalex <- scale_x_continuous(name = xtitle)
    } else{
        xmin <- min(xbreaks)
        xmax <- max(xbreaks)
        if(is.null(xlimits)){
          scalex <- scale_x_continuous(breaks = xbreaks, limits = c(xmin, xmax), name = xtitle)
        } else{
          scalex <- scale_x_continuous(breaks = xbreaks, limits = xlimits, name = xtitle)
        }
        
    }

    if(missing(ybreaks)){
        scaley <- scale_y_continuous(name = ytitle)
    } else{
        ymin <- min(ybreaks)
        ymax <- max(ybreaks)
        scaley <- scale_y_continuous(breaks = ybreaks, limits = c(ymin, ymax), name = ytitle)
    }

    # Feed data
    df <- data.frame(x=x, weight=weight)


    # Full plot
    if(!is.null(color_val)){
        if(length(color_val) == 1){
            df$particular <- as.factor(df$x == color_val)
        } else{
            color_val <- color_val[!is.na(color_val)]
            if(length(color_val) != length(x)){
                print(length(color_val))
                print(length(x))
                stop("In Custom.Hist, Invalid x and color_var dimensions")
            }
            df$particular <- as.factor(color_val)
        }
        
        # Histogram part
        if(!missing(binwidth)){
          if(count){
            hist_part <- geom_histogram(col=graph_par$line_col, binwidth=binwidth, boundary=0)
          } else{
            hist_part <- geom_histogram(aes(y=..count../sum(..count..)), col=graph_par$line_col, binwidth=binwidth, boundary=0)
          }
        } else{
          if(count){
            hist_part <- geom_histogram(col=graph_par$line_col, bins=bins, boundary=0)
          } else{
            hist_part <- geom_histogram(aes(y=..count../sum(..count..)), col=graph_par$line_col, bins=bins, boundary=0)
          }
        }
        plt <- ggplot(df, aes(x, weight=weight, fill=particular)) + 
            hist_part + grid_settings + scalex + scaley +
            scale_fill_manual(name=legend_name, values = c(fill_col, part_col)) + 
            theme(legend.position = legend_pos)
    } else{
        # Histogram part
        if(!missing(binwidth)){
          if(count){
            hist_part <- geom_histogram(fill=fill_col, col=graph_par$line_col, binwidth=binwidth, boundary=0)
          } else{
            hist_part <- geom_histogram(aes(y=..count../sum(..count..)), fill=fill_col, col=graph_par$line_col, binwidth=binwidth, boundary=0)
          }
        } else{
          if(count){
            hist_part <- geom_histogram(fill=fill_col, col=graph_par$line_col, bins=bins, boundary=0)
          } else{
            hist_part <- geom_histogram(aes(y=..count../sum(..count..)), fill=fill_col, col=graph_par$line_col, bins=bins, boundary=0)
          }
        }
        plt <- ggplot(df, aes(x, weight=weight)) + hist_part + grid_settings + scalex + scaley
    }

    # Add axis titles
    plt <- plt + theme(axis.title.y = element_text(size = graph_par$y_title_size, vjust = graph_par$y_title_move), 
                        axis.text.y = element_text(size = graph_par$axis_size),
                        axis.text.x = element_text(size = graph_par$axis_size), 
                        axis.title.x = element_text(size = graph_par$x_title_size)) 

    # Main title
    if(!is.null(main_title)){
        plt <- plt + ggtitle(main_title) + theme(plot.title = element_text(size = graph_par$main_title_size, hjust=0.5, face="bold"))
    }
  
    # Vertical line 
    if(vline){
        if(stat == "median"){
            if(length(weight)==1){
                vloc <- wtd.quantile(x, NULL, 0.5)
            } else{
                vloc <- wtd.quantile(x, weight, 0.5)
            }
        } else if(stat == "mean"){
            if(length(weight)==1){
                vloc <- wtd.mean(x, NULL)
            } else{
                vloc <- wtd.mean(x, weight)
            }
        } else{
            stop("In Custom.Hist, Unknown stat type")
        }
        plt <- plt + geom_vline(xintercept=vloc, col=graph_par$vline_col, size=graph_par$stat_line_width)
    }
    
    # Annotate
    if(vline & annotate){
        build  <- ggplot_build(plt)
        ranges <- build$layout$panel_ranges[[1]]
        xrange <- ranges$x.range
        yrange <- ranges$y.range
        
        yloc   <- graph_par$stat_text_height*yrange[2] + (1-graph_par$stat_text_height)*yrange[1]
        
        # If the median value is in the leftmost 25% of the x axis:
        align_text_left <- (vloc < (xrange[1]*(1-graph_par$stat_text_xalign) + xrange[2]*graph_par$stat_text_xalign))
        xloc <- vloc*(ifelse(align_text_left, graph_par$stat_move, 2-graph_par$stat_move))

        plt <- plt + annotate("text", x=xloc, y = yloc, label = round(vloc, graph_par$round), 
                              hjust = ifelse(align_text_left, 0, 1),
                              size = graph_par$label_size / ggplot2:::.pt, fontface="bold")
    }

    return(plt)
}















