
# --------------------------------- #
# Custom Histogram Function
# --------------------------------- #

library(ggplot2)
library(Hmisc)
library(gridExtra)
library(grid)
library(scales)

Plot.Hist <- function(data,
                      selected_column    = NA,
                      strategy_hist      = TRUE,
                      selected_universe  = NA,
                      normalize_selected = FALSE,
                      vline_stat         = c("mean",
                                             "median"),
                      vline_text         = TRUE,
                      binwidth,
                      x_breaks,
                      x_name,
                      y_breaks,
                      y_name,
                      y_type,
                      emphasize_color,
                      compare_color,
                      transparency,
                      save_to_file = TRUE,
                      output_location = NA,
                      text_size = 13) {
  if (strategy_hist) {
    coef <- data$c
    se <- data$s
    
    data <- data[!is.na(coef)]
    # print(nrow(data))
    coef <- coef[!is.na(coef)]
    
    if (selected_universe == "significant") {
      condition <- data$p < 0.05
    } else if (selected_universe == "positive_p0.005") {
      condition <- data$p < 0.005 & data$c > 0
    } else if (selected_universe == "positive_significant") {
      condition <- data$p < 0.05 & data$c > 0
    }
    
    x <- coef[condition==T]
    x <- x[!is.na(x)]
    
    if (vline_stat == "median") {
      selected_vloc <- median(x, na.rm = T)
      full_vloc     <- median(coef, na.rm = T)
      
    } else if (vline_stat == "mean") {
      selected_vloc <- mean(x, na.rm = T)
      full_vloc     <- mean(coef, na.rm = T)
    }
    
    xbreaks <- round(x_breaks, 2)
    xmin <- min(xbreaks)
    xmax <- max(xbreaks)
    scalex <- scale_x_continuous(breaks = xbreaks, 
                                 limits = c(xmin, xmax), 
                                 name = x_name)
    
    
    
    ybreaks <- y_breaks
    ymin <- min(ybreaks)
    ymax <- max(ybreaks)
    scaley <- scale_y_continuous(name = y_name,
                                 breaks = ybreaks,
                                 limits = c(ymin, ymax))
    
    df <- data.frame(x=x)
    
    if (normalize_selected) {
      plt <- ggplot() + 
        geom_histogram(data = data.table(c = coef),
                       aes(c,
                           y = ..count../sum(..count..)),
                       fill  = compare_color,
                       alpha = transparency,
                       binwidth = binwidth,
                       col = "grey60",
                       boundary = 0) +
        geom_histogram(data = data.table(c = x),
                       aes(c,
                           y = ..count../sum(..count..)),
                       fill  = emphasize_color,
                       alpha = transparency,
                       binwidth = binwidth,
                       col = "grey60",
                       boundary = 0)
      
      if (y_type == "percent") {
        plt <- ggplot() + 
          geom_histogram(data = data.table(c = coef),
                         aes(c,
                             y = ..count../sum(..count..)*100),
                         fill  = compare_color,
                         alpha = transparency,
                         binwidth = binwidth,
                         col = "grey60",
                         boundary = 0) +
          geom_histogram(data = data.table(c = x),
                         aes(c,
                             y = ..count../sum(..count..)*100),
                         fill  = emphasize_color,
                         alpha = transparency,
                         binwidth = binwidth,
                         col = "grey60",
                         boundary = 0)
      }
      
      if (vline_text) {
        plt <- plt + 
          geom_vline(xintercept = full_vloc,
                     color      = compare_color,
                     size       = 0.6) +
          geom_text(data = data.table(text = full_vloc),
                    mapping = aes(x = text,
                                  y = 0.9 * ymax,
                                  label = as.character(round(text, 4))),
                    hjust = 1.1,
                    vjust = 1,
                    size  = (text_size / 3)) + 
          geom_vline(xintercept = selected_vloc,
                     color      = emphasize_color,
                     size       = 0.6) +
          geom_text(data = data.table(text = selected_vloc),
                    mapping = aes(x = text,
                                  y = 0.9 * ymax,
                                  label = as.character(round(text, 4))),
                    hjust = -0.1,
                    vjust = 1,
                    size  = (text_size / 3))
      } else {
        
        plt <- plt + 
          geom_vline(xintercept = full_vloc,
                     color      = compare_color,
                     size       = 0.6) +
          geom_vline(xintercept = selected_vloc,
                     color      = emphasize_color,
                     size       = 0.6)
      }
      
      
    } else {
      plt <- ggplot() + 
        geom_histogram(data = data.table(c = coef),
                       aes(c,
                           y = ..count../sum(..count..),
                           fill = condition),
                       alpha = transparency,
                       binwidth = binwidth,
                       col = "grey60",
                       boundary = 0) +
        scale_fill_manual(values = c(compare_color, emphasize_color))
      
      if (y_type == "percent") {
        plt <- ggplot() + 
          geom_histogram(data = data.table(c = coef),
                         aes(c,
                             y = ..count../sum(..count..)*100,
                             fill = condition),
                         alpha = transparency,
                         binwidth = binwidth,
                         col = "grey60",
                         boundary = 0) +
          scale_fill_manual(values = c(compare_color, emphasize_color))
      }
      
      if (vline_text) {
        plt <- plt + 
          geom_vline(xintercept = full_vloc,
                     color      = "red",
                     size       = 0.6) +
          geom_text(data = data.table(text = full_vloc),
                    mapping = aes(x = text,
                                  y = 0.9 * ymax,
                                  label = as.character(round(text, 4))),
                    hjust = -0.1,
                    vjust = 1,
                    size  = (text_size / 3))
      } else {
        plt <- plt + 
          geom_vline(xintercept = full_vloc,
                     color      = "red",
                     size       = 0.6)
      }
      
    }
    
  } else {
    x <- data[, get(selected_column)]
    x <- x[!is.na(x)]
    
    if (vline_stat == "median") {
      full_vloc     <- median(x, na.rm = T)
    } else if (vline_stat == "mean") {
      full_vloc     <- mean(x, na.rm = T)
    }
    
    xbreaks <- round(x_breaks, 2)
    xmin <- min(xbreaks)
    xmax <- max(xbreaks)
    scalex <- scale_x_continuous(breaks = xbreaks, 
                                 limits = c(xmin, xmax), 
                                 name   = x_name)
    
    
    
    ybreaks <- y_breaks
    ymin <- min(ybreaks)
    ymax <- max(ybreaks)
    scaley <- scale_y_continuous(name   = y_name,
                                 breaks = ybreaks,
                                 limits = c(ymin, ymax))
    
    df <- data.frame(x=x)
    
    plt <- ggplot() + 
      geom_histogram(data = data.table(c = x),
                     aes(c,
                         y = ..count../sum(..count..)),
                     alpha = transparency,
                     binwidth = binwidth,
                     fill = emphasize_color,
                     col = "grey60",
                     boundary = 0)
    
    if (y_type == "percent") {
      plt <- ggplot() + 
        geom_histogram(data = data.table(c = x),
                       aes(c,
                           y = ..count../sum(..count..)*100),
                       alpha = transparency,
                       binwidth = binwidth,
                       fill = emphasize_color,
                       col = "grey60",
                       boundary = 0)
    }
    
    
    if (vline_text) {
      plt <- plt + 
        geom_vline(xintercept = full_vloc,
                   color      = "red",
                   size       = 0.6) +
        geom_text(data = data.table(text = full_vloc),
                  mapping = aes(x = text,
                                y = 0.9 * ymax,
                                label = as.character(round(text, 3))),
                  hjust = -0.1,
                  vjust = 1,
                  size  = (text_size / 3))
    } else {
      plt <- plt + 
        geom_vline(xintercept = full_vloc,
                   color      = "red",
                   size       = 0.6)
    }
  }
  
  
  plt <- plt + 
    scalex + 
    scaley +
    theme_bw(base_size = text_size) %+replace%
    theme(panel.background = element_blank(),
          panel.grid.minor = element_blank(),
          panel.grid.major = element_line(colour = "grey90"),
          axis.text        = element_text(size = text_size),
          axis.title       = element_text(size = text_size),
          axis.title.x     = element_text(size = text_size),
          legend.key       = element_blank(),
          legend.position  = "none",
          panel.border     = element_blank(),
          axis.line        = element_line())
  
  if (save_to_file) {
    
    pdf(output_location,
        width = 6.4, height = 6)
    
    print(plt)
    
    dev.off()
    
  } else {
    
    print(plt)
  }
}

Plot.Scatter <- function(x_data,
                         y_data,
                         merge_cols,
                         x_breaks,
                         x_name,
                         y_breaks = NA,
                         y_name,
                         outline_color   = "black",
                         fill_color      = "violetred1",
                         point_size      = 2.3 * 0.8,
                         transparency,
                         save_to_file    = TRUE,
                         output_location = NA,
                         text_size = 13) {
  data <- merge(x_data,
                y_data,
                by = merge_cols)
  
  coef.1 <- data$c.x
  coef.2 <- data$c.y
  
  data <- data[!is.na(c.x)&!is.na(c.y)]

  coef.1 <- coef.1[!is.na(coef.1)]
  coef.2 <- coef.2[!is.na(coef.2)]
  
  xbreaks <- round(x_breaks, 2)
  xmin <- min(xbreaks)
  xmax <- max(xbreaks)
  scalex <- scale_x_continuous(breaks = xbreaks, 
                               limits = c(xmin, xmax), 
                               name   = x_name)
  
  if (is.na(y_breaks)) {
    y_breaks <- copy(x_breaks)
  }
  
  ybreaks <- y_breaks
  ymin <- min(ybreaks)
  ymax <- max(ybreaks)
  scaley <- scale_y_continuous(name   = y_name,
                               breaks = ybreaks,
                               limits = c(ymin, ymax))
  
  plt <- ggplot(data = data) +
    geom_abline(intercept = 0, 
                slope = 1,
                color = "gray72",
                size  = 1) +
    geom_point(aes(x = c.x,
                   y = c.y),
               shape = 21,
               color = outline_color,
               fill  = fill_color,
               size  = point_size,
               stroke = 0.2,
               alpha = transparency) +
    scalex + 
    scaley +
    theme_bw(base_size = text_size) %+replace%
    theme(panel.background = element_blank(),
          panel.grid.minor = element_blank(),
          panel.grid.major = element_line(colour = "grey90"),
          axis.text        = element_text(size = text_size),
          axis.title       = element_text(size = text_size),
          axis.title.x     = element_text(size = text_size),
          legend.key       = element_blank(),
          legend.position  = "none",
          panel.border     = element_blank(),
          axis.line        = element_line())
  
  if (save_to_file) {
    
    pdf(output_location,
        width = 6.4, height = 6)
    
    print(plt)
    
    dev.off()
    
  } else {
    
    print(plt)
  }
}
