# --------------------------------- #
# ROI CI graph
# --------------------------------- #

rm(list = ls())
library(data.table)
library(bit64)
library(lubridate)
library(Hmisc)
library(ggplot2)
library(ggExtra)

source("./Generic-Plot-Settings-v2.R")

Hist.plt <- function(data,
                     output_dir = NA,
                     y_limit = 30,
                     save_to_file = F) {
   data <- data[Cost==1]
   
   data <- data[Margin %in% c(0.2, 0.3, 0.4) | Margin == "0.3"]
   data[CI_lower > 0,  type := "Positive & p < 0.05"]
   data[CI_higher < 0, type := "Negative & p < 0.05"]
   data[CI_lower <= 0 & CI_higher >= 0, type := "CI includes ROI > 0"]
   
   data[, count := 1]
   data[, sum_group := sum(count),
           by = .(Margin)]
   
   data[, freq := count/sum_group * 100]
   
   median_table = unique(data[, median := median(ROI, na.rm=T),
                                 by = .(Margin)][, 
                                                 .(Margin, median)])
   
   data[, outlier := NULL]
   
   outlier_num = unique(data[, outlier := sum(ifelse(ROI < -500 | ROI > 500,
                                                        1,
                                                        0),
                                                 na.rm = T),
                                by = .(Margin)][, 
                                                .(Margin, outlier)])
   
   outlier_num = c("0.15" = "Margin m = 0.15", 
                   "0.2" = "Margin m = 0.2", 
                   "0.25" = "Margin m = 0.25", 
                   "0.3" = "Margin m = 0.3", 
                   "0.35" = "Margin m = 0.35", 
                   "0.4" = "Margin m = 0.4")
   
   plt <- ggplot(data = data) +
      geom_vline(mapping = aes(xintercept = 0),
                 color = "grey60",
                 linetype = "dotdash") +
      geom_histogram(aes(x = ROI,
                         weight = freq,
                         fill = type),
                     color = "grey60",
                     boundary = 0,
                     binwidth = 50) + 
      geom_vline(data = median_table,
                 mapping = aes(xintercept = median),
                 color = "violetred3") +
      geom_text(data = median_table,
                mapping = aes(x = median,
                              y = 0.9 * y_limit,
                              label = as.character(round(median,3))),
                hjust = -0.1,
                vjust = 1,
                size  = 13/3,
                inherit.aes = F) +
      scale_x_continuous("ROI (%)",
                         limits = c(-500.01, 500.01),
                         breaks = seq(-500, 500, 250)) +
      scale_y_continuous("Percent",
                         limits = c(0, y_limit),
                         breaks = seq(0, y_limit, 10)) +
      facet_wrap(~ Margin,
                 nrow = 1,
                 labeller = as_labeller(outlier_num)) +
      scale_fill_manual(values = c("Positive & p < 0.05" = "violetred3",
                                   "Negative & p < 0.05" = "gray88",
                                   "CI includes ROI > 0"    = "lightsteelblue2")) +
      labs(title = "") +
      theme_bw(base_size = 13) %+replace%
      theme(panel.background = element_blank(),
            panel.grid.minor = element_blank(),
            panel.grid.major = element_line(colour = "grey90"),
            axis.text        = element_text(size = 13),
            axis.title       = element_text(size = 13),
            axis.title.x     = element_text(size = 13),
            legend.key       = element_blank(),
            axis.line        = element_line()) +
      theme(strip.background = element_rect(fill = "gray96"),
            strip.text       = element_text(size = 13),
            panel.spacing    = unit(0.2, "in"),
            legend.title     = element_blank(),
            legend.position  = "bottom")
   
   if (save_to_file) {
      pdf(output_dir,
          height = 5,
          width  = 12)
      
      print(plt)
      
      dev.off()
   }
   
}

# ----------------------------------------------------------------
# Overall average
# ----------------------------------------------------------------
load("../../8-ROI-Analyses/Overall/Processed-Data/base-avg-ROI-overall-observed.RData")
load("../../8-ROI-Analyses/Overall/Processed-Data/base-avg-overall-std-err-observed.RData")

data <- merge(avg_overall_observed[, Margin := as.character(Margin)],
              avg_overall_std_err[, Margin := as.character(Margin)],
              by = c("Module", "Brand", "Margin", "Cost", "Cost.value"))

data[, `:=` (ROI    = ROI * 100,
             ROI_se = ROI_se * 100)]

data[, CI_lower  := ROI - 1.96*ROI_se]
data[, CI_higher := ROI + 1.96*ROI_se]

data[CI_lower < -2000, CI_lower := -2000]
data[CI_higher > 2000, CI_higher:= 2000]

data <- data[complete.cases(data)]

Hist.plt(data,
         save_to_file = T,
         output_dir = "../Figures/fig3_b.pdf")


# ----------------------------------------------------------------
# Weekly Average
# ----------------------------------------------------------------
load("../../8-ROI-Analyses/Weekly/Processed-Data/base-average-ROI-observed.RData")
load("../../8-ROI-Analyses/Weekly/Processed-Data/base-average-std-err-observed.RData")

mean_avg_observed <- average_observed[, .(ROI = mean(ROI)),
                      by = .(Module, Brand, Margin, Cost)]
mean_avg_std_err  <- average_std_err_observed[, .(ROI_se = mean(ROI_se)),
                                              by = .(Module, Brand, Margin, Cost, Beta, Beta_se)]


data <- merge(mean_avg_observed[, Margin := as.character(Margin)],
              mean_avg_std_err[, Margin := as.character(Margin)],
              by = c("Module", "Brand", "Margin", "Cost"))

data[, `:=` (ROI    = ROI * 100,
             ROI_se = ROI_se * 100)]

data[, CI_lower  := ROI - 1.96*ROI_se]
data[, CI_higher := ROI + 1.96*ROI_se]

data[CI_lower < -2000, CI_lower := -2000]
data[CI_higher > 2000, CI_higher:= 2000]

Hist.plt(data,
         save_to_file = T,
         y_limit = 50,
         output_dir = "../Figures/fig3_a.pdf")
